//##************************
//##* This program is free software; you can use it, redistribute it and/or modify it
//##* under the terms of the GNU General Public License version 2 as published by the
//##* Free Software Foundation. The full text of the GNU General Public License
//##* version 2 can be found in the file named 'COPYING' that accompanies this
//##* program. This source code is (C)copyright Geoffrey French 1999-2007.
//##************************
#ifndef BITLIST_H__
#define BITLIST_H__

#include <memory.h>

#include <assert.h>

#define UNIT_SIZE_IN_BITS 32
#define UNIT_SIZE_IN_BYTES 4


#define DIVIDE_BY_UNIT_SIZE_IN_BITS_RIGHT_SHIFT		5
#define REMAINDER_UNIT_SIZE_IN_BITS_MASK 			0x1f
#define MULTIPLY_UNIT_SIZE_IN_BYTES_LEFT_SHIFT		2


#define FULL_MASK		0xffffffff



/********************************************************************************
						BitList
 ********************************************************************************/

class GS_DllExport BitList
{
private:
	typedef unsigned int Unit;

	Unit *m_data;
	int m_size, m_dataSize;


public:
	inline BitList(int sz = 0)
			: m_size( sz )
	{
		if ( m_size != 0 )
		{
			m_dataSize = m_size >> DIVIDE_BY_UNIT_SIZE_IN_BITS_RIGHT_SHIFT;
			if ( ( m_size & REMAINDER_UNIT_SIZE_IN_BITS_MASK )  >  0 )
			{
				m_dataSize++;
			}
			m_data = new Unit[m_dataSize];
		}
		else
		{
			m_data = NULL;
			m_dataSize = 0;
		}

		clear();
	}

	inline BitList(const BitList &b)
			: m_size( b.m_size ), m_dataSize( b.m_dataSize )
	{
		if ( m_size != 0 )
		{
			m_data = new Unit[m_dataSize];
			memcpy( m_data, b.m_data, m_dataSize << MULTIPLY_UNIT_SIZE_IN_BYTES_LEFT_SHIFT );
		}
		else
		{
			m_data = NULL;
		}
	}


	inline ~BitList()
	{
		if ( m_data != NULL )
		{
			delete[] m_data;
			m_data = NULL;
		}
	}


	inline BitList & operator=(const BitList &b)
	{
		if ( m_dataSize != b.m_dataSize )
		{
			if ( m_data != NULL )
			{
				delete[] m_data;
				m_data = NULL;
			}

			m_dataSize = b.m_dataSize;

			if ( m_dataSize != 0 )
			{
				m_data = new Unit[m_dataSize];
			}
			else
			{
				m_data = NULL;
			}
		}

		m_size = b.m_size;
		if ( m_dataSize != 0 )
		{
			memcpy( m_data, b.m_data, m_dataSize << MULTIPLY_UNIT_SIZE_IN_BYTES_LEFT_SHIFT );
		}

		return *this;
	}


	inline bool operator[](int index) const
	{
		return get( index );
	}


	inline bool get(int index) const
	{
		int dataIndex = index >> DIVIDE_BY_UNIT_SIZE_IN_BITS_RIGHT_SHIFT;
		int bitIndex = index & REMAINDER_UNIT_SIZE_IN_BITS_MASK;
		return ( m_data[dataIndex] >> bitIndex )  &  0x1;
	}

	inline void set(int index, bool value) const
	{
		int x = value  ?  1  :  0;
		int dataIndex = index >> DIVIDE_BY_UNIT_SIZE_IN_BITS_RIGHT_SHIFT;
		int bitIndex = index & REMAINDER_UNIT_SIZE_IN_BITS_MASK;
		m_data[dataIndex] = ( m_data[dataIndex] & ~( 0x1 << bitIndex ) )  |  ( x  <<  bitIndex );
	}


	inline bool findSetBit(int start, int end) const
	{
		int startIndex = start >> DIVIDE_BY_UNIT_SIZE_IN_BITS_RIGHT_SHIFT;
		int startBit = start & REMAINDER_UNIT_SIZE_IN_BITS_MASK;

		int endIndex = end >> DIVIDE_BY_UNIT_SIZE_IN_BITS_RIGHT_SHIFT;
		int endBit = end & REMAINDER_UNIT_SIZE_IN_BITS_MASK;

		if ( ( m_data[startIndex] >> startBit )  !=  0 )
		{
			return true;
		}

		for (int i = startIndex + 1; i < endIndex; i++)
		{
			if ( m_data[i] != 0 )
			{
				return true;
			}
		}

		if ( endBit != 0 )
		{
			Unit mask = FULL_MASK  >>  ( UNIT_SIZE_IN_BITS - endBit );
			if ( ( m_data[endIndex] & mask )  !=  0 )
			{
				return true;
			}
		}

		return false;
	}


	inline void clear()
	{
		if ( m_data != NULL )
		{
			memset( m_data, 0, m_dataSize << MULTIPLY_UNIT_SIZE_IN_BYTES_LEFT_SHIFT );
		}
	}

	inline int size() const
	{
		return m_size;
	}
};


#undef UNIT_SIZE_IN_BITS
#undef UNIT_SIZE_IN_BYTES

#undef DIVIDE_BY_UNIT_SIZE_IN_BITS_RIGHT_SHIFT
#undef REMAINDER_UNIT_SIZE_IN_BITS_MASK
#undef MULTIPLY_UNIT_SIZE_IN_BYTES_LEFT_SHIFT

#undef FULL_MASK


#endif
