##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from gSculpt.Util.SignalSlot import ClassSignal

from gSculpt.Math.Math import *
import gSculpt.Math.MathField

from gSculpt.Kernel.DB import Row, FieldColumn

import GSGraphCommandTracker



class GSGraphItemList (object):
	appendSignal = ClassSignal()
	removeSignal = ClassSignal()


	def __init__(self):
		super( GSGraphItemList, self ).__init__()
		self._items = []


	def __contains__(self, item):
		return item in self._items

	def __getitem__(self, index):
		return self._items[index]

	def __len__(self):
		return len( self._items )

	def __iter__(self):
		return iter( self._items )


	def _p_append(self, item):
		self._items.append( item )
		self.appendSignal.emit( item )

	def _p_remove(self, item):
		self.removeSignal.emit( item )
		self._items.remove( item )



class GSGraphPin (object):
	nameSignal = ClassSignal()


	def __init__(self, name):
		super( GSGraphPin, self ).__init__()
		self._name = name
		self._node = None


	def _p_getName(self):
		return self._name

	def _p_setName(self, name):
		self._name = name
		self.nameSignal.emit( self, name )


	def _p_getNode(self):
		return self._node



	name = property( _p_getName, _p_setName )
	node = property( _p_getNode )




class GSGraphSource (GSGraphPin):
	def __init__(self, name):
		super( GSGraphSource, self ).__init__( name )
		self._links = []


	def __len__(self):
		return len( self._links )

	def __getitem__(self, index):
		return self._links[index]

	def __iter__(self):
		return iter( self._links )


	def _p_clearOutputs(self):
		while len( self._links ) > 0:
			link = self._links[-1]
			link.sink._p_removeInputLink( link )


	def _p_addLink(self, link):
		assert link not in self._links, 'link already present in list'
		self._links.append( link )
		self._node._p_addOutputLink( link )

	def _p_removeLink(self, link):
		assert link in self._links, 'link not present in list'
		self._links.remove( link )
		self._node._p_removeOutputLink( link )






class GSGraphSink (GSGraphPin):
	trackerClass = GSGraphCommandTracker.GSGraphSinkCommandTracker

	addInputSignal = ClassSignal()
	removeInputSignal = ClassSignal()


	def __init__(self, name):
		super( GSGraphSink, self ).__init__( name )
		self._commandTracker_ = None


	def checkAddInputForCycles(self, source):
		return self._node._p_checkAddInputForCycles( source._node )


	def addInput(self, source):
		assert self._node is not None, 'no node'
		assert len( self ) < self.maxLen() or  self.maxLen() is None, 'cannot add another input'
		assert source not in self, 'input already present'
		link = GSGraphLink( self._node._graph, source, self )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onSinkInputAdded( self, source, link )
		self.addInputSignal.emit( self, source, link )
		self._node._p_sinkAddInput( self, source )
		return link

	def removeInput(self, source):
		assert self._node is not None, 'no node'
		link = self.getLinkTo( source )
		assert link is not None, 'input not present'
		self.removeInputSignal.emit( self, source, link )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onSinkInputRemoved( self, source, link )
		self._node._p_sinkRemoveInput( self, source )
		link._p_destroy()
		return link

	def _p_removeInputLink(self, link):
		assert self._node is not None, 'no node'
		self.removeInputSignal.emit( self, link.source, link )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onSinkInputRemoved( self, link.source, link )
		self._node._p_sinkRemoveInput( self, link.source )
		link._p_destroy()
		return link


	def _p_clearInputs(self):
		assert self._node is not None, 'no node'
		while len( self ) > 0:
			link = self[-1]
			self.removeInputSignal.emit( self, link.source, link )
			if self._commandTracker_ is not None:
				self._commandTracker_._f_onSinkInputRemoved( self, link.source, link )
			self._node._p_sinkRemoveInput( self, link.source )
			link._p_destroy()


	def canAddInput(self):
		return len( self ) < self.maxLen() or  self.maxLen() is None


	def __contains__(self, sourceOrLink):
		if isinstance( sourceOrLink, GSGraphLink ):
			return self._p_containsLink( sourceOrLink )
		elif isinstance( sourceOrLink, GSGraphSource ):
			return self.getLinkTo( sourceOrLink ) is not None
		else:
			raise TypeError, 'unknown type'

	def __len__(self):
		assert False, 'abstract'

	def index(self, sourceOrLink):
		if isinstance( sourceOrLink, GSGraphLink ):
			return self._p_indexOfLink( sourceOrLink )
		elif isinstance( sourceOrLink, GSGraphSource ):
			return self._p_indexOfSource( sourceOrLink )
		else:
			raise TypeError, 'unknown type'

	def maxLen(self):
		assert False, 'abstract'

	def __getitem__(self, index):
		assert False, 'abstract'

	def inputSource(self, index):
		return self[index]._source

	def __iter__(self):
		assert False, 'abstract'

	def inputSources(self):
		for link in self:
			yield link._source

	def getLinkTo(self, source):
		assert False, 'abstract'

	def copyLinksFrom(self, sink):
		inputSources = list( self.inputSources )
		self._p_clearInputs()
		for source in inputSources:
			self.addInput( source )




	def _p_addLink(self, link):
		assert False, 'abstract'

	def _p_removeLink(self, link):
		assert False, 'abstract'

	def _p_containsLink(self, link):
		assert False, 'abstract'

	def _p_indexOfLink(self, link):
		assert False, 'abstract'

	def _p_indexOfSource(self, source):
		assert False, 'abstract'






class GSGraphSinkSingle (GSGraphSink):
	def __init__(self, name):
		super( GSGraphSinkSingle, self ).__init__( name )
		self._link = None


	def __len__(self):
		if self._link is None:
			return 0
		else:
			return 1

	def maxLen(self):
		return 1

	def __getitem__(self, index):
		if isinstance( index, slice ):
			return [ self._link ][index]
		else:
			if index > 0  or  index < -1:
				raise IndexError
			return self._link

	def __iter__(self):
		if self._link is not None:
			yield self._link

	def getLinkTo(self, source):
		if self._link is None:
			return None
		else:
			if self._link._source is source:
				return self._link
			else:
				return None


	def _p_addLink(self, link):
		assert self._link is None, 'link already set'
		self._link = link
		self._node._p_addInputLink( link )

	def _p_removeLink(self, link):
		assert link is self._link, '@link does not match my link'
		self._link = None
		self._node._p_removeInputLink( link )

	def _p_containsLink(self, link):
		return link is self._link

	def _p_indexOfLink(self, link):
		if link is self._link:
			return 0
		else:
			raise ValueError, 'link not in list'

	def _p_indexOfSource(self, source):
		if self._link._source is source:
			return 0
		else:
			raise ValueError, 'not linked to source'




class GSGraphSinkMultiple (GSGraphSink):
	def __init__(self, name):
		super( GSGraphSinkMultiple, self ).__init__( name )
		self._links = []


	def __len__(self):
		return len( self._links )

	def maxLen(self):
		return None

	def __getitem__(self, index):
		return self._links[index]

	def __iter__(self):
		return iter( self._links )

	def getLinkTo(self, source):
		for link in self._links:
			if link._source is source:
				return link
		return None


	def _p_addLink(self, link):
		assert link not in self._links, 'link already present in list'
		self._links.append( link )
		self._node._p_addInputLink( link )

	def _p_removeLink(self, link):
		assert link in self._links, 'link not present in list'
		self._links.remove( link )
		self._node._p_removeInputLink( link )

	def _p_containsLink(self, link):
		return link in self._links

	def _p_indexOfLink(self, link):
		return self._links.index( link )

	def _p_indexOfSource(self, source):
		for i, link in enumerate( self ):
			if link._source is source:
				return i
		raise ValueError, 'not linked to source'



class GSGraphLink (object):
	def __init__(self, graph, source, sink):
		super( GSGraphLink, self ).__init__()
		self._source = source
		self._sink = sink
		self._graph = graph
		self._graph._p_addLink( self )
		self._source._p_addLink( self )
		self._sink._p_addLink( self )

	def _p_destroy(self):
		self._source._p_removeLink( self )
		self._sink._p_removeLink( self )
		self._graph._p_removeLink( self )
		self._source = None
		self._sink = None
		self._graph = None


	def erase(self):
		self._sink._p_removeInputLink( self )


	def _p_getSource(self):
		return self._source

	def _p_getSink(self):
		return self._sink


	source = property( _p_getSource )
	sink = property( _p_getSink )



class GSGraphNode (Row):
	viewPosition = FieldColumn( Point2, Point2() )
	bHasViewPosition = FieldColumn( bool, False )


	def __init__(self):
		super( GSGraphNode, self ).__init__()
		self._inputLinks = set()
		self._outputLinks = set()
		self._sources = GSGraphItemList()
		self._sinks = GSGraphItemList()
		self._inputNodeCounts = {}
		self._inputNodes = GSGraphItemList()
		self._inputSourceCounts = {}
		self._inputSources = GSGraphItemList()
		self._graph = None


	def addSource(self, source):
		source._node = self
		self._sources._p_append( source )

	def removeSource(self, source):
		self._sources._p_remove( source )
		source._node = None


	def addSink(self, sink):
		sink._node = self
		self._sinks._p_append( sink )
		for source in sink.inputSources():
			self._p_sinkAddInput( sink, source )

	def removeSink(self, sink):
		for source in sink.inputSources():
			self._p_sinkRemoveInput( sink, source )
		self._sinks._p_remove( sink )
		sink._node = None


	def hasOutputLinks(self):
		return len( self._outputLinks ) > 0


	def hasInput(self, node):
		return node in self._inputNodes



	def _p_checkAddInputForCycles(self, node):
		if node is self:
			return True
		for inputNode in node._inputNodes:
			if self._p_checkAddInputForCycles( inputNode ):
				return True
		return False



	def _p_addInputLink(self, link):
		assert link not in self._inputLinks, 'input link already present in list'
		self._inputLinks.add( link )

	def _p_removeInputLink(self, link):
		assert link in self._inputLinks, 'input link not present in list'
		self._inputLinks.remove( link )

	def _p_addOutputLink(self, link):
		assert link not in self._outputLinks, 'output link already present in list'
		self._outputLinks.add( link )

	def _p_removeOutputLink(self, link):
		assert link in self._outputLinks, 'output link not present in list'
		self._outputLinks.remove( link )


	def _p_sinkAddInput(self, sink, source):
		try:
			self._inputNodeCounts[source.node] += 1
		except KeyError:
			self._inputNodeCounts[source.node] = 1
			self._inputNodes._p_append( source.node )

		try:
			self._inputSourceCounts[source] += 1
		except KeyError:
			self._inputSourceCounts[source] = 1
			self._inputSources._p_append( source )


	def _p_sinkRemoveInput(self, sink, source):
		self._inputNodeCounts[source.node] -= 1
		count = self._inputNodeCounts[source.node]
		if count == 0:
			del self._inputNodeCounts[source.node]
			self._inputNodes._p_remove( source.node )

		self._inputSourceCounts[source] -= 1
		count = self._inputSourceCounts[source]
		if count == 0:
			del self._inputSourceCounts[source]
			self._inputSources._p_remove( source )


	def _p_clear(self):
		for source in self._sources:
			source._p_clearOutputs()

		assert len( self._outputLinks ) == 0

		for sink in self._sinks:
			sink._p_clearInputs()

		assert len( self._inputLinks ) == 0


	def _p_getSources(self):
		return self._sources

	def _p_getSinks(self):
		return self._sinks

	def _p_getInputNodes(self):
		return self._inputNodes

	def _p_getInputSources(self):
		return self._inputSources


	sources = property( _p_getSources )
	sinks = property( _p_getSinks )
	inputNodes = property( _p_getInputNodes )
	inputSources = property( _p_getInputSources )





class GSGraph (object):
	trackerClass = GSGraphCommandTracker.GSGraphCommandTracker

	nodeAddedSignal = ClassSignal()
	removeNodeSignal = ClassSignal()
	nodeRemovedSignal = ClassSignal()


	def __init__(self):
		super( GSGraph, self ).__init__()
		self._nodes = []
		self._links = GSGraphItemList()
		self._commandTracker_ = None




	def addNode(self, node):
		node._graph = self
		self._nodes.append( node )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onGraphNodeAdded( self, node )
		self.nodeAddedSignal.emit( self, node )


	def removeNode(self, node):
		self.removeNodeSignal.emit( self, node )
		node._p_clear()
		self._nodes.remove( node )
		node._graph = None
		self.nodeRemovedSignal.emit( self, node )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onGraphNodeRemoved( self, node )


	def _p_addLink(self, link):
		self._links._p_append( link )

	def _p_removeLink(self, link):
		self._links._p_remove( link )


	def _p_getNodes(self):
		return self._nodes

	def _p_getLinks(self):
		return self._links


	nodes = property( _p_getNodes )
	links = property( _p_getLinks )





if __name__ == '__main__':
	graph = GSGraph()

	node1 = GSGraphNode()
	node1.addSource( GSGraphSource( 'out' ) )
	node1.addSink( GSGraphSinkSingle( 'in1' ) )
	node1.addSink( GSGraphSinkSingle( 'in2' ) )
	graph.addNode( node1 )

	node2 = GSGraphNode()
	node2.addSource( GSGraphSource( 'out' ) )
	graph.addNode( node2 )

	node3 = GSGraphNode()
	node3.addSource( GSGraphSource( 'out' ) )
	graph.addNode( node3 )

	node1.sinks[0].addInput( node2.sources[0] )
	node1.sinks[1].addInput( node3.sources[0] )
