##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
import copy
import weakref

from gSculpt.Util import SignalSlot

from gSculpt.FileIO import IOXml

from gSculpt.Kernel.ObjectSourceProxy import ObjectSourceProxy
from gSculpt.Kernel.CachedObjectSource import CachedObjectSource
from gSculpt.Kernel import KMeta
from gSculpt.Kernel.Variable import GenericReadOnlyVarInterface, GenericVarInterface, VarInterfaceClass, VarInterface, VarProxy
from gSculpt.Kernel import DBCommandTracker






#//////////////////////////////////////////////////////////////////////////////
#				FieldBaseClass
#//////////////////////////////////////////////////////////////////////////////

class FieldBaseClass (VarInterfaceClass):
	"""Field base metaclass"""
	def __init__(cls, clsName, clsBases, clsDict):
		"""Constructor
		@clsName - class name
		@clsBases - base classes
		@clsDict - class dict"""

		# Register the class with the XML IO system
		IOXml.ioObjectFactoryRegister( clsName, cls )

		# Can use an XML node of the value type, when reading
		cls.__readxml_readable_types__ = [ cls.valueClass ]

		# Init XML IO
		if cls.valueClass is not None:
			try:
				# Get the @xmlNodeValueReader class attribute
				xmlNodeValueReader = clsDict['xmlNodeValueReader']
			except KeyError:
				# Not found; ensure that the class is XML readable (implements __readxml__)
				if not IOXml.isXmlReadable( cls.valueClass ):
					raise TypeError, 'value class \'%s\' is not XML readable, and no value reader was provided' % ( cls.valueClass.__name__, )
				cls.s_xmlNodeValueReader = None
			else:
				# Rename @xmlNodeValueReader to @s_xmlNodeValueReader
				del cls.xmlNodeValueReader
				cls.s_xmlNodeValueReader = ( xmlNodeValueReader, )

			try:
				# Get the @xmlNodeValueWriter class attribute
				xmlNodeValueWriter = clsDict['xmlNodeValueWriter']
			except KeyError:
				# Not found; ensure that the class is XML writeable (implements __writexml__)
				if not IOXml.isXmlWriteable( cls.valueClass ):
					raise TypeError, 'value class \'%s\' is not XML writeable, and no value writer was provided' % ( cls.valueClass.__name__, )
				cls.s_xmlNodeValueWriter = None
			else:
				# Rename @xmlNodeValueWriter to @s_xmlNodeValueWriter
				del cls.xmlNodeValueWriter
				cls.s_xmlNodeValueWriter = ( xmlNodeValueWriter, )

		super( FieldBaseClass, cls ).__init__( clsName, clsBases, clsDict )






#//////////////////////////////////////////////////////////////////////////////
#				FieldClass
#//////////////////////////////////////////////////////////////////////////////

class FieldClass (FieldBaseClass):
	"""Field metaclass"""
	def __init__(cls, clsName, clsBases, clsDict):
		"""Constructor
		@clsName - class name
		@clsBases - base classes
		@clsDict - class dict"""

		# Ensure that @defaultValue is present
		if not clsDict.has_key( 'defaultValue' ):
			raise AttributeError, 'defaultValue not specified'

		# Ensure that @defaultvalue is an instance of @valueClass
		if cls.valueClass is not None:
			if not isinstance( cls.defaultValue, cls.valueClass )  and  cls.defaultValue is not None:
				raise TypeError, 'defaultValue must be an instance of valueClass'

		super( FieldClass, cls ).__init__( clsName, clsBases, clsDict )





#//////////////////////////////////////////////////////////////////////////////
#				Field
#//////////////////////////////////////////////////////////////////////////////

class Field (VarInterface):
	"""Field class
	Must provide the following class attributes:
	@valueClass - the class of the field value
	@defaultValue - the default value"""

	__metaclass__ = FieldClass
	__slots__ = [ '_value', '_ownerRow', '_name' ]
	valueClass = None
	defaultValue = None



	def __init__(self, src=None, _p_initial=None):
		"""Constructor:
		@src - source of value, either an object of the value type, or a Variable object from which to obtain the value"""
		super( Field, self ).__init__()

		if _p_initial is not None:
			self._value = _p_initial
		else:
			if src is None:
				# Set the value to be a copy of the default value
				initial = copy.copy( self.defaultValue )
			else:
				if src.__class__ == self.valueClass:
					# src is a @valueClass
					initial = copy.copy( src )
				elif src.__class__ == self.__class__:
					# src is a Field
					initial = copy.copy( src._value )
				else:
					# Cannot initialise
					raise VarInitialisationTypeException, 'Cannot initialse variable value; source value type is "%s", value class is "%s", self class is "%s"' % \
													( src.__class__, self.valueClass, self.__class__ )
			self._value = initial

		self._ownerRow = None
		self._name = None



	def _p_isRegistered(self):
		"""Private - Determine if the field is registered"""
		return self._ownerRow is not None


	def isValid(self):
		"""Determine if the variable is valid"""
		return True



	def getValue(self):
		"""Get the field value"""
		self._p_assertRegistered()
		self._ownerRow._f_onGet( self._name )
		return self._value

	def setValue(self, value):
		"""Set the field value
		@value - the new value"""
		self._p_assertRegistered()
		self._o_checkValueType( value )
		self._p_preValueChange( self._value )
		self._o_unrefValue( self._value )
		self._value = value
		self._o_refValue( value )
		self._p_valueChanged( value )



	def _p_preValueChange(self, oldValue):
		"""Private - notify of imminent value change"""
		self._p_assertRegistered()
		self._ownerRow._f_onChildFieldPreModify( self._name )
		super( Field, self )._p_preValueChange( oldValue )

	def _p_valueChanged(self, value):
		"""Private - notify of value change"""
		self._p_assertRegistered()
		self._ownerRow._f_onChildFieldModify( self._name )
		super( Field, self )._p_valueChanged( value )

	def _f_register(self, row, name):
		"""Private - register field
		@row - the row that contains @self
		@name - the field name"""
		self._ownerRow = row
		self._name = name

	def _p_assertRegistered(self):
		"""Private - Assert that the field is registered"""
		assert self._p_isRegistered(), 'field not registered'

	def _o_checkValueType(self, value):
		if not isinstance( value, self.valueClass ):
			raise TypeError, 'value is not a subclass of the value class of this field'



	def __readxml__(self, xmlNode):
		"""Read value from an XML node
		@xmlNode - the XML node"""
		if xmlNode.isValid():
			if self.s_xmlNodeValueReader is not None:
				if self.s_xmlNodeValueReader[0] is not None:
					self.setValue( self.s_xmlNodeValueReader[0]( xmlNode, self.getValue() ) )
			else:
				v = self.getValue()
				xmlNode  >>  v
				self.setValue( v )

	def __writexml__(self, xmlNode):
		"""Write value to an XML node
		@xmlNode - the XML node"""
		if xmlNode.isValid():
			if self.s_xmlNodeValueWriter is not None:
				if self.s_xmlNodeValueWriter[0] is not None:
					self.s_xmlNodeValueWriter[0]( xmlNode, self.getValue() )
			else:
				xmlNode  <<  self.getValue()



	def __copy__(self):
		"""Copy helper"""
		c = self.__class__()
		c._value[0] = self._value[0]
		return c

	def __deepcopy__(self, memo):
		"""Deep copy helper"""
		c = self.__class__()
		c._value[0] = copy.deepcopy( self._value[0], memo )
		return c


	value = property( getValue, setValue )






#//////////////////////////////////////////////////////////////////////////////
#			ObjectField
#//////////////////////////////////////////////////////////////////////////////

class ObjectField (Field):
	"""Object field class"""
	valueClass = None
	defaultValue = None


	def __init__(self, src=None):
		"""Constructor:
		@src - source of value, either an object of the value type, or a Variable object from which to obtain the value"""

		if src is None:
			super( ObjectField, self ).__init__( _p_initial=self.defaultValue )
		else:
			if isinstance( src, self.valueClass ):
				super( ObjectField, self ).__init__( _p_initial=src )
			elif isinstance( src, ObjectField ):
				super( ObjectField, self ).__init__( _p_initial=src.getValue() )



	def _o_checkValueType(self, value):
		if value is not None:
			if not isinstance( value, self.valueClass ):
				raise TypeError, 'value is not a subclass of the value class of this field'



	def __readxml__(self, xmlNode):
		"""Read value from an XML node
		@xmlNode - the XML node"""
		if xmlNode.isValid():
			self.setValue( xmlNode.readObject() )

	def __writexml__(self, xmlNode):
		"""Write value to an XML node
		@xmlNode - the XML node"""
		if xmlNode.isValid():
			xmlNode.writeObject( self.getValue() )






#//////////////////////////////////////////////////////////////////////////////
#			Primitive field types
#//////////////////////////////////////////////////////////////////////////////

class BoolField (Field):
	valueClass = bool
	defaultValue = False
	xmlNodeValueReader = IOXml.ioXmlReadBoolNode
	xmlNodeValueWriter = IOXml.ioXmlWriteBoolNode

class IntField (Field):
	valueClass = int
	defaultValue = 0
	xmlNodeValueReader = IOXml.ioXmlReadIntNode
	xmlNodeValueWriter = IOXml.ioXmlWriteIntNode

class FloatField (Field):
	valueClass = float
	defaultValue = 0.0
	xmlNodeValueReader = IOXml.ioXmlReadFloatNode
	xmlNodeValueWriter = IOXml.ioXmlWriteFloatNode

class StringField (Field):
	valueClass = str
	defaultValue = ''
	xmlNodeValueReader = IOXml.ioXmlReadStringNode
	xmlNodeValueWriter = IOXml.ioXmlWriteStringNode




#//////////////////////////////////////////////////////////////////////////////
#			Field type registry
#//////////////////////////////////////////////////////////////////////////////

g_fieldTypeRegistry = {}


def registerFieldType(valueClass, t):
	"Register a field type by value class; @name is typically the class of the value type (e.g. 'int')"
	global g_fieldTypeRegistry
	g_fieldTypeRegistry[valueClass] = t

def getFieldType(valueClass):
	"Get a field type by name"
	global g_fieldTypeRegistry
	if issubclass( valueClass, Row ):
		return RowRefField
	else:
		return g_fieldTypeRegistry[valueClass]


registerFieldType( bool, BoolField )
registerFieldType( int, IntField )
registerFieldType( float, FloatField )
registerFieldType( str, StringField )







#//////////////////////////////////////////////////////////////////////////////
#		ObjectSourceProxyField
#//////////////////////////////////////////////////////////////////////////////

class ObjectSourceProxyField (ObjectSourceProxy):
	"""Proxy field
	Adds registration to @VarProxy"""
	def __init__(self, target=None):
		"""Constructor
		@target - the target variable of the proxy"""
		super( ObjectSourceProxyField, self ).__init__( target )

		self._ownerRow = None
		self._name = ''


	def _f_register(self, ownerRow, name):
		"""Friendly - register the field"""
		self._ownerRow = ownerRow
		self._name = name


	def getValue(self):
		"""Get the value"""
		v = super( ObjectSourceProxyField, self ).getValue()
		self._ownerRow._f_onGet( self._name )
		if isinstance( v, Row ):
			self._ownerRow._f_onGetExternal( v )
		return v






#//////////////////////////////////////////////////////////////////////////////
#				VarProxyField
#//////////////////////////////////////////////////////////////////////////////

class VarProxyField (VarProxy):
	"""Proxy field
	Adds registration to @VarProxy"""
	def __init__(self, target=None):
		"""Constructor
		@target - the target variable of the proxy"""
		super( VarProxyField, self ).__init__( target )

		self._ownerRow = None
		self._name = ''


	def _f_register(self, ownerRow, name):
		"""Friendly - register the field"""
		self._ownerRow = ownerRow
		self._name = name


	def getValue(self):
		"""Get the value"""
		v = super( VarProxyField, self ).getValue()
		self._ownerRow._f_onGet( self._name )
		if isinstance( v, Row ):
			self._ownerRow._f_onGetExternal( v )
		return v





#//////////////////////////////////////////////////////////////////////////////
#				FunctionField
#//////////////////////////////////////////////////////////////////////////////

class FunctionField (CachedObjectSource):
	"""Function field
	Computes its value using a supplied function"""
	def __init__(self, function):
		"""Constructor
		@function - a method of the row that owns this field, whose result will be returned by getValue()"""
		super( FunctionField, self ).__init__()

		self._function = function

		self._ownerRow = None
		self._name = ''
		self._paramRowsAndFieldNames = set()
		self._paramOwnerRows = set()


	def getValue(self):
		"""Get the value"""
		value = super( FunctionField, self ).getValue()
		self._ownerRow._f_onGet( self._name )
		return value


	def _f_register(self, ownerRow, name):
		"""Friendly - register the field"""
		self._ownerRow = ownerRow
		self._name = name
		self._ownerRow.fieldModifiedSignal.connect( self._p_onFieldModified )
		self._paramOwnerRows = set( [ self._ownerRow ] )


	def _p_evaluateSource(self):
		"""Evaluate"""

		self._paramRowsAndFieldNames = set()
		self._paramOwnerRows = set( [ self._ownerRow ] )


		# Add get listeners
		self._ownerRow.addGetListener( self._p_onRowGet )
		self._ownerRow.addGetExternalListener( self._p_onRowGetExternal )

		# Invoke the function
		result = self._function( self._ownerRow )

		# Remove get listeners
		for row in self._paramOwnerRows:
			row.removeGetListener( self._p_onRowGet )
			row.removeGetExternalListener( self._p_onRowGetExternal )

		# Disconnect from external row signals
		for _row in self._paramOwnerRows:
			if row is not self._ownerRow:
				row.fieldModifiedSignal.disconnect( self._p_onFieldModified )

		SignalSlot.unblockSlot( self._p_onFieldModified )

		return result


	def _p_onRowGet(self, row, fieldName):
		self._paramRowsAndFieldNames.add( ( row, fieldName ) )

	def _p_onRowGetExternal(self, externalRow):
		if externalRow not in self._paramOwnerRows:
			self._paramOwnerRows.add( externalRow )
			externalRow.fieldModifiedSignal.connect( self._p_onFieldModified )
			externalRow.addGetExternalListener( self._p_onRowGetExternal )
			externalRow.addGetListener( self._p_onRowGet )



	def _p_onFieldModified(self, row, fieldName):
		if row in self._paramOwnerRows:
			if ( row, fieldName ) in self._paramRowsAndFieldNames:
				SignalSlot.blockSlot( self._p_onFieldModified )
				self._p_invalidateCache()
				self._ownerRow._f_onFieldModify( self._name )



	value = property( getValue )








#//////////////////////////////////////////////////////////////////////////////
#		ObjectSourceColumnInterface
#//////////////////////////////////////////////////////////////////////////////

class ObjectSourceColumnInterface (object):
	def _f_getObjectSourceFromInstance(self, instance):
		"""Get the object source corresponding to @self from @instance
		@instance - an instance of @Row"""
		assert False, 'abstract'

	def _f_getValueFromInstance(self, instance):
		"""Get the value of the variable managed by @self in @instance"""
		assert False, 'abstract'






#//////////////////////////////////////////////////////////////////////////////
#		ObjectSourceColumnBase
#//////////////////////////////////////////////////////////////////////////////

class ObjectSourceColumnBase (KMeta.KMetaMember, ObjectSourceColumnInterface):
	"""Field column base class
	NOTE: subclasses: NEVER use self as a key for retrieving data incase the column is overridden"""
	def __init__(self, doc=''):
		KMeta.KMetaMember.__init__( self, doc )
		ObjectSourceColumnInterface.__init__( self )
		self._fieldAttrName = None


	def _f_metaMember_initMetaMember(self, cls, name):
		"""Initialise a meta-member; called by the metaclass constructor
		@cls - the class that this meta-member is part of
		@name - the attribute name of this meta-member"""
		super( ObjectSourceColumnBase, self )._f_metaMember_initMetaMember( cls, name )
		self._fieldAttrName = intern( ObjectSourceColumnBase._p_computeFieldAttrName( name ) )


	@staticmethod
	def _p_computeFieldAttrName(name):
		"""Private - compute the full field attribute name"""
		return '_field_' + name






#//////////////////////////////////////////////////////////////////////////////
#			FieldColumnInterface
#//////////////////////////////////////////////////////////////////////////////

class FieldColumnInterface (ObjectSourceColumnInterface):
	def _f_getObjectSourceFromInstance(self, instance):
		"""Get the object source corresponding to @self from @instance
		@instance - an instance of @Row"""
		return self._f_getVarFromInstance( instance )

	def _f_getVarFromInstance(self, instance):
		"""Get the variable interface corresponding to @self from @instance
		@instance - an instance of @Row"""
		assert False, 'abstract'

	def _f_setValueInInstance(self, instance, value):
		"""Set the value of the variable managed by @self in @instance"""
		assert False, 'abstract'





#//////////////////////////////////////////////////////////////////////////////
#			FieldColumnBase
#//////////////////////////////////////////////////////////////////////////////

class FieldColumnBase (ObjectSourceColumnBase, FieldColumnInterface):
	"""Field column base class
	NOTE: subclasses: NEVER use self as a key for retrieving data incase the column is overridden"""
	pass






#//////////////////////////////////////////////////////////////////////////////
#			FieldColumnBaseWithXml
#//////////////////////////////////////////////////////////////////////////////

class FieldColumnBaseWithXml (FieldColumnBase):
	"""Field column base class
	Adds XML serialisation interface to @FieldColumnBase"""
	def __init__(self, xmlName=None, doc=''):
		super( FieldColumnBaseWithXml, self ).__init__( doc )
		self._xmlName = xmlName



	def _f_metaMember_initMetaMember(self, cls, name):
		"""Friends - initialise the column
		@name - the column name"""
		super( FieldColumnBaseWithXml, self )._f_metaMember_initMetaMember( cls, name )
		if self._xmlName is None:
			self._xmlName = self._o_computeXmlName( name )


	def _f_readInstanceFieldXml(self, instance, instanceXmlNode):
		"""Read the field managed by @self in @instance from an XML node"""
		pass

	def _f_writeInstanceFieldXml(self, instance, instanceXmlNode):
		"""Write the field managed by @self in @instance to an XML node"""
		pass






#//////////////////////////////////////////////////////////////////////////////
#			FieldColumn
#//////////////////////////////////////////////////////////////////////////////

class FieldColumn (FieldColumnBaseWithXml):
	def __init__(self, valueType, initialValue, xmlName=None, bWriteIfDefault=True, doc=''):
		"""Constructor
		@valueType - the value type
		@initialValue - the initialValue
		@xmlName - the xml node name [None]
		@doc - doc string ['']"""
		super( FieldColumn, self ).__init__( xmlName, doc )
		self._valueType = valueType
		self._initialValue = initialValue
		self._bWriteIfDefault = bWriteIfDefault
		self._fieldType = getFieldType( valueType )



	def _f_metaMember_initInstance(self, instance, srcInstance=None):
		"""Initialise instance
		@instance - instance to initialise
		@srcInstance - source instance to copy values from"""
		super( FieldColumn, self )._f_metaMember_initInstance( instance, srcInstance )
		if srcInstance is None:
			field = self._fieldType( self._initialValue )
		else:
			field = self._fieldType( self._f_getValueFromInstance( srcInstance ) )
		setattr( instance, self._fieldAttrName, field )
		field._f_register( instance, self._name )


	def __get__(self, obj, objtype = None):
		"""Descriptor - get"""
		if obj is not None:
			# Return the field value
			return self._f_getValueFromInstance( obj )
		else:
			# Return the field declaration
			return self

	def __set__(self, obj, value):
		"""Descriptor - set"""
		self._f_setValueInInstance( obj, value )

	def __delete__(self, obj):
		"""Descriptor - delete"""
		raise TypeError, 'cannot delete field value'


	def _f_getVarFromInstance(self, instance):
		"""Get the variable interface corresponding to @self from @instance
		@instance - an instance of @Row"""
		return getattr( instance, self._fieldAttrName )



	def _f_getValueFromInstance(self, instance):
		"""Get the value of the field managed by @self in @instance"""
		return getattr( instance, self._fieldAttrName ).getValue()


	def _f_setValueInInstance(self, instance, value):
		"""Set the value of the field managed by @self in @instance"""
		getattr( instance, self._fieldAttrName ).setValue( value )



	def _f_readInstanceFieldXml(self, instance, instanceXmlNode):
		"""Read the field managed by @self in @instance from an XML node"""
		super( FieldColumn, self )._f_readInstanceFieldXml( instance, instanceXmlNode )
		instanceXmlNode.getChild( self._xmlName )  >>  self._f_getVarFromInstance( instance )

	def _f_writeInstanceFieldXml(self, instance, instanceXmlNode):
		"""Write the field managed by @self in @instance to an XML node"""
		super( FieldColumn, self )._f_writeInstanceFieldXml( instance, instanceXmlNode )
		if self._bWriteIfDefault  or  self._f_getValueFromInstance( instance )  !=  self._initialValue:
			instanceXmlNode.addChild( self._xmlName )  <<  self._f_getVarFromInstance( instance )





#//////////////////////////////////////////////////////////////////////////////
#			CompositeColumn
#//////////////////////////////////////////////////////////////////////////////

class CompositeColumn (FieldColumnBaseWithXml):
	"""Field declaration"""
	def __init__(self, targetRowClass, xmlName=None, doc=''):
		"""Constructor
		@rowClass - the row class
		@xmlName - the xml node name [None]
		@doc - doc string ['']
		@traits - field traits [None]"""
		super( CompositeColumn, self ).__init__( xmlName, doc )
		assert isinstance( targetRowClass, RowClass )  and  issubclass( targetRowClass, Row ), 'targetRowClass must be an instance of RowClass and a subclass of Row'
		self._targetRowClass = targetRowClass




	def _f_metaMember_overload(self, superMetaMember, superClass):
		if not issubclass( self._targetRowClass, superMetaMember._targetRowClass ):
			raise ValueError, 'cannot overload composite column %s.%s; the target row class (%s) must be a subclass of the base column target row class (%s)' %		\
							( self._cls.__name__, self._name, self._targetRowClass.__name__, superMetaMember._targetRowClass.__name__ )


	def _f_metaMember_initInstance(self, instance, srcInstance=None):
		"""Initialise instance
		@instance - instance to initialise
		@srcInstance - source instance to copy values from"""
		super( CompositeColumn, self )._f_metaMember_initInstance( instance, srcInstance )
		# Set the value attribute of the instance
		if srcInstance is None:
			field = CompositeField( self._targetRowClass )
		else:
			field = CompositeField( self._targetRowClass, copy.deepcopy( self._f_getValueFromInstance( srcInstance ) ) )
		setattr( instance, self._fieldAttrName, field )
		field._f_register( instance, self._name )



	def __get__(self, obj, objtype = None):
		"""Descriptor - get"""
		if obj is not None:
			# Return the field value
			return self._f_getValueFromInstance( obj )
		else:
			# Return the proxy declaration
			return self

	def __set__(self, obj, value):
		"""Descriptor - set"""
		self._f_setValueInInstance( obj, value )

	def __delete__(self, obj):
		"""Descriptor - delete"""
		raise TypeError, 'cannot delete field value'



	def _f_getVarFromInstance(self, instance):
		"""Get the variable interface corresponding to @self from @instance
		@instance - an instance of @Row"""
		return getattr( instance, self._fieldAttrName )


	def _f_getValueFromInstance(self, instance):
		"""Get the value of the field managed by @self in @instance"""
		return getattr( instance, self._fieldAttrName ).getValue()


	def _f_setValueInInstance(self, instance, value):
		"""Set the value of the field managed by @self in @instance"""
		getattr( instance, self._fieldAttrName ).setValue( value )


	def _f_readInstanceFieldXml(self, instance, instanceXmlNode):
		"""Read the field managed by @self in @instance from an XML node"""
		super( CompositeColumn, self )._f_readInstanceFieldXml( instance, instanceXmlNode )
		instanceXmlNode.getChild( self._xmlName )  >>  self._f_getVarFromInstance( instance )

	def _f_writeInstanceFieldXml(self, instance, instanceXmlNode):
		"""Write the field managed by @self in @instance to an XML node"""
		super( CompositeColumn, self )._f_writeInstanceFieldXml( instance, instanceXmlNode )
		instanceXmlNode.addChild( self._xmlName )  <<  self._f_getVarFromInstance( instance )








#//////////////////////////////////////////////////////////////////////////////
#			RowRefColumn
#//////////////////////////////////////////////////////////////////////////////

class RowRefColumn (FieldColumnBaseWithXml):
	"""Row-reference column"""
	class _ObjectSourceMemberRef (ObjectSourceColumnInterface):
		"""Row-reference column member"""
		def __init__(self, ownerColumn, targetColumn):
			super( RowRefColumn._ObjectSourceMemberRef, self ).__init__()

			if not isinstance( targetColumn, ObjectSourceColumnInterface ):
				raise TypeError, 'targetColumn must reference a ObjectSourceColumnInterface'

			self._ownerColumn = ownerColumn
			self._targetColumn = targetColumn
			self._fieldAttrName = None


		def _f_initMetaMember(self):
			"""Initialise the meta-member"""
			self._fieldAttrName = intern( '_proxy_' + self._ownerColumn._name + '.' + self._targetColumn._name )


		def _p_initInstance(self, instance):
			"""Initialise instance
			@instance - instance to initialise"""
			rowRefVar = self._ownerColumn._f_getVarFromInstance( instance )

			targetInstance = self._ownerColumn._f_getValueFromInstance( instance )

			if targetInstance is not None:
				targetVar = self._targetColumn._f_getVarFromInstance( targetInstance )
			else:
				targetVar = None

			proxy = ObjectSourceProxyField( targetVar )
			proxy._f_register( instance, self._fieldAttrName )

			def onRefValue(targetInstance):
				if targetInstance is not None:
					targetVar = self._targetColumn._f_getVarFromInstance( targetInstance )
				else:
					targetVar = None
				proxy.target = targetVar

			rowRefVar.valueSignal.connect( onRefValue )

			setattr( instance, self._fieldAttrName, proxy )



		def _f_getObjectSourceFromInstance(self, instance):
			"""Get the object source corresponding to @self from @instance
			@instance - an instance of @Row"""
			if not hasattr( instance, self._fieldAttrName ):
				self._p_initInstance( instance )
			return getattr( instance, self._fieldAttrName )


		def _f_getValueFromInstance(self, instance):
			"""Get the value of the field managed by @self in @instance"""
			if not hasattr( instance, self._fieldAttrName ):
				self._p_initInstance( instance )
			return getattr( instance, self._fieldAttrName ).getValue()





	class _FieldMemberRef (FieldColumnInterface):
		"""Row-reference column member"""
		def __init__(self, ownerColumn, targetColumn):
			super( RowRefColumn._FieldMemberRef, self ).__init__()

			if not isinstance( targetColumn, FieldColumnInterface ):
				raise TypeError, 'targetColumn must reference a FieldColumnInterface'

			self._ownerColumn = ownerColumn
			self._targetColumn = targetColumn
			self._fieldAttrName = None


		def _f_initMetaMember(self):
			"""Initialise the meta-member"""
			self._fieldAttrName = intern( '_proxy_' + self._ownerColumn._name + '.' + self._targetColumn._name )


		def _p_initInstance(self, instance):
			"""Initialise instance
			@instance - instance to initialise"""
			rowRefVar = self._ownerColumn._f_getVarFromInstance( instance )

			targetInstance = self._ownerColumn._f_getValueFromInstance( instance )

			if targetInstance is not None:
				targetVar = self._targetColumn._f_getVarFromInstance( targetInstance )
			else:
				targetVar = None

			proxy = VarProxyField( targetVar )
			proxy._f_register( instance, self._fieldAttrName )

			def onRefValue(targetInstance):
				if targetInstance is not None:
					targetVar = self._targetColumn._f_getVarFromInstance( targetInstance )
				else:
					targetVar = None
				proxy.target = targetVar

			rowRefVar.valueSignal.connect( onRefValue )

			setattr( instance, self._fieldAttrName, proxy )



		def _f_getVarFromInstance(self, instance):
			"""Get the variable interface corresponding to @self from @instance
			@instance - an instance of @Row"""
			if not hasattr( instance, self._fieldAttrName ):
				self._p_initInstance( instance )
			return getattr( instance, self._fieldAttrName )


		def _f_getValueFromInstance(self, instance):
			"""Get the value of the field managed by @self in @instance"""
			if not hasattr( instance, self._fieldAttrName ):
				self._p_initInstance( instance )
			return getattr( instance, self._fieldAttrName ).getValue()


		def _f_setValueInInstance(self, instance, value):
			"""Set the value of the field managed by @self in @instance"""
			if not hasattr( instance, self._fieldAttrName ):
				self._p_initInstance( instance )
			getattr( instance, self._fieldAttrName ).setValue( value )



	def __init__(self, targetRowClass, xmlName=None, doc=''):
		"""Constructor
		@targetRowClass - the target row class
		@xmlName - the xml node name [None]
		@doc - doc string ['']
		@traits - field traits [None]"""
		super( RowRefColumn, self ).__init__( xmlName, doc )
		if not ( isinstance( targetRowClass, RowClass )  and  issubclass( targetRowClass, Row ) ):
			raise TypeError, 'targetRowClass must be an instance of RowClass and a subclass of Row'
		self._targetRowClass = targetRowClass
		self._targetColumnToRefTable = {}



	def _f_metaMember_initMetaMember(self, cls, name):
		super( RowRefColumn, self )._f_metaMember_initMetaMember( cls, name )
		for ref in self._targetColumnToRefTable.values():
			ref._f_initMetaMember()


	def _f_metaMember_overload(self, superMetaMember, superClass):
		"""Inform a meta-member that it is overloading an existing meta-member
		@superMetaMember - the meta-member being overloaded
		@superClass - the base class"""
		if not issubclass( self._targetRowClass, superMetaMember._targetRowClass ):
			raise ValueError, 'cannot overload row-reference column %s.%s; the target row class (%s) must be a subclass of the base column target row class (%s)' % ( self._cls.__name__, self._name, self._targetRowClass.__name__, superMetaMember._targetRowClass.__name__ )


	def _f_metaMember_initInstance(self, instance, srcInstance=None):
		"""Initialise an instance of the class of which this meta-member is part
		@instance - the instance to initialise
		@srcMember - instance of this meta-member, part of another object, from which to copy data"""
		super( RowRefColumn, self )._f_metaMember_initInstance( instance, srcInstance )

		if srcInstance is None:
			field = RowRefField( self._targetRowClass )
		else:
			field = RowRefField( self._targetRowClass, self._f_getValueFromInstance( srcInstance ) )
		setattr( instance, self._fieldAttrName, field )
		field._f_register( instance, self._name )




	def __getattr__(self, name):
		"""Attribute accessor - Create a column wrapper for a specified column from the target row class"""
		try:
			targetColumn = getattr( self._targetRowClass, name )
		except AttributeError:
			raise AttributeError, 'Row class \'%s\' has no attribute named \'%s\'' % ( self._targetRowClass.__name__, name )
		else:
			try:
				memberRef = self._targetColumnToRefTable[targetColumn]
			except KeyError:
				if isinstance( targetColumn, FieldColumnInterface ):
					memberRef = RowRefColumn._FieldMemberRef( self, targetColumn )
				elif isinstance( targetColumn, ObjectSourceColumnInterface ):
					memberRef = RowRefColumn._ObjectSourceMemberRef( self, targetColumn )
				else:
					raise TypeError, 'Attempting to access member of class \'%s\'; it is not an instance of \'ObjectSourceColumnInterface\''  %  ( self._targetRowClass.__name__, )
				self._targetColumnToRefTable[targetColumn] = memberRef
			return memberRef



	def __get__(self, obj, objtype = None):
		"""Descriptor - get"""
		if obj is not None:
			# Return the field value
			return self._f_getValueFromInstance( obj )
		else:
			# Return the proxy declaration
			return self

	def __set__(self, obj, value):
		"""Descriptor - set"""
		self._f_setValueInInstance( obj, value )

	def __delete__(self, obj):
		"""Descriptor - delete"""
		raise TypeError, 'cannot delete field value'





	def _f_getVarFromInstance(self, instance):
		"""Get the variable interface corresponding to @self from @instance
		@instance - an instance of @Row"""
		return getattr( instance, self._fieldAttrName )


	def _f_getValueFromInstance(self, instance):
		"""Get the value of the field managed by @self in @instance"""
		return getattr( instance, self._fieldAttrName ).getValue()


	def _f_setValueInInstance(self, instance, value):
		"""Set the value of the field managed by @self in @instance"""
		getattr( instance, self._fieldAttrName ).setValue( value )



	def _f_readInstanceFieldXml(self, instance, instanceXmlNode):
		"""Read the field managed by @self in @instance from an XML node"""
		super( RowRefColumn, self )._f_readInstanceFieldXml( instance, instanceXmlNode )
		instanceXmlNode.getChild( self._xmlName )  >>  self._f_getVarFromInstance( instance )

	def _f_writeInstanceFieldXml(self, instance, instanceXmlNode):
		"""Write the field managed by @self in @instance to an XML node"""
		super( RowRefColumn, self )._f_writeInstanceFieldXml( instance, instanceXmlNode )
		instanceXmlNode.addChild( self._xmlName )  <<  self._f_getVarFromInstance( instance )








#//////////////////////////////////////////////////////////////////////////////
#   			ObjectSourceProxyColumn
#//////////////////////////////////////////////////////////////////////////////

class ObjectSourceProxyColumn (ObjectSourceColumnBase):
	"""Field proxy declaration"""
	def __init__(self, memberRef):
		"""Constructor
		@targetColumn - the target column"""
		super( ObjectSourceProxyColumn, self ).__init__( '[PROXY] ' + memberRef._targetColumn.__doc__ )

		if not isinstance( memberRef, RowRefColumn._ObjectSourceMemberRef )  and  not isinstance( memberRef, RowRefColumn._FieldMemberRef ):
			raise TypeError, '@columnRef must be a RowRefColumn member reference (a RowRefColumn._ObjectSourceMemberRef or a RowRefColumn._FieldMemberRef)'

		self._o_addDependency( memberRef._ownerColumn )

		self._memberRef = memberRef


	def _f_metaMember_initClass(self):
		super( ObjectSourceProxyColumn, self )._f_metaMember_initClass()

		if not issubclass( self._cls, self._memberRef._ownerColumn._cls ):
			raise ValueError, 'field proxy (%s.%s) cannot target a column from a class that is not a superclass of %s, it tried to target a column from %s' %	\
							( cls.__name__, name, cls.__name__, self._memberRef._ownerColumn._cls.__name__ )


	def _f_metaMember_initInstance(self, instance, srcInstance=None):
		"""Initialise instance
		@instance - instance to initialise"""
		proxy = self._memberRef._f_getObjectSourceFromInstance( instance )
		proxy._f_register( instance, self._name )

		setattr( instance, self._fieldAttrName, proxy )

		def onValueChanged():
			instance._f_onFieldModify( self._name )

		proxy.changedSignal.connect( onValueChanged )





	def __get__(self, obj, objtype = None):
		"""Descriptor - get"""
		if obj is not None:
			# Return the field value
			return self._f_getValueFromInstance( obj )
		else:
			# Return the proxy declaration
			return self

	def __set__(self, obj, value):
		"""Descriptor - set"""
		raise TypeError, 'cannot set value of object source proxy'

	def __delete__(self, obj):
		"""Descriptor - delete"""
		raise TypeError, 'cannot delete field value'



	def _f_getObjectSourceFromInstance(self, instance):
		"""Get the field proxy from a RowProxy instance that corresponds to @self
		@instance - the instance from which to retrieve the field proxy"""
		return getattr( instance, self._fieldAttrName )


	def _f_getValueFromInstance(self, instance):
		"""Get the value of the field proxy managed by @self in @instance"""
		return getattr( instance, self._fieldAttrName ).getValue()






#//////////////////////////////////////////////////////////////////////////////
#   			FieldProxyColumn
#//////////////////////////////////////////////////////////////////////////////

class FieldProxyColumn (FieldColumnBase):
	"""Field proxy declaration"""
	def __init__(self, memberRef):
		"""Constructor
		@targetColumn - the target column"""
		super( FieldProxyColumn, self ).__init__( '[PROXY] ' + memberRef._targetColumn.__doc__ )

		if not isinstance( memberRef, RowRefColumn._FieldMemberRef ):
			raise TypeError, '@columnRef must be a RowRefColumn member reference (a RowRefColumn._FieldMemberRef)'

		self._o_addDependency( memberRef._ownerColumn )

		self._memberRef = memberRef


	def _f_metaMember_initClass(self):
		super( FieldProxyColumn, self )._f_metaMember_initClass()

		if not issubclass( self._cls, self._memberRef._ownerColumn._cls ):
			raise ValueError, 'field proxy (%s.%s) cannot target a column from a class that is not a superclass of %s, it tried to target a column from %s' %	\
							( cls.__name__, name, cls.__name__, self._memberRef._ownerColumn._cls.__name__ )


	def _f_metaMember_initInstance(self, instance, srcInstance=None):
		"""Initialise instance
		@instance - instance to initialise"""
		proxy = self._memberRef._f_getObjectSourceFromInstance( instance )
		proxy._f_register( instance, self._name )

		setattr( instance, self._fieldAttrName, proxy )

		def onValueChanged():
			instance._f_onFieldModify( self._name )

		proxy.changedSignal.connect( onValueChanged )





	def __get__(self, obj, objtype = None):
		"""Descriptor - get"""
		if obj is not None:
			# Return the field value
			return self._f_getValueFromInstance( obj )
		else:
			# Return the proxy declaration
			return self

	def __set__(self, obj, value):
		"""Descriptor - set"""
		self._f_setValueInInstance( obj, value )

	def __delete__(self, obj):
		"""Descriptor - delete"""
		raise TypeError, 'cannot delete field value'



	def _f_getVarFromInstance(self, instance):
		"""Get the field proxy from a RowProxy instance that corresponds to @self
		@instance - the instance from which to retrieve the field proxy"""
		return getattr( instance, self._fieldAttrName )


	def _f_getValueFromInstance(self, instance):
		"""Get the value of the field proxy managed by @self in @instance"""
		return getattr( instance, self._fieldAttrName ).getValue()


	def _f_setValueInInstance(self, instance, value):
		"""Set the value of the field proxy managed by @self in @instance"""
		getattr( instance, self._fieldAttrName ).setValue( value )







#//////////////////////////////////////////////////////////////////////////////
#   			FunctionColumn
#//////////////////////////////////////////////////////////////////////////////

class FunctionColumn (FieldColumnBase):
	"""Function column"""
	def __init__(self, function, doc=''):
		"""Constructor
		@function - the function used to compute the value
		@doc - doc string ['']"""
		super( FunctionColumn, self ).__init__( doc )
		self._function = function



	def _f_metaMember_initInstance(self, instance, srcInstance=None):
		"""Initialise instance
		@instance - instance to initialise
		@srcInstance - source instance to copy values from"""
		super( FunctionColumn, self )._f_metaMember_initInstance( instance, srcInstance )
		field = FunctionField( self._function )
		setattr( instance, self._fieldAttrName, field )
		field._f_register( instance, self._name )


	def __get__(self, obj, objtype = None):
		"""Descriptor - get"""
		if obj is not None:
			# Return the field value
			return self._f_getValueFromInstance( obj )
		else:
			# Return the field declaration
			return self

	def __set__(self, obj, value):
		"""Descriptor - set"""
		raise TypeError, 'cannot set function field value'

	def __delete__(self, obj):
		"""Descriptor - delete"""
		raise TypeError, 'cannot delete field value'


	def _f_getVarFromInstance(self, instance):
		"""Get the variable interface corresponding to @self from @instance
		@instance - an instance of @Row"""
		return getattr( instance, self._fieldAttrName )



	def _f_getValueFromInstance(self, instance):
		"""Get the value of the field managed by @self in @instance"""
		return getattr( instance, self._fieldAttrName ).getValue()






#//////////////////////////////////////////////////////////////////////////////
#			_FieldAccessor
#//////////////////////////////////////////////////////////////////////////////

class _FieldAccessor (object):
	__slots__ = [ '_rowClass', '_rowInstance' ]

	"""_FieldAccessor - helper class"""
	def __init__(self, rowClass, rowInstance):
		"""Constructor
		@fieldColumn - field column"""
		super( _FieldAccessor, self ).__init__()
		self._rowClass = rowClass
		self._rowInstance = rowInstance


	def __getattr__(self, attrName):
		try:
			column = self._rowClass._Row_fieldColumns[attrName]
		except KeyError:
			raise AttributeError, 'no field named %s in class %d'  %  ( attrName, self._rowClass )

		if self._rowInstance is None:
			return column
		else:
			return column._f_getVarFromInstance( self._rowInstance )


	def __setattr__(self, attrName, value):
		if attrName == '_rowClass'  or  attrName == '_rowInstance':
			object.__setattr__( self, attrName, value )
		else:
			raise AttributeError, 'fields cannot be set'


	def __delattr__(self, attrName):
		raise AttributeError, 'fields cannot be deleted'





#//////////////////////////////////////////////////////////////////////////////
#			_FieldAccessorDescriptor
#//////////////////////////////////////////////////////////////////////////////

class _FieldAccessorDescriptor (object):
	"""Fields can be accessed from here"""
	def __get__(self, obj, objtype = None):
		"""Descriptor - get"""
		# Return a _FieldAccessor
		return _FieldAccessor( objtype, obj )


	def __set__(self, obj, value):
		"""Descriptor - set"""
		raise AttributeError, 'cannot set fields'


	def __delete__(self, obj):
		"""Descriptor - delete"""
		raise AttributeError, 'cannot delete fields'






#//////////////////////////////////////////////////////////////////////////////
#			RowClass
#//////////////////////////////////////////////////////////////////////////////

class RowClass (KMeta.KClass):
	"""Field container metaclass"""
	def __init__(cls, clsName, clsBases, clsDict):
		"""Constructor
		@clsName - class name
		@bases - base classes
		@clsDict - class dict"""
		super( RowClass, cls ).__init__( clsName, clsBases, clsDict )

		cls._Row_fieldColumns = cls._o_gatherDictFromBases( clsBases, '_Row_fieldColumns' )


		# Process field columns
		# Gather a list of field columns, and a list of composite columns
		for name, value in clsDict.items():
			if isinstance( value, FieldColumnBase ):
				# Register the column
				cls._Row_fieldColumns[name] = value



		# Register the class with the XML IO system
		xmlName = clsDict.get( '__iotypename__', clsName )
		IOXml.ioObjectFactoryRegister( xmlName, cls )



	def getFieldColumnByName(cls, name):
		"""Get a column by name
		@name - the name of the column to retrieve"""
		return cls._Row_fieldColumns[name]








#//////////////////////////////////////////////////////////////////////////////
#   			Row
#//////////////////////////////////////////////////////////////////////////////

class Row (KMeta.KObject):
	__metaclass__ = RowClass
	trackerClass = DBCommandTracker.RowCommandTracker

	fieldModifiedSignal = SignalSlot.ClassSignal()

	"""Field container base class"""
	def __init__(self, src = None):
		"""Constructor"""
		self._getListeners = set()
		self._getExternalListeners = set()
		self._commandTracker_ = None
		super( Row, self ).__init__( src )



	def addGetListener(self, listener):
		self._getListeners.add( listener )

	def removeGetListener(self, listener):
		self._getListeners.remove( listener )


	def addGetExternalListener(self, listener):
		self._getExternalListeners.add( listener )

	def removeGetExternalListener(self, listener):
		self._getExternalListeners.remove( listener )



	def _f_onChildFieldPreModify(self, fieldName):
		"""Notify of field imminent value change
		@fieldName - the name of the field that is to be modified"""
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onRowChildFieldPreModify( self, fieldName )

	def _f_onChildFieldModify(self, fieldName):
		"""Notify of field value change
		@fieldName - the name of the field that has been modified"""
		self._f_onFieldModify( fieldName )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onRowChildFieldModified( self, fieldName )


	def _f_onFieldModify(self, fieldName):
		"""Notify of field value change
		@fieldName - the name of the field that has been modified"""
		self.fieldModifiedSignal.emit( self, fieldName )


	def _f_onGet(self, fieldName):
		for listener in self._getListeners:
			listener( self, fieldName )

	def _f_onGetExternal(self, row):
		for listener in self._getExternalListeners:
			listener( row )


	def getFieldByName(self, name):
		"""Get a field by name
		@fieldName - the name of the field to get"""
		fieldColumn = getattr( type( self ), name )
		assert isinstance( fieldColumn, FieldColumnInterface )
		return fieldColumn._f_getVarFromInstance( self )


	def _o_getVersion(self):
		try:
			currentVersion = self.__version__
		except AttributeError:
			return None
		else:
			return currentVersion




	def __readxml__(self, xmlNode):
		"""XML IO read helper"""
		for fieldColumn in self._Row_fieldColumns.values():
			if isinstance( fieldColumn, FieldColumnBaseWithXml ):
				fieldColumn._f_readInstanceFieldXml( self, xmlNode )
		# Set the version if the version property exists, OR the object already has a version
		version = xmlNode.getVersionProperty()
		self.__version__ = version


	def __writexml__(self, xmlNode):
		"""XML IO write helper"""
		for fieldColumn in self._Row_fieldColumns.values():
			if isinstance( fieldColumn, FieldColumnBaseWithXml ):
				fieldColumn._f_writeInstanceFieldXml( self, xmlNode )
		try:
			currentVersion = self.__version__
		except AttributeError:
			pass
		else:
			if currentVersion is not None:
				xmlNode.setVersionProperty( currentVersion )



	def __copy__(self):
		"""Copy helper"""
		return self.__class__( src=self )



	def _f_getCompositeFields(self):
		"""Get a list of the composite fields in this row (used by the command tracker)"""
		return [ column._f_getVarFromInstance( self )   for column in self._Row_fieldColumns.values()   if isinstance( column, CompositeColumn ) ]


	fields = _FieldAccessorDescriptor()






#//////////////////////////////////////////////////////////////////////////////
#				CompositeField
#//////////////////////////////////////////////////////////////////////////////

class CompositeField (GenericReadOnlyVarInterface):
	__metaclass__ = FieldBaseClass
	valueClass = Row

	"""Composite Field class"""
	def __init__(self, targetRowClass=None, initialValue=None):
		"""Constructor:
		@targetRowClass - the class of the target row
		@initialValue - the initial row reference"""
		super( CompositeField, self ).__init__()

		assert initialValue is None  or  isinstance( initialValue, rowClass ), 'initialValue must be an instance of rowClass'

		self._ownerRow = None
		self._name = None

		if initialValue is None:
			self._value = targetRowClass()
		else:
			self._value = initialValue

		self._value.fieldModifiedSignal.connect( self._p_onRowValueModify )

		if targetRowClass is None:
			targetRowClass = Row
		self._targetRowClass = targetRowClass



	def _p_isRegistered(self):
		"""Private - Determine if the field is registered"""
		return self._ownerRow is not None


	def isValid(self):
		"""Determine if the variable is valid"""
		return True


	def getValue(self):
		"""Get the field value"""
		self._p_assertRegistered()
		self._ownerRow._f_onGet( self._name )
		self._ownerRow._f_onGetExternal( self._value )
		return self._value


	def __readxml__(self, xmlNode):
		"""Read value from an XML node
		@xmlNode - the XML node"""
		if xmlNode.isValid():
			xmlNode  >>  self._value

	def __writexml__(self, xmlNode):
		"""Write value to an XML node
		@xmlNode - the XML node"""
		if xmlNode.isValid():
			xmlNode  <<  self._value



	def _p_onRowValueModify(self, row, fieldName):
		"""Private - notify of value change"""
		self._p_assertRegistered()
		self._ownerRow._f_onFieldModify( self._name )
		super( CompositeField, self )._p_valueChanged( self._value )

	def _f_register(self, row, name):
		"""Private - register field
		@row - the row that contains @self
		@name - the field name"""
		self._ownerRow = row
		self._name = name

	def _p_assertRegistered(self):
		"""Private - Assert that the field is registered"""
		assert self._p_isRegistered(), 'field not registered'




	def __copy__(self):
		"""Copy helper"""
		c = self.__class__()
		c._value = self._value
		return c

	def __deepcopy__(self, memo):
		"""Deep copy helper"""
		c = self.__class__()
		c._vaue = copy.deepcopy( self._value, memo )
		return c


	value = property( getValue, None )







#//////////////////////////////////////////////////////////////////////////////
#				RowRefField
#//////////////////////////////////////////////////////////////////////////////

class RowRefField (GenericVarInterface):
	__metaclass__ = FieldBaseClass
	valueClass = Row

	"""Row Reference class"""
	def __init__(self, targetRowClass=None, initialValue=None):
		"""Constructor:
		@targetRowClass - the class of the target row
		@initialValue - the initial row reference"""
		super( RowRefField, self ).__init__()

		assert initialValue is None  or  isinstance( initialValue, rowClass ), 'initialValue must be an instance of rowClass'

		self._ownerRow = None
		self._name = None

		self._value = initialValue

		if targetRowClass is None:
			targetRowClass = Row
		self._targetRowClass = targetRowClass



	def _p_isRegistered(self):
		"""Private - Determine if the field is registered"""
		return self._ownerRow is not None


	def isValid(self):
		"""Determine if the variable is valid"""
		return True


	def getValue(self):
		"""Get the field value"""
		self._p_assertRegistered()
		self._ownerRow._f_onGet( self._name )
		if self._value is not None:
			self._ownerRow._f_onGetExternal( self._value )
		return self._value

	def setValue(self, value):
		"""Set the field value
		@value - the new value"""
		self._p_assertRegistered()
		assert value is None  or  isinstance( value, self._targetRowClass )
		self._p_preValueChange( self._value )
		self._o_unrefValue( self._value )
		self._value = value
		self._o_refValue( value )
		self._p_valueChanged( value )


	def __readxml__(self, xmlNode):
		"""Read value from an XML node
		@xmlNode - the XML node"""
		if xmlNode.isValid():
			self.setValue( xmlNode.readObject() )

	def __writexml__(self, xmlNode):
		"""Write value to an XML node
		@xmlNode - the XML node"""
		if xmlNode.isValid():
			xmlNode.writeObject( self.getValue() )



	def _p_preValueChange(self, oldValue):
		"""Private - notify of imminent value change"""
		self._p_assertRegistered()
		self._ownerRow._f_onChildFieldPreModify( self._name )
		super( RowRefField, self )._p_preValueChange( oldValue )

	def _p_valueChanged(self, value):
		"""Private - notify of value change"""
		self._p_assertRegistered()
		self._ownerRow._f_onChildFieldModify( self._name )
		super( RowRefField, self )._p_valueChanged( value )

	def _f_register(self, row, name):
		"""Private - register field
		@row - the row that contains @self
		@name - the field name"""
		self._ownerRow = row
		self._name = name

	def _p_assertRegistered(self):
		"""Private - Assert that the field is registered"""
		assert self._p_isRegistered(), 'field not registered'




	def __copy__(self):
		"""Copy helper"""
		c = self.__class__()
		c._value = self._value
		return c

	def __deepcopy__(self, memo):
		"""Deep copy helper"""
		c = self.__class__()
		c._vaue = copy.deepcopy( self._value, memo )
		return c


	value = property( getValue, setValue )







if __name__ == '__main__':
	import traceback

	class RowA (Row):
		a = FieldColumn( int, 0 )
		b = FieldColumn( float, 0.0 )

	class RowB (Row):
		c = CompositeColumn( RowA )
		d = FieldColumn( int, 1 )

	class RowC (Row):
		x = RowRefColumn( RowA )
		xa = FieldProxyColumn( x.a )

	class RowD (Row):
		x = RowRefColumn( RowA )
		xa = FieldProxyColumn( x.a )
		y = RowRefColumn( RowB )
		e = FieldColumn( int, 2 )
		q = CompositeColumn( RowA )

		def _f(self):
			if self.xa is None:
				return 0
			else:
				return self.xa + self.e + self.q.a

		def _g(self):
			return self.f * self.f

		def _h(self):
			if self.x is not None  and  self.y is not None:
				return self.x.b + self.y.c.b
			else:
				return 0

		f = FunctionColumn( _f )
		g = FunctionColumn( _g )
		h = FunctionColumn( _h )


	def connectPrintValue(field, greeting):
		def makePrinter(greeting, field):
			def printValue():
				value = field.getValue()
				print greeting, ':', value
			return printValue
		field.changedSignal.connect( makePrinter( greeting, field ) )

	def connectTracebackPrintValue(field, greeting):
		def makePrinter(greeting, field):
			def printValue():
				value = field.getValue()
				print greeting, ':', value
				#traceback.print_stack()
			return printValue
		field.changedSignal.connect( makePrinter( greeting, field ) )


	a = RowA()
	connectPrintValue( a.fields.a, 'a.a' )
	connectPrintValue( a.fields.b, 'a.b' )

	b = RowB()
	connectPrintValue( b.c.fields.a, 'b.c.a' )
	connectPrintValue( b.c.fields.b, 'b.c.b' )
	connectPrintValue( b.fields.d, 'b.d' )

	c = RowC()
	connectPrintValue( c.fields.x, 'c.x' )
	connectPrintValue( c.fields.xa, 'c.xa' )

	d = RowD()
	connectTracebackPrintValue( d.fields.f, 'd.f' )
	connectTracebackPrintValue( d.fields.g, 'd.g' )
	connectTracebackPrintValue( d.fields.h, 'd.h' )
	print d.f
	print d.g
	print d.h


	commands = [	\
		'a.a = 1',
		'a.b = 2.0',
		'b.c.a = 3',
		'b.c.b = 4.0',
		'b.d = 5',
		'c.x = a',
		'c.xa = 6',
		'a.a = 7',
		'c.x = None',
		'd.x = a',
		'd.e = 20',
		'a.a = 5',
		'd.q.a = 50',
		'd.y = b',
		'a.b = 5.3',
		'b.c.b = 2.7',
	]

	for cmd in commands:
		print '>>>', cmd
		exec( cmd )
		print '\n'




