##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
import pygtk
pygtk.require( '2.0' )
import gtk

import Variable
from gSculpt.VarEdit import VarEdit

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.Kernel.KMeta import KMetaMember, KClass, KObject

from gSculpt.Kernel.DB import Row, RowClass, RowRefColumn





class RowEditContext (object):
	def __init__(self, row, rowEditor, view, editorSettings, parentWindow, commandHistory):
		self.row = row
		self.rowEditor = rowEditor
		self.view = view
		self.editorSettings = editorSettings
		self.parentWindow = parentWindow
		self.commandHistory = commandHistory



class ColumnEditor (KMetaMember):
	"Column editor interface"
	def _f_getVarEdit(self, rowEditor):
		return getattr( rowEditor, self._name )


	def createElement(self, rowEditContext):
		"""Create an element
		@rowEditContext - the context in which this element is being used"""
		assert False, 'abstract'

	def destroyElement(self, element, rowEditContext):
		"""Destroy a previously created element
		@element; the element to destroy
		@rowEditContext - the context in which this element is being used"""
		pass

	def initialiseElement(self, element, rowEditContext):
		"""Initialise a previously created element
		@element; the element to initialise
		@rowEditContext - the context in which this element is being used"""
		pass

	def shutdownElement(self, element, rowEditContext):
		"""Shutdown a previously created element
		@element; the element to shutdown
		@rowEditContext - the context in which this element is being used"""
		pass


	def installElementEventHandlers(self, element, rowEditContext):
		"""Install the event handlers of a previously created element
		@element; the element in question
		@rowEditContext - the context in which this element is being used"""
		pass

	def uninstallElementEventHandlers(self, element, rowEditContext):
		"""Uninstall the event handlers of a previously created element
		@element; the element in question
		@rowEditContext - the context in which this element is being used"""
		pass


	def installElementAccelerators(self, element, accels, rowEditContext):
		"""Install the accelerators of a previously created element
		@element; the element in question
		@accels - accelerator list
		@rowEditContext - the context in which this element is being used"""
		pass

	def uninstallElementAccelerators(self, element, accels, rowEditContext):
		"""Uninstall the accelerators of a previously created element
		@element; the element in question
		@accels - accelerator list
		@rowEditContext - the context in which this element is being used"""
		pass


	@staticmethod
	def _p_getVar(column, rowEditContext):
		"""Private - get a variable from a function context given the column
		@column - the column
		@rowEditContext - the context in which this element is being used"""
		return column._f_getVarFromInstance( rowEditContext.rowEditor )


	@staticmethod
	def _p_getObjectSource(column, rowEditContext):
		"""Private - get an object source from a function context given the column
		@column - the column
		@rowEditContext - the context in which this element is being used"""
		return column._f_getObjectSourceFromInstance( rowEditContext.rowEditor )




class ColumnWidgetEditor (ColumnEditor):
	"Widget based column editor interface"
	def __init__(self):
		super( ColumnWidgetEditor, self ).__init__()
		self._sensitiveVarColumn = None
		self._bSensitivityInverted = False


	def initialiseElement(self, element, rowEditContext):
		super( ColumnWidgetEditor, self ).initialiseElement( element, rowEditContext )
		# Handle sensitivity
		if self._sensitiveVarColumn is not None:
			# Get the field
			sensitiveVar = self._p_getVar( self._sensitiveVarColumn, rowEditContext )
			# Attach as sensitivity source
			element.attachSensitivityVariable( sensitiveVar, self._bSensitivityInverted )

	def shutdownElement(self, element, rowEditContext):
		super( ColumnWidgetEditor, self ).shutdownElement( element, rowEditContext )
		# Handle sensitivity
		if self._sensitiveVarColumn is not None:
			# Get the field
			sensitiveVar = self._p_getVar( self._sensitiveVarColumn, rowEditContext )
			# Detach as sensitivity source
			element.detachSensitivityVariable( sensitiveVar )


	def showElement(self, element):
		"""Show a previously created element
		@element; the element to destroy"""
		element.show()

	def hideElement(self, element):
		"""Hide a previously created element
		@element; the element to destroy"""
		element.hide()



	def _p_setSensitivityFieldColumn(self, column):
		"""Set a field column as a source of a variable that determines widget sensitivity"""
		self._sensitiveVarColumn = column
		self._bSensitivityInverted = False

	def _p_setInverseSensitivityFieldColumn(self, column):
		"""Set a field column as a source of a variable that determines the inverse of the widget sensitivity"""
		self._sensitiveVarColumn = column
		self._bSensitivityInverted = True


	sensitivity = property( None, _p_setSensitivityFieldColumn )
	inverseSensitivity = property( None, _p_setInverseSensitivityFieldColumn )






class ColumnEditorComponent (KMetaMember):
	"""Column editor component
	A UI element can be composed of multiple components. A 3D vector editing component would be composed of components to edit the X, Y, and Z axes.
	A factory component is used to retrieve the component of the UI element created by an element factory."""
	def getFactory(self):
		"Get the factory that owns this factory component"
		assert False, 'abstract'

	def getElementWidget(self, element):
		"""Get the widget of @element that is represented by this factory component
		@element - the UI element"""
		assert False, 'abstract'



class ColumnEditorAttributeComponent (ColumnEditorComponent):
	"""Column editor attribute component
	Keeps a reference to the factory, for @getFactory() to return, and a widget attribute name; the name of the attribute in the variable editor that stores the widget represented by
	this component factory, when given a UI element created by the element factory"""
	def __init__(self, factory, attributeName):
		"""Constructor
		@factory - the UI element factory
		@widgetFunctor - a function which returns the widget that is represented by this component factory, when given a UI element"""
		self._factory = factory
		self._attributeName = attributeName


	def getFactory(self):
		"Get the factory that owns this factory component"
		return self._factory

	def getElementWidget(self, element):
		"""Get the widget of @element that is represented by this factory component
		@element - the UI element"""
		return getattr( element, self._attributeName )




class BasicColumnWidgetEditor (ColumnWidgetEditor):
	def __init__(self, componentAttributeNames):
		super( BasicColumnWidgetEditor, self ).__init__()

		for attrName in componentAttributeNames:
			component = ColumnEditorAttributeComponent( self, attrName )
			setattr( self, attrName, component )






class CompositeColumnEditorClass (type):
	def __init__(cls, clsName, bases, clsDict):
		super( CompositeColumnEditorClass, cls ).__init__( clsName, bases, clsDict )
		if not clsDict.has_key( 'compositeRowEditorClass' ):
			raise AttributeError, '\'compositeRowEditorClass\' not specified'



class CompositeColumnEditor (ColumnEditor):
	__metaclass__ = CompositeColumnEditorClass

	compositeRowEditorClass = None

	class CompositeComponent (ColumnEditorComponent):
		def __init__(self, factory):
			"""Constructor
			@factory - the UI element factory"""
			self._factory = factory


		def getFactory(self):
			"Get the factory that owns this factory component"
			return self._factory

		def getElementWidget(self, element):
			"""Get the widget of @element that is represented by this factory component
			@element - the UI element"""
			return element.getRowEditorWidget()



	def __init__(self, fieldColumn):
		super( CompositeColumnEditor, self ).__init__()
		self._fieldColumn = fieldColumn
		self.rowEditor = CompositeColumnEditor.CompositeComponent( self )


	def createElement(self, rowEditContext):
		element = self.compositeRowEditorClass()
		var = self._p_getVar( self._fieldColumn, rowEditContext )
		element.attachRow( var.getValue(), rowEditContext.view, rowEditContext.editorSettings, rowEditContext.parentWindow, rowEditContext.commandHistory )
		element.createRowEditorUI()
		element.layoutRowEditorUI()
		return element

	def destroyElement(self, element, rowEditContext):
		super( CompositeColumnEditor, self ).initialiseElement( element, rowEditContext )
		element.destroyRowEditorUI()
		element.detachRow()

	def initialiseElement(self, element, rowEditContext):
		super( CompositeColumnEditor, self ).initialiseElement( element, rowEditContext )
		element.initialiseRowEditorUI()

	def shutdownElement(self, element, rowEditContext):
		super( CompositeColumnEditor, self ).shutdownElement( element, rowEditContext )
		element.shutdownRowEditorUI()

	def installElementEventHandlers(self, element, rowEditContext):
		element.rowEditorInstallEventHandlers()

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.rowEditorUninstallEventHandlers()


	def installElementAccelerators(self, element, accels, rowEditContext):
		element.rowEditorInstallAccelerators( accels )

	def uninstallElementAccelerators(self, element, accels, rowEditContext):
		element.rowEditorUninstallAccelerators( accels )








g_functionToolClassTable = {}




class RowEditorClass (RowClass):
	def __init__(cls, clsName, bases, clsDict):
		super( RowEditorClass, cls ).__init__( clsName, bases, clsDict )

		global g_functionToolClassTable
		if not clsDict.has_key( 'rowClass' ):
			raise AttributeError, 'rowClass not specified'

		if not ( isinstance( cls.rowClass, RowClass )  and  issubclass( cls.rowClass, Row ) ):
			raise TypeError, 'rowClass must be a subclass of RowClass and an instance of Row'


		# Get column editors, layout factory, and data sources from class / class dictionary
		columnEditors = RowEditorClass._o_gatherListFromBases( bases, 's_RowEdit_columnEditors' )
		columnEditorComponents = RowEditorClass._o_gatherListFromBases( bases, 's_RowEdit_columnEditorComponents' )
		layoutFactories = RowEditorClass._o_gatherListFromBases( bases, 's_RowEdit_layoutFactories' )
		cls.s_RowEdit_columnEditors = columnEditors
		cls.s_RowEdit_columnEditorComponents = columnEditorComponents
		cls.s_RowEdit_layoutFactories = layoutFactories

		# Handle dictionary contents
		for name, value in clsDict.items():
			if isinstance( value, ColumnEditor ):
				columnEditors.append( value )
			elif isinstance( value, ColumnEditorComponent ):
				columnEditorComponents.append( value )


		columnEditorsByDep = ColumnEditor._f_computeDependencySortedList( columnEditors )
		cls.s_RowEdit_columnEditorsByDep = columnEditorsByDep


		# Handle layout
		try:
			layout = clsDict['layout']
			if not isinstance( layout, FormLayoutFactory )  and  not callable( layout ):
				raise TypeError, 'layout is not an instance of FormLayoutFactory and is not callable'
			layoutFactories.append( layout )
		except KeyError:
			pass






class RowEditor (Row):
	"Function tool base class"
	__metaclass__ = RowEditorClass

	rowClass = Row


	row = RowRefColumn( Row )



	def __init__(self):
		"Constructor"
		super( RowEditor, self ).__init__()
		self._view = None
		self._editorSettings = None
		self._commandHistory = None
		self._rowEditContext = None
		self._uiElements = []
		self._columnEditorToElement = {}
		self._rowEditorWidget = None


		for columnEditor in self.s_RowEdit_columnEditorsByDep:
			setattr( self, columnEditor._name, None )

		for component in self.s_RowEdit_columnEditorComponents:
			setattr( self, component._name, None )


	def getRowEditorWidget(self):
		return self._rowEditorWidget



	def attachRow(self, row, view, editorSettings, parentWindow, commandHistory):
		"""Attach a Row object to this row editor
		@row - the row to attach
		@view - modeller view
		@parentWindow - top level window
		@commandHistory - the command history"""
		assert self.row is None, 'Row already attached'
		assert isinstance( row, self.rowClass ), '@func is not an instance of ' + self.rowClass.__name__
		self.row = row
		self._view = view
		self._editorSettings = editorSettings
		self._parentWindow = parentWindow
		self._commandHistory = commandHistory
		self._rowEditContext = self._o_createRowEditContext()

	def detachRow(self):
		"""Detach the currently attached row"""
		assert self.row is not None, 'No row attached'
		self.row = None
		self._view = None
		self._editorSettings = None
		self._parentWindow = None
		self._commandHistory = None
		self._rowEditContext = None


	def _o_createRowEditContext(self):
		return RowEditContext( self.row, self, self._view, self._editorSettings, self._parentWindow, self._commandHistory )





	def createRowEditorUI(self):
		for columnEditor in self.s_RowEdit_columnEditorsByDep:
			element = columnEditor.createElement( self._rowEditContext )
			self._uiElements.append( element )
			self._columnEditorToElement[columnEditor] = element
			if isinstance( columnEditor, ColumnWidgetEditor ):
				columnEditor.showElement( element )
			setattr( self, columnEditor._name, element )

		for component in self.s_RowEdit_columnEditorComponents:
			columnEditor = component.getFactory()
			element = columnEditor._f_getVarEdit( self )
			elementWidget = component.getElementWidget( element )
			setattr( self, component._name, elementWidget )

		self._rowEditorWidget = gtk.VBox()
		self._rowEditorWidget.show()


	def layoutRowEditorUI(self):
		for layoutFactory in self.s_RowEdit_layoutFactories:
			if isinstance( layoutFactory, FormLayoutFactory ):
				layoutWidget = layoutFactory.makeLayout( self._p_getUIElementWidget )
			elif callable( layoutFactory ):
				layoutWidget = layoutFactory( self )
				assert layoutWidget is not None, 'layout factory function must not return None'
			else:
				raise TypeError, 'layout is not an instance of FormLayoutFactory and is not callable'
			layoutWidget.show()
			self._rowEditorWidget.pack_start( layoutWidget, False, False )



	def destroyRowEditorUI(self):
		for columnEditor in reversed( self.s_RowEdit_columnEditorsByDep ):
			columnEditor.destroyElement( self._p_getUIElementForColumnEditor( columnEditor ), self._rowEditContext )
			setattr( self, columnEditor._name, None )

		for component in self.s_RowEdit_columnEditorComponents:
			setattr( self, component._name, None )

		self._rowEditorWidget = None



	def initialiseRowEditorUI(self):
		for columnEditor in self.s_RowEdit_columnEditorsByDep:
			columnEditor.initialiseElement( self._p_getUIElementForColumnEditor( columnEditor ), self._rowEditContext )


	def shutdownRowEditorUI(self):
		for columnEditor in reversed( self.s_RowEdit_columnEditorsByDep ):
			columnEditor.shutdownElement( self._p_getUIElementForColumnEditor( columnEditor ), self._rowEditContext )


	def rowEditorInstallEventHandlers(self):
		for columnEditor in self.s_RowEdit_columnEditorsByDep:
			columnEditor.installElementEventHandlers( self._p_getUIElementForColumnEditor( columnEditor ), self._rowEditContext )

	def rowEditorUninstallEventHandlers(self):
		for columnEditor in reversed( self.s_RowEdit_columnEditorsByDep ):
			columnEditor.uninstallElementEventHandlers( self._p_getUIElementForColumnEditor( columnEditor ), self._rowEditContext )


	def rowEditorInstallAccelerators(self, accels):
		assert self._rowEditContext is not None, 'no row attached'
		for columnEditor in self.s_RowEdit_columnEditorsByDep:
			columnEditor.installElementAccelerators( self._p_getUIElementForColumnEditor( columnEditor ), accels, self._rowEditContext )

	def rowEditorUninstallAccelerators(self, accels):
		assert self._rowEditContext is not None, 'no row attached'
		for columnEditor in reversed( self.s_RowEdit_columnEditorsByDep ):
			columnEditor.uninstallElementAccelerators( self._p_getUIElementForColumnEditor( columnEditor ), accels, self._rowEditContext )






	def _p_getUIElementForColumnEditor(self, factory):
		"""Private - get the UI element given the factory that manufactured it
		@factory - the factory"""
		return self._columnEditorToElement[factory]

	def _p_getUIElementWidget(self, factoryComponent):
		"""Private - get the widget given a factory component
		@factoryComponent - the factory component"""
		factory = factoryComponent.getFactory()
		element = self._p_getUIElementForColumnEditor( factory )
		return factoryComponent.getElementWidget( element )



	@classmethod
	def _o_getColumnEditors(self):
		return self.s_RowEdit_columnEditors



	def _p_getView(self):
		return self._view

	view = property( _p_getView )








if __name__ == '__main__':
	import pygtk
	pygtk.require( '2.0' )
	import gtk


	from gSculpt.Kernel.DB import FieldColumn
	from gSculpt.Kernel.ColumnEditors import IntColumnEditEntry
	from gSculpt.UI.FormLayout import FormLayoutFactory
	from gSculpt.CommandHistory.CommandHistory import CommandHistory


	class ExampleRow (Row):
		a = FieldColumn( int, 1 )

	class ExampleEditor (RowEditor):
		rowClass = ExampleRow
		aEdit = IntColumnEditEntry( ExampleRow.a, 0, 1048576, 1, 10 )

		layout = FormLayoutFactory()
		layout.row()  <<  'A:'  <<  aEdit
		layout.row()
		layout.finalise()


	def onDelete(widget, event, data):
		gtk.main_quit()


	commandHistory = CommandHistory()


	window = gtk.Window( gtk.WINDOW_TOPLEVEL )
	window.connect( 'delete-event', onDelete )
	window.show()

	row = ExampleRow()
	editor = ExampleEditor()
	editor.attachRow( row, commandHistory )
	editor.createRowEditorUI()
	editor.initialiseRowEditorUI()

	window.add( editor.getWidget() )



