##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
import copy

from gSculpt.Util.SignalSlot import ClassSignal

from gSculpt.Kernel.ObjectSource import ObjectSource
from gSculpt.Kernel.ObjectSourceProxy import ObjectSourceProxy


class VarInitialisationTypeException: pass



#//////////////////////////////////////////////////////////////////////////////
#		Generic Variable interface
#//////////////////////////////////////////////////////////////////////////////



class GenericReadOnlyVarInterface (ObjectSource):
	valueSignal = ClassSignal()
	preChangeSignal = ClassSignal()

	def _p_preValueChange(self, oldValue):
		"""Private - notify of forthcoming value change"""
		self.preChangeSignal.emit()

	def _p_valueChanged(self, value):
		"""Private - notify of value change"""
		self.valueSignal.emit( value )
		self.changedSignal.emit()



class GenericVarInterface (GenericReadOnlyVarInterface):
	def setValue(self):
		"""Set the variable value"""
		assert False, 'abstract'


	@classmethod
	def _o_refValue(cls, value):
		"""Protected - overload - called when the value is changed and @value is the new value"""
		pass

	@classmethod
	def _o_unrefValue(cls, value):
		"""Protected - overload - called when the value is changed and @value is replaced"""
		pass


	value = property( GenericReadOnlyVarInterface.getValue, setValue )




#//////////////////////////////////////////////////////////////////////////////
#			Variable interface
#//////////////////////////////////////////////////////////////////////////////



class VarInterfaceClass (type):
	def __init__(cls, clsName, bases, clsDict):
		if not clsDict.has_key( 'valueClass' ):
			raise AttributeError, 'valueClass not specified'

		super( VarInterfaceClass, cls ).__init__( clsName, bases, clsDict )



class VarInterface (GenericVarInterface):
	__metaclass__ = VarInterfaceClass
	valueClass = None

	def getValueClass(self):
		"""Get the value class"""
		return self.valueClass





#//////////////////////////////////////////////////////////////////////////////
#				Variable
#//////////////////////////////////////////////////////////////////////////////

class VariableClass (VarInterfaceClass):
	pass



class Variable (VarInterface):
	__metaclass__ = VariableClass
	valueClass = None

	def __init__(self, src=None, _p_initial=None):
		"""Constructor:
		@src - source of value, either an object of the value type, or a Variable object from which to obtain the value"""
		super( Variable, self ).__init__()

		if _p_initial is not None:
			self._value = _p_initial
		else:
			if src is None:
				self._value = self.valueClass()
			else:
				if src.__class__ == self.valueClass:
					self._value = copy.copy( src )
				elif src.__class__ == self.__class__:
					self._value = copy.copy( src._value )
				else:
					raise VarInitialisationTypeException, 'Cannot initialse variable value; source value type is "%s", value class is "%s", variable class is "%s"' % \
													( src.__class__, self.valueClass, self.__class__ )


	def getValue(self):
		"""Get the variable value"""
		return self._value

	def setValue(self, value):
		"""Set the variable value"""
		self._o_checkValueType( value )
		self._p_preValueChange( self._value )
		self._o_unrefValue( self._value )
		self._value = value
		self._o_refValue( value )
		self._p_valueChanged( value )


	def _o_checkValueType(self, value):
		assert isinstance( value, self.valueClass )


	def __copy__(self):
		"Copy helper"
		c = self.__class__()
		c._value = self._value
		return c

	def __deepcopy__(self, memo):
		"Deep copy helper"
		c = self.__class__()
		c._vaue = copy.deepcopy( self._value, memo )
		return c


	value = property( getValue, setValue )




#//////////////////////////////////////////////////////////////////////////////
#			Object Variable
#//////////////////////////////////////////////////////////////////////////////

class ObjectVar (Variable):
	valueClass = None

	def __init__(self, src=None):
		"""Constructor:
		@src - source of value, either an object of the value type, or a Variable object from which to obtain the value"""

		if src is None:
			super( ObjectVar, self ).__init__( _p_initial=self.defaultValue )
		else:
			if isinstance( src, self.valueClass ):
				super( ObjectVar, self ).__init__( _p_initial=src )
			elif isinstance( src, ObjectField ):
				super( ObjectVar, self ).__init__( _p_initial=src.getValue() )
			else:
				raise VarInitialisationTypeException, 'Cannot initialse variable value; source value type is "%s", value class is "%s", variable class is "%s"' % \
												( src.__class__, self.valueClass, self.__class__ )


	def _o_checkValueType(self, value):
		assert value is None  or  isinstance( value, self.valueClass )




#//////////////////////////////////////////////////////////////////////////////
#			Variable Proxy
#//////////////////////////////////////////////////////////////////////////////

class VarProxy (GenericVarInterface, ObjectSourceProxy):
	"""Variable proxy class; forwards signals from the target variable, and forwards calls to the target variable"""
	def setValue(self, value):
		"""Set the variable value"""
		if self._target is not None:
			return self._target.setValue( value )


	def setTarget(self, target):
		"""Set the target variable interface
		@target - the target"""
		if target is not self._target:
			# Keep the old target
			oldTarget = self._target
			# Get the old value
			oldValue = self.getValue()
			# Notify of imminent value change
			self._p_preValueChange( oldValue )

			# Record the validity
			bWasValid = self.isValid()

			# Disconnect signals
			if self._target is not None:
				self._target.validitySignal.chainDisconnect( self.validitySignal )
				self._target.valueSignal.chainDisconnect( self.valueSignal )
				self._target.changedSignal.chainDisconnect( self.changedSignal )
				self._target.preChangeSignal.chainDisconnect( self.preChangeSignal )

			self._target = target

			bIsValid = self.isValid()

			# If the variable is becoming invalid, then emit the validity signal BEFORE emitting the value signal
			if bWasValid and not bIsValid:
				self.validitySignal.emit( False )

			# Get the new value
			newValue = self.getValue()
			# Notify of value change
			self._p_valueChanged( newValue )

			# Connect signals
			if self._target is not None:
				self._target.validitySignal.chainConnect( self.validitySignal )
				self._target.valueSignal.chainConnect( self.valueSignal )
				self._target.changedSignal.chainConnect( self.changedSignal )
				self._target.preChangeSignal.chainConnect( self.preChangeSignal )

			# If the variable is becoming valid, emit the validity signal AFTER emitting the value signal
			if not bWasValid and bIsValid:
				self.validitySignal.emit( True )


	def _o_initTarget(self, target):
		"""Set the target variable interface
		@target - the target"""
		if target is not None:
			self._target = target

			# Connect signals
			self._target.validitySignal.chainConnect( self.validitySignal )
			self._target.valueSignal.chainConnect( self.valueSignal )
			self._target.changedSignal.chainConnect( self.changedSignal )
			self._target.preChangeSignal.chainConnect( self.preChangeSignal )



	value = property( ObjectSourceProxy.getValue, setValue )
	target = property( ObjectSourceProxy.getTarget, setTarget )



#//////////////////////////////////////////////////////////////////////////////
#			Variable types
#//////////////////////////////////////////////////////////////////////////////

class BoolVarInterface (VarInterface):
	valueClass = bool

class IntVarInterface (VarInterface):
	valueClass = int

class FloatVarInterface (VarInterface):
	valueClass = float

class StrVarInterface (VarInterface):
	valueClass = str


class BoolVar (Variable):
	valueClass = bool

class IntVar (Variable):
	valueClass = int

class FloatVar (Variable):
	valueClass = float

class StrVar (Variable):
	valueClass = str



