##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from gSculpt.Util.SignalSlot import ClassSignal

from gSculpt.Math.Math import Point3

from gSculpt.Kernel.ObjectSource import ObjectSource

from gSculpt.WorkArea.Viewport3d import Viewport3d

from ScreenSpaceHandle import ScreenSpace1Handle, ScreenSpace2Handle

from gSculpt.Manipulator.Manipulator import Manipulator




class RealManipulator (Manipulator):
	class Scale (object):
		def scale(self, viewport, x):
			assert False, 'abstract'


	class FixedScale (Scale):
		def __init__(self, scaleFactor = 1.0):
			self.scaleFactor = scaleFactor

		def scale(self, viewport, x):
			return x * self.scaleFactor


	class FocalDistanceScale (Scale):
		def scale(self, viewport, x):
			assert isinstance( viewport, Viewport3d ), 'FocalDistanceScale does not work unless the viewport is a Viewport3d'
			return viewport.camera.transformDistanceAcrossScreenToDistanceThruWorld( x )


	class WorldSpaceScale (Scale):
		def __init__(self, focalPoint):
			"""Constructor
			@focalPoint - the focal point; a Point3  -OR- an ObjectSource that creates a Point3 or None  -OR-  a callable object that returns a Point3"""
			self.focalPoint = focalPoint

		def scale(self, viewport, x):
			assert isinstance( viewport, Viewport3d ), 'WorldSpaceScale does not work unless the viewport is a Viewport3d'
			if self.focalPoint is not None:
				if isinstance( self.focalPoint, Point3 ):
					focalPoint = self.focalPoint
				elif isinstance( self.focalPoint, ObjectSource ):
					focalPoint = self.focalPoint.getValue()
				elif callable( self.focalPoint ):
					focalPoint = self.focalPoint()

				if focalPoint is None:
					return viewport.camera.transformDistanceAcrossScreenToDistanceThruWorld( x )
				else:
					assert isinstance( focalPoint, Point3 )

					cam = viewport.camera
					pointDist = cam.transformWorldToCameraZ( focalPoint )
					return cam.transformDistanceAcrossScreenToDistanceThruWorld( x, pointDist )
			else:
				return viewport.camera.transformDistanceAcrossScreenToDistanceThruWorld( x )







class Real1Manipulator (RealManipulator):
	startSignal = ClassSignal()
	manipulateSignal = ClassSignal()
	finishSignal = ClassSignal()


	def __init__(self, lower, upper, scale, snapping=None):
		super( Real1Manipulator, self ).__init__()
		self._view = None
		self._handle = ScreenSpace1Handle()
		self._handle.startSignal.connect( self._p_onStart )
		self._handle.dragSignal.connect( self._p_onDrag )
		self._handle.finishSignal.connect( self._p_onFinish )
		self._var = None
		self.lower = lower
		self.upper = upper
		self.scale = scale
		self._initial = 0.0
		self.snapping = snapping

		self._o_setUISensitivity( False )




	def manipAttachToView(self, view):
		assert self._view is None, 'already attached to a view'
		self._view = view
		if self._var is not None:
			self._handle.attachToView( self._view )

	def manipDetachFromView(self):
		assert self._view is not None, 'not attached to a view'
		if self._var is not None:
			self._handle.detachFromView()
		self._view = None


	def attachVariable(self, var):
		assert self._var is None,  'variable already attached'
		self._var = var

		self._o_watchVarValidity( self._var )
		self._o_refreshSensitivity()

		if self._view is not None:
			self._handle.attachToView( self._view )

	def detachVariable(self):
		assert self._var is not None,  'no variable attached'

		self._o_stopWatchingVarValidity( self._var )

		if self._view is not None:
			self._handle.detachFromView()
		self._var = None
		self._o_refreshSensitivity()



	def _o_setUISensitivity(self, bSensitive):
		self._handle.setSensitive( bSensitive )

	def _o_computeSensitivity(self):
		if self._var is not None:
			return self._var.isValid()
		else:
			return False


	def _p_onStart(self, handle, viewport):
		if self._var is not None:
			self._initial = self._var.value
			self._p_commandHistoryBreak()
			self.startSignal.emit( self, viewport )

	def _p_onDrag(self, handle, viewport, delta):
		if self._var is not None:
			value = self._initial  +  self.scale.scale( viewport, delta )
			value = min( max( value, self.lower ), self.upper );
			if self.snapping is not None:
				value = self.snapping.snap( value, viewport.pointerStatus.bControlKey, viewport.pointerStatus.bShiftKey )
			self._var.value = value
			self.manipulateSignal.emit( self, viewport )

	def _p_onFinish(self, handle, viewport):
		self._initial = None
		self._p_commandHistoryBreak()
		if self._var is not None:
			self.finishSignal.emit( self, viewport )






class Real2Manipulator (Manipulator):
	startSignal = ClassSignal()
	manipulateSignal = ClassSignal()
	finishSignal = ClassSignal()


	def __init__(self, xLower, xUpper, xScale, yLower, yUpper, yScale, xSnapping=None, ySnapping=None):
		super( Real2Manipulator, self ).__init__()
		self._view = None
		self._handle = ScreenSpace2Handle()
		self._handle.startSignal.connect( self._p_onStart )
		self._handle.dragSignal.connect( self._p_onDrag )
		self._handle.finishSignal.connect( self._p_onFinish )
		self._xVar = None
		self._yVar = None
		self.xLower = xLower
		self.xUpper = xUpper
		self.xScale = xScale
		self.yLower = yLower
		self.yUpper = yUpper
		self.yScale = yScale
		self._xInitial = 0.0
		self._yInitial = 0.0
		self.xSnapping = xSnapping
		self.ySnapping = ySnapping

		self._o_setUISensitivity( False )




	def manipAttachToView(self, view):
		assert self._view is None, 'already attached to a view'
		self._view = view
		self._handle.attachToView( self._view )

	def manipDetachFromView(self):
		assert self._view is not None, 'not attached to a view'
		self._handle.detachFromView()
		self._view = None


	def attachVariables(self, xVar, yVar):
		assert self._xVar is None  and  self._yVar is None,  'already attached to a variables'
		self._xVar = xVar
		self._yVar = yVar

		self._o_watchVarValidity( self._xVar )
		self._o_watchVarValidity( self._yVar )
		self._o_refreshSensitivity()

		if self._view is not None:
			self._handle.attachToView( self._view )

	def detachVariables(self):
		assert self._xVar is not None  or  self._yVar is not None,  'not attached to variables'

		self._o_stopWatchingVarValidity( self._xVar )
		self._o_stopWatchingVarValidity( self._yVar )

		if self._view is not None:
			self._handle.detachFromView()
		self._xVar = None
		self._yVar = None
		self._o_refreshSensitivity()



	def _o_setUISensitivity(self, bSensitive):
		self._handle.setSensitive( bSensitive )

	def _o_computeSensitivity(self):
		if self._xVar is not None  and  self._yVar is not None:
			return self._xVar.isValid()  and  self._yVar.isValid()
		else:
			return False


	def _p_onStart(self, handle, viewport):
		if self._xVar is not None:
			self._xInitial = self._xVar.value
		if self._yVar is not None:
			self._yInitial = self._yVar.value
		if self._xVar is not None  or  self._yVar is not None:
			self._p_commandHistoryBreak()
			self.startSignal.emit( self, viewport )

	def _p_onDrag(self, handle, viewport, deltaX, deltaY, constraintsX, constraintsY):
		if self._xVar is not None:
			value = self._xInitial  +  self.xScale.scale( viewport, deltaX )
			value = min( max( value, self.xLower ), self.xUpper );
			if self.xSnapping is not None  and  constraintsX != 0.0:
				value = self.xSnapping.snap( value, viewport.pointerStatus.bControlKey, viewport.pointerStatus.bShiftKey )
			self._xVar.value = value

		if self._yVar is not None:
			value = self._yInitial  +  self.yScale.scale( viewport, deltaY )
			value = min( max( value, self.yLower ), self.yUpper );
			if self.ySnapping is not None  and  constraintsY != 0.0:
				value = self.ySnapping.snap( value, viewport.pointerStatus.bControlKey, viewport.pointerStatus.bShiftKey )
			self._yVar.value = value

		if self._xVar is not None  or  self._yVar is not None:
			self.manipulateSignal.emit( self, viewport )

	def _p_onFinish(self, handle, viewport):
		self._xInitial = None
		self._yInitial = None
		if self._xVar is not None  or  self._yVar is not None:
			self._p_commandHistoryBreak()
			self.finishSignal.emit( self, viewport )


