##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from OpenGL.GL import *

from copy import copy

from gSculpt.Util.SignalSlot import ClassSignal

from gSculpt.Math.Math import BBox2

from gSculpt.Graphics.Graphics import Region3d, RectangleRegion2d, ProjectedRegion3d

from gSculpt.WorkArea.Viewport import ViewportEventHandler
from gSculpt.WorkArea.Viewport3d import Viewport3d, Viewport3dPainter

from gSculpt.View.View import PaintLayer

from gSculpt.Model import ModelDraw

from gSculpt.ProceduralTool import GSProcedureTool

from gSculpt.MarkTools.MarkRegionManipulator import MarkRegionManipulator




class RectangularMarkRegionManipulator (MarkRegionManipulator, ViewportEventHandler, Viewport3dPainter):
	startSignal = ClassSignal()
	manipulateSignal = ClassSignal()
	finishSignal = ClassSignal()


	def __init__(self):
		super( RectangularMarkRegionManipulator, self ).__init__()
		self._regionViewport = None

		self._p1 = None
		self._p2 = None
		self._bButtonDown = False


	def manipAttachToView(self, view):
		super( RectangularMarkRegionManipulator, self ).manipAttachToView( view )
		self._view.addEventHandler( self, Viewport3d )
		self._view.addPainter( self, Viewport3d )
		self._view.postRedraw()

	def manipDetachFromView(self):
		assert self._view is not None, 'not attached to a view'
		self._view.removeEventHandler( self, Viewport3d )
		self._view.removePainter( self, Viewport3d )
		self._view.postRedraw()
		super( RectangularMarkRegionManipulator, self ).manipDetachFromView()



	def evPrimaryButtonDown(self, viewport):
		super( RectangularMarkRegionManipulator, self ).evPrimaryButtonDown( viewport )

		# Emit start signal
		self.startSignal.emit( viewport )

		# Handle key status
		self._o_handleKeyStatus( viewport )

		# Initialise rectangle
		self._p1 = copy( viewport.pointerStatus.positionF )
		self._p2 = copy( self._p1 )
		self._regionViewport = viewport

		# Button pressed
		self._bButtonDown = True

		# Create region
		self._o_createNewRegion( viewport )


	def evMotion(self, viewport, point):
		super( RectangularMarkRegionManipulator, self ).evMotion( viewport, point )

		# Refresh the region
		if self._bButtonDown:
			self._p2 = copy( viewport.pointerStatus.positionF )

			if self._regionVar is not None:
				cam = viewport.camera
				region2d = RectangleRegion2d( self._p1, self._p2 )
				region3d = ProjectedRegion3d( region2d, cam.projection )
				self._regionVar.value = region3d

			self.manipulateSignal.emit( viewport )

			self._view.forcePendingRedraw()


	def evPrimaryButtonUp(self, viewport):
		super( RectangularMarkRegionManipulator, self ).evPrimaryButtonUp( viewport )

		# Button released
		self._bButtonDown = False

		# Finished
		self._p_commandHistoryBreak()
		self._regionViewport = None
		self.finishSignal.emit()

		# Redraw
		self._view.postRedraw()



	def evPaint3d(self, viewport, paintLayer):
		# Paint the region
		self._o_paintRegion( viewport, paintLayer )


		if paintLayer == PaintLayer.PAINTLAYER_OVERLAY_BLEND_NODEPTH:
			if viewport is self._regionViewport:
				glDepthMask( False )

				glMatrixMode( GL_PROJECTION )
				glPushMatrix()
				viewport.glProjectionIdentity()

				glMatrixMode( GL_MODELVIEW )
				glPushMatrix()
				viewport.glModelViewIdentity()

				box = BBox2( self._p1, self._p2 )
				ModelDraw.drawMarkBox( box )

				glPopMatrix()
				glMatrixMode( GL_PROJECTION )
				glPopMatrix()
				glMatrixMode( GL_MODELVIEW )

				glDepthMask( True )





class RectangularMarkRegionColummnManipulator (GSProcedureTool.GSProcedureToolColumnManipulator):
	def __init__(self, predicateFieldColumn, regionFieldColumn, backfaceCullFieldColumn, bEnabledFieldColumn):
		super( RectangularMarkRegionColummnManipulator, self ).__init__()
		self._predicateFieldColumn = predicateFieldColumn
		self._regionFieldColumn = regionFieldColumn
		self._backfaceCullFieldColumn = backfaceCullFieldColumn
		self._bEnabledFieldColumn = bEnabledFieldColumn


	def createElement(self, rowEditContext):
		return RectangularMarkRegionManipulator()

	def initialiseElement(self, element, rowEditContext):
		if self._predicateFieldColumn is not None:
			predicate = self._p_getVar( self._predicateFieldColumn, rowEditContext )
			element.attachPredicateVar( predicate )
		if self._regionFieldColumn is not None:
			region = self._p_getVar( self._regionFieldColumn, rowEditContext )
			element.attachRegionVar( region )
		if self._backfaceCullFieldColumn is not None:
			backfaceCull = self._p_getVar( self._backfaceCullFieldColumn, rowEditContext )
			element.attachBackfaceCullVar( backfaceCull )
		if self._bEnabledFieldColumn is not None:
			bEnabled = self._p_getVar( self._bEnabledFieldColumn, rowEditContext )
			element.attachEnabledVar( bEnabled )
		if rowEditContext.commandHistory is not None:
			element.commandHistory = rowEditContext.commandHistory

	def shutdownElement(self, element, rowEditContext):
		element.detachVariables()
		element.commandHistory = None


	def installElementEventHandlers(self, element, rowEditContext):
		element.manipAttachToView( rowEditContext.view )

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.manipDetachFromView()
