##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from gSculpt.FileIO.IOXml import IOXmlEnum, ioObjectFactoryRegisterPrimitive

from gSculpt.UI.FormLayout import FormLayout

from gSculpt.Graphics.Graphics import BackfaceCull, Region3d
from gSculpt.Graphics.BackfaceCullField import BackfaceCullField
from gSculpt.Graphics.Region3dField import Region3dField

from gSculpt.Kernel.DB import *
from gSculpt.Kernel.ColumnEditors import BoolColumnEditCheck
from gSculpt.Kernel.Enum import Enum
from gSculpt.Kernel.Variable import VarInterface, Variable

from gSculpt.Model.Model import MarkPredicate
import gSculpt.Model.MarkPredicateVar
from gSculpt.Model.MarkPredicateColumnEditor import MarkPredicateColumnEditRadioGroup

from gSculpt.ProceduralCore.GSFunction import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *

from gSculpt.Mesh.Mesh import GSProductMesh





class MeshMarkRegionTarget (Enum):
	VERTICES = 0
	EDGES = 1
	FACES = 2



targetIOXmlEnum = IOXmlEnum( MeshMarkRegionTarget, 'MeshMarkRegionTarget', MeshMarkRegionTarget.VERTICES )
targetIOXmlEnum.value( MeshMarkRegionTarget.VERTICES, 'vertices' )
targetIOXmlEnum.value( MeshMarkRegionTarget.EDGES, 'edges' )
targetIOXmlEnum.value( MeshMarkRegionTarget.FACES, 'faces' )

ioXmlReadMeshMarkRegionTargetProp = targetIOXmlEnum.propReader
ioXmlReadMeshMarkRegionTargetNode = targetIOXmlEnum.nodeReader
ioXmlWriteMeshMarkRegionTargetProp = targetIOXmlEnum.propWriter
ioXmlWriteMeshMarkRegionTargetNode = targetIOXmlEnum.nodeWriter

ioObjectFactoryRegisterPrimitive( 'MeshMarkRegionTarget', MeshMarkRegionTarget )



# Variables and fields for MeshMarkRegionTarget

class MeshMarkRegionTargetVarInterface (VarInterface):
	valueClass = MeshMarkRegionTarget


class MeshMarkRegionTargetVar (Variable):
	valueClass = MeshMarkRegionTarget


class MeshMarkRegionTargetField (Field):
	valueClass = MeshMarkRegionTarget
	defaultValue = MeshMarkRegionTarget.VERTICES
	xmlNodeValueReader = ioXmlReadMeshMarkRegionTargetNode
	xmlNodeValueWriter = ioXmlWriteMeshMarkRegionTargetNode

registerFieldType( MeshMarkRegionTarget, MeshMarkRegionTargetField )






class ProcMeshMarkRegion (GSProcedure):
	target = FieldColumn( MeshMarkRegionTarget, MeshMarkRegionTarget.VERTICES )
	predicate = FieldColumn( MarkPredicate, MarkPredicate.MARK )
	region = FieldColumn( Region3d, None )
	backfaceCull = FieldColumn( BackfaceCull, None )
	bEnabled = FieldColumn( bool, False )
	bCullEnabled = FieldColumn( bool, False )
	bCompletelyContained = FieldColumn( bool, False )

	description = _( 'Abstract region select' )
	inputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		if self.bEnabled:
			cull = None
			if self.bCullEnabled:
				cull = self.backfaceCull

			target = self.target
			if target == MeshMarkRegionTarget.VERTICES:
				inputObject.markVertices_region( self.region, cull, self.predicate )
			elif target == MeshMarkRegionTarget.EDGES:
				inputObject.markEdges_region( self.region, cull, self.predicate, self.bCompletelyContained )
			elif target == MeshMarkRegionTarget.FACES:
				inputObject.markFaces_region( self.region, cull, self.predicate, self.bCompletelyContained )
		return inputObject





class MeshMarkRegionAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcMeshMarkRegion

	def save(self):
		if self._proc.target == MeshMarkRegionTarget.VERTICES:
			self._inputObject.markVertices_save()
		elif self._proc.target == MeshMarkRegionTarget.EDGES:
			self._inputObject.markEdges_save()
		elif self._proc.target == MeshMarkRegionTarget.FACES:
			self._inputObject.markFaces_save()

	def restore(self):
		if self._proc.target == MeshMarkRegionTarget.VERTICES:
			self._inputObject.markVertices_restore()
		elif self._proc.target == MeshMarkRegionTarget.EDGES:
			self._inputObject.markEdges_restore()
		elif self._proc.target == MeshMarkRegionTarget.FACES:
			self._inputObject.markFaces_restore()

	def adjust(self):
		if self._proc.bEnabled:
			cull = None
			if self._proc.bCullEnabled:
				cull = self._proc.backfaceCull

			if self._proc.target == MeshMarkRegionTarget.VERTICES:
				self._inputObject.markVertices_region( self._proc.region, cull, self._proc.predicate )
			elif self._proc.target == MeshMarkRegionTarget.EDGES:
				self._inputObject.markEdges_region( self._proc.region, cull, self._proc.predicate, self._proc.bCompletelyContained )
			elif self._proc.target == MeshMarkRegionTarget.FACES:
				self._inputObject.markFaces_region( self._proc.region, cull, self._proc.predicate, self._proc.bCompletelyContained )

	def getResult(self):
		return self._inputObject




class ToolMeshMarkRegion (GSProcedureTool.GSProcedureTool):
	row = RowRefColumn( ProcMeshMarkRegion )

	functionClass = ProcMeshMarkRegion
	adjustableInvoker = MeshMarkRegionAdjustableInvoker
	pageTitle = _( 'Reg.Sel.' )
	titleText = _( 'Abstract Region Select' )


	predicateRadio = MarkPredicateColumnEditRadioGroup( row.predicate )
	bCullEnabledCheck = BoolColumnEditCheck( row.bCullEnabled, _( 'Ignore back-facing' ) )
	bCompletelyContainedCheck = BoolColumnEditCheck( row.bCompletelyContained, _( 'Completely contained' ) )

	def layout(self):
		layout = FormLayout()
		layout.row()  <<  _( 'Mode:' )
		layout.row()  <<  self.predicateRadio.mark
		layout.row()  <<  self.predicateRadio.unmark
		layout.row()  <<  self.predicateRadio.add
		layout.row()  <<  self.predicateRadio.filter
		layout.row()  <<  self.predicateRadio.invert
		layout.row()
		layout.row()  <<  self.bCullEnabledCheck.check
		if self.row.target != MeshMarkRegionTarget.VERTICES:
			layout.row()  <<  self.bCompletelyContainedCheck.check
		layout.row()
		layout.finalise()
		layout.getWidget().show()
		return layout.getWidget()



