##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import deepcopy

from gSculpt.Math.Math import Colour3f, Point3, Segment3

from gSculpt.View.View import PaintLayer

from gSculpt.WorkArea.Viewport3d import Viewport3d

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.Kernel.DB import *

from gSculpt.ProceduralCore.GSFunction import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *

from gSculpt.Model import ModelDraw, Pick3d
from gSculpt.Model.Model import pyMPickList_to_MPickList, MPickList
from gSculpt.Model.TargetList import TargetList
from gSculpt.Model.TargetListColumnEditor import TargetListColumnEditor

from gSculpt.Mesh.Mesh import GSProductMesh
from gSculpt.Mesh.MeshPickHelper import MeshPickHelper

from gSculpt.Manipulator.Manipulator import Manipulator




class PickPathState (object):
	def __init__(self):
		self.prevVertexIndex = -1



class ProcMeshPickEdgeLoopPath (GSProcedure):
	targetList = FieldColumn( TargetList, TargetList() )

	description = _( 'Pick edge loop path' )
	inputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		mpicks = MPickList()
		pyMPickList_to_MPickList( [ pick.toMPick()  for  pick in self.targetList ], mpicks )
		inputObject.markEdgeLoopPath_pick( mpicks )
		return inputObject




class MeshPickEdgeLoopPathManipulator (Manipulator, MeshPickHelper):
	def __init__(self):
		super( MeshPickEdgeLoopPathManipulator, self ).__init__()
		self._var = None
		self.pathState = None

		self._prevVertexPosition = None
		self._targetVertexIndex = None
		self._targetVertexPosition = None
		self._pathEdgeSegments = None
		self._bTargetValid = False

		self._o_setUISensitivity( False )


	def manipAttachToView(self, view):
		super( MeshPickEdgeLoopPathManipulator, self ).manipAttachToView( view )
		self.attachView( view )
		self._view.addEventHandler( self, Viewport3d )		# HACK: need to handle other viewport types
		self._view.addPainter( self, Viewport3d )
		self._view.postRedraw()

	def manipDetachFromView(self):
		self._view.removeEventHandler( self, Viewport3d )
		self._view.removePainter( self, Viewport3d )
		self._view.postRedraw()
		self.detachView()
		super( MeshPickEdgeLoopPathManipulator, self ).manipDetachFromView()


	def attachVariable(self, var):
		assert self._var is None, 'already attached to a variable'
		self._var = var
		self._var.changedSignal.connect( self._p_onVarChanged )
		self._o_watchVarValidity( self._var )
		self._o_refreshSensitivity()

	def detachVariable(self):
		assert self._var is not None, 'not attached to a variable'
		self._var.changedSignal.disconnect( self._p_onVarChanged )
		self._o_stopWatchingVarValidity( self._var )
		self._var = None
		self._o_refreshSensitivity()



	def _o_setUISensitivity(self, bSensitive):
		self._o_setMeshPickHelperSensitivity( bSensitive )

	def _o_computeSensitivity(self):
		if self._var is not None:
			return self._var.isValid()
		else:
			return False



	def _p_refreshHit(self, viewport, mesh, pick):
		if self.pathState is not None  and  self.pathState.prevVertexIndex != -1:
			self._prevVertexPosition = mesh.getVertexPosition( self.pathState.prevVertexIndex )
		else:
			self._prevVertexPosition = None
		vertexIndex, thruFaceIndex = mesh.pickVertex( pick.toMPick() )
		if vertexIndex != -1:
			if vertexIndex != self._targetVertexIndex:
				self._targetVertexIndex = vertexIndex
				self._targetVertexPosition = mesh.getVertexPosition( vertexIndex )
				if self.pathState is not None  and  self.pathState.prevVertexIndex != -1:
					self._bTargetValid, pathEdgeIndices = mesh.discoverEdgeLoopPath( self.pathState.prevVertexIndex, vertexIndex )
					if self._bTargetValid:
						self._pathEdgeSegments = [ mesh.getEdgeSegment( pathEdgeIndex )   for pathEdgeIndex in pathEdgeIndices ]
					else:
						self._pathEdgeSegments = None
				else:
					self._pathEdgeSegments = None
					self._bTargetValid = True
			return True
		else:
			self._targetVertexIndex = None
			self._targetVertexPosition = None
			self._pathEdgeSegments = None
			self._bTargetvalid = False
			return False



	def _p_paintHighlight(self, viewport, paintLayer):
		if self._bTargetValid:
			ModelDraw.drawVertexPickHighlight( self._targetVertexPosition, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )
		else:
			ModelDraw.drawVertexPickHighlight( self._targetVertexPosition, ModelDraw.INVALID_TARGET_HIGHLIGHT_COLOUR, paintLayer )

		if self._pathEdgeSegments is not None:
			for seg in self._pathEdgeSegments:
				ModelDraw.drawEdgePickHighlight( seg, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )

		if self._prevVertexPosition is not None:
			ModelDraw.drawVertexPickHighlight( self._prevVertexPosition, ModelDraw.PATH_HIGHLIGHT_COLOUR, paintLayer )


	def _p_pickConfirmed(self, viewport, pick):
		"Protected - call to commit a pick"
		assert self._var is not None, 'no variable attached'
		self._var.value = self._var.value + [ deepcopy( pick ) ]
		self._p_commandHistoryBreak()


	def _o_resetHighlight(self):
		"Protected - reset the highlight"
		self._prevVertexPosition = None
		self._targetVertexIndex = None
		self._targetVertexPosition = None
		self._pathEdgeSegments = None
		self._bTargetValid = False


	def _p_onVarChanged(self):
		"Variable changed"
		self._o_resetHighlight()
		self._o_requestHighlightRefresh()





class MeshPickEdgeLoopPathColumnManipulator (GSProcedureTool.GSProcedureToolColumnManipulator):
	def __init__(self, fieldColumn):
		super( MeshPickEdgeLoopPathColumnManipulator, self ).__init__()
		self._fieldColumn = fieldColumn


	def createElement(self, rowEditContext):
		return MeshPickEdgeLoopPathManipulator()

	def initialiseElement(self, element, rowEditContext):
		field = self._p_getVar( self._fieldColumn, rowEditContext )
		element.attachVariable( field )
		element.pathState = rowEditContext.rowEditor._adjustableInvoker.pathState
		if rowEditContext.commandHistory is not None:
			element.commandHistory = rowEditContext.commandHistory

	def shutdownElement(self, element, rowEditContext):
		element.detachVariable()
		element.commandHistory = None
		element.pathState = None


	def installElementEventHandlers(self, element, rowEditContext):
		element.manipAttachToView( rowEditContext.view )

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.manipDetachFromView()


	def attachInputProductSource(self, element, source):
		element.attachMeshSource( source )

	def detachInputProductSource(self, element, source):
		element.detachMeshSource()




class MeshPickEdgeLoopPathAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcMeshPickEdgeLoopPath

	def __init__(self, *args, **kwargs):
		super( MeshPickEdgeLoopPathAdjustableInvoker, self ).__init__( *args, **kwargs )
		self.pathState = PickPathState()


	def save(self):
		self._inputObject.markEdges_save()

	def restore(self):
		self._inputObject.markEdges_restore()

	def adjust(self):
		mpicks = MPickList()
		pyMPickList_to_MPickList( [ pick.toMPick()  for  pick in self._proc.targetList ], mpicks )
		self._inputObject.markEdgeLoopPath_pick( mpicks )

		if len( self._proc.targetList ) > 0:
			self.pathState.prevVertexIndex, thruFaceIndex = self._inputObject.pickVertex( self._proc.targetList[-1].toMPick() )
		else:
			self.pathState.prevVertexIndex = -1

	def getResult(self):
		return self._inputObject




class ToolMeshPickEdgeLoopPath (GSProcedureTool.GSProcedureTool):
	row = RowRefColumn( ProcMeshPickEdgeLoopPath )

	functionClass = ProcMeshPickEdgeLoopPath
	adjustableInvoker = MeshPickEdgeLoopPathAdjustableInvoker
	pageTitle = _( 'LoopPth.' )
	titleText = _( 'Pick Edge Loop Path' )


	targetListEditor = TargetListColumnEditor( row.targetList )

	targetListManip = MeshPickEdgeLoopPathColumnManipulator( row.targetList )
	targetListManip.enable()


	layout = FormLayoutFactory()
	layout.row()  <<  targetListEditor.numTargets
	layout.row()  <<  None  <<  ( targetListEditor.removeTarget, 2 )  <<  None
	layout.row()
	layout.finalise()



