##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import deepcopy

from gSculpt.Math.Math import Colour3f, Point3, Segment3

from gSculpt.View.View import PaintLayer

from gSculpt.WorkArea.Viewport3d import Viewport3d

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.Kernel.DB import *

from gSculpt.ProceduralCore.GSFunction import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *

from gSculpt.Model import ModelDraw, Pick3d
from gSculpt.Model.Model import pyMPickList_to_MPickList, MPickList
from gSculpt.Model.TargetList import TargetList
from gSculpt.Model.TargetListColumnEditor import TargetListColumnEditor

from gSculpt.Mesh.Mesh import GSProductMesh
from gSculpt.Mesh.MeshPickHelper import MeshPickHelper

from gSculpt.Manipulator.Manipulator import Manipulator




class PickPathState (object):
	def __init__(self):
		self.prevEdgeIndex = -1



class ProcMeshPickEdgeRingPath (GSProcedure):
	targetList = FieldColumn( TargetList, TargetList() )

	description = _( 'Pick edge ring path' )
	inputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		mpicks = MPickList()
		pyMPickList_to_MPickList( [ pick.toMPick()  for  pick in self.targetList ], mpicks )
		inputObject.markEdgeRingPath_pick( mpicks )
		return inputObject




class MeshPickEdgeRingPathManipulator (Manipulator, MeshPickHelper):
	def __init__(self):
		super( MeshPickEdgeRingPathManipulator, self ).__init__()
		self._var = None
		self.pathState = None

		self._prevEdgeSegment = None
		self._targetEdgeIndex = None
		self._targetEdgeSegment = None
		self._pathEdgeSegments = None
		self._pathPoints = None
		self._bTargetValid = False

		self._o_setUISensitivity( False )



	def manipAttachToView(self, view):
		super( MeshPickEdgeRingPathManipulator, self ).manipAttachToView( view )
		self.attachView( view )
		self._view.addEventHandler( self, Viewport3d )		# HACK: need to handle other viewport types
		self._view.addPainter( self, Viewport3d )
		self._view.postRedraw()

	def manipDetachFromView(self):
		self._view.removeEventHandler( self, Viewport3d )
		self._view.removePainter( self, Viewport3d )
		self._view.postRedraw()
		self.detachView()
		super( MeshPickEdgeRingPathManipulator, self ).manipDetachFromView()


	def attachVariable(self, var):
		assert self._var is None, 'already attached to a variable'
		self._var = var
		self._var.changedSignal.connect( self._p_onVarChanged )
		self._o_watchVarValidity( self._var )
		self._o_refreshSensitivity()

	def detachVariable(self):
		assert self._var is not None, 'not attached to a variable'
		self._var.changedSignal.disconnect( self._p_onVarChanged )
		self._o_stopWatchingVarValidity( self._var )
		self._var = None
		self._o_refreshSensitivity()



	def _o_setUISensitivity(self, bSensitive):
		self._o_setMeshPickHelperSensitivity( bSensitive )

	def _o_computeSensitivity(self):
		if self._var is not None:
			return self._var.isValid()
		else:
			return False



	def _p_refreshHit(self, viewport, mesh, pick):
		edgeIndex, closestPoint, thruFaceIndex = mesh.pickEdge( pick.toMPick() )
		if self.pathState is not None  and  self.pathState.prevEdgeIndex != -1:
			self._prevEdgeSegment = mesh.getEdgeSegment( self.pathState.prevEdgeIndex )
		else:
			self._prevEdgeSegment = None
		if edgeIndex != -1:
			if edgeIndex != self._targetEdgeIndex:
				self._targetEdgeIndex = edgeIndex
				self._targetEdgeSegment = mesh.getEdgeSegment( edgeIndex )
				if self.pathState is not None  and  self.pathState.prevEdgeIndex != -1:
					self._bTargetValid, pathEdgeIndices = mesh.discoverEdgeRingPath( self.pathState.prevEdgeIndex, edgeIndex )
					if self._bTargetValid:
						self._pathEdgeSegments = [ mesh.getEdgeSegment( pathEdgeIndex )   for pathEdgeIndex in pathEdgeIndices ]
						self._pathPoints = [ seg.getMidPoint()   for seg in self._pathEdgeSegments ]
					else:
						self._pathEdgeSegments = None
						self._pathPoints = None
				else:
					self._pathEdgeSegments = None
					self._pathPoints = None
					self._bTargetValid = True
			return True
		else:
			self._targetEdgeIndex = None
			self._targetEdgeSegment = None
			self._pathEdgeSegments = None
			self._pathPoints = None
			self._bTargetvalid = False
			return False



	def _p_paintHighlight(self, viewport, paintLayer):
		if self._pathEdgeSegments is not None:
			for seg in self._pathEdgeSegments:
				ModelDraw.drawEdgePickHighlight( seg, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )
			if self._pathPoints is not None:
				ModelDraw.drawPickPath( self._pathPoints, ModelDraw.PATH_HIGHLIGHT_COLOUR, paintLayer )
		else:
			if self._bTargetValid:
				ModelDraw.drawEdgePickHighlight( self._targetEdgeSegment, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )
			else:
				ModelDraw.drawEdgePickHighlight( self._targetEdgeSegment, ModelDraw.INVALID_TARGET_HIGHLIGHT_COLOUR, paintLayer )
		if self._prevEdgeSegment is not None:
			ModelDraw.drawEdgePickHighlight( self._prevEdgeSegment, ModelDraw.PATH_HIGHLIGHT_COLOUR, paintLayer )



	def _p_pickConfirmed(self, viewport, pick):
		"Protected - call to commit a pick"
		assert self._var is not None, 'no variable attached'
		self._var.value = self._var.value + [ deepcopy( pick ) ]
		self._p_commandHistoryBreak()


	def _o_resetHighlight(self):
		"Protected - reset the highlight"
		self._prevEdgeSegment = None
		self._targetEdgeIndex = None
		self._targetEdgeSegment = None
		self._pathEdgeSegments = None
		self._pathPoints = None
		self._bTargetValid = False


	def _p_onVarChanged(self):
		"Variable changed"
		self._o_resetHighlight()
		self._o_requestHighlightRefresh()





class MeshPickEdgeRingPathColumnManipulator (GSProcedureTool.GSProcedureToolColumnManipulator):
	def __init__(self, fieldColumn):
		super( MeshPickEdgeRingPathColumnManipulator, self ).__init__()
		self._fieldColumn = fieldColumn


	def createElement(self, rowEditContext):
		return MeshPickEdgeRingPathManipulator()

	def initialiseElement(self, element, rowEditContext):
		field = self._p_getVar( self._fieldColumn, rowEditContext )
		element.attachVariable( field )
		element.pathState = rowEditContext.rowEditor._adjustableInvoker.pathState
		if rowEditContext.commandHistory is not None:
			element.commandHistory = rowEditContext.commandHistory

	def shutdownElement(self, element, rowEditContext):
		element.detachVariable()
		element.commandHistory = None
		element.pathState = None


	def installElementEventHandlers(self, element, rowEditContext):
		element.manipAttachToView( rowEditContext.view )

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.manipDetachFromView()


	def attachInputProductSource(self, element, source):
		element.attachMeshSource( source )

	def detachInputProductSource(self, element, source):
		element.detachMeshSource()




class MeshPickEdgeRingPathAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcMeshPickEdgeRingPath

	def __init__(self, *args, **kwargs):
		super( MeshPickEdgeRingPathAdjustableInvoker, self ).__init__( *args, **kwargs )
		self.pathState = PickPathState()


	def save(self):
		self._inputObject.markEdges_save()

	def restore(self):
		self._inputObject.markEdges_restore()

	def adjust(self):
		mpicks = MPickList()
		pyMPickList_to_MPickList( [ pick.toMPick()  for  pick in self._proc.targetList ], mpicks )
		self._inputObject.markEdgeRingPath_pick( mpicks )

		if len( self._proc.targetList ) > 0:
			self.pathState.prevEdgeIndex, closestPoint, thruFaceIndex = self._inputObject.pickEdge( self._proc.targetList[-1].toMPick() )
		else:
			self.pathState.prevEdgeIndex = -1

	def getResult(self):
		return self._inputObject




class ToolMeshPickEdgeRingPath (GSProcedureTool.GSProcedureTool):
	row = RowRefColumn( ProcMeshPickEdgeRingPath )

	functionClass = ProcMeshPickEdgeRingPath
	adjustableInvoker = MeshPickEdgeRingPathAdjustableInvoker
	pageTitle = _( 'RingPth.' )
	titleText = _( 'Pick Edge Ring Path' )


	targetListEditor = TargetListColumnEditor( row.targetList )

	targetListManip = MeshPickEdgeRingPathColumnManipulator( row.targetList )
	targetListManip.enable()


	layout = FormLayoutFactory()
	layout.row()  <<  targetListEditor.numTargets
	layout.row()  <<  None  <<  ( targetListEditor.removeTarget, 2 )  <<  None
	layout.row()
	layout.finalise()



