##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import copy, deepcopy

from OpenGL.GL import *
from OpenGL.GLU import *

from gSculpt.Math.Math import Colour3f, Point3, Segment3

from gSculpt.FileIO.IOXml import IOXmlEnum, ioObjectFactoryRegisterPrimitive, ioObjectFactoryRegister

from gSculpt.View.View import PaintLayer

from gSculpt.WorkArea.Viewport import ViewportEventHandler
from gSculpt.WorkArea.Viewport3d import Viewport3d, Viewport3dPainter

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.Graphics.Graphics import BackfaceCull
from gSculpt.Graphics.BackfaceCullField import BackfaceCullField

from gSculpt.Kernel.DB import *
from gSculpt.Kernel.ColumnEditors import BoolColumnEditCheck

from gSculpt.ProceduralCore.GSFunction import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *

from gSculpt.Model import ModelDraw
from gSculpt.Mesh.Mesh import GSProductMesh, MeshKnifeTargetConstraint, CppMeshKnifeTarget

from gSculpt.Model.Pick3d import Pick3d

from gSculpt.Manipulator.Manipulator import Manipulator





constraintTypeIOXmlEnum = IOXmlEnum( MeshKnifeTargetConstraint, 'MeshKnifeTargetConstraint', MeshKnifeTargetConstraint.NOTHING )
constraintTypeIOXmlEnum.value( MeshKnifeTargetConstraint.NOTHING, 'nothing' )
constraintTypeIOXmlEnum.value( MeshKnifeTargetConstraint.VERTEX, 'vertex' )
constraintTypeIOXmlEnum.value( MeshKnifeTargetConstraint.EDGE, 'edge' )

ioXmlReadMeshKnifeTargetConstraintProp = constraintTypeIOXmlEnum.propReader
ioXmlReadMeshKnifeTargetConstraintNode = constraintTypeIOXmlEnum.nodeReader
ioXmlWriteMeshKnifeTargetConstraintProp = constraintTypeIOXmlEnum.propWriter
ioXmlWriteMeshKnifeTargetConstraintNode = constraintTypeIOXmlEnum.nodeWriter

ioObjectFactoryRegisterPrimitive( 'MeshKnifeTargetConstraint', MeshKnifeTargetConstraint )





class MeshKnifeTarget (object):
	def __init__(self, pick=None, constraint=MeshKnifeTargetConstraint.NOTHING):
		super( MeshKnifeTarget, self ).__init__()
		if pick is None:
			pick = Pick3d()
		self.pick = pick
		self.constraint = constraint


	def toCppMeshKnifeTarget(self):
		return CppMeshKnifeTarget( self.pick.toMPick(), self.constraint )


	def __readxml__(self, xmlNode):
		xmlNode.getChild( 'pick' )  >>  self.pick
		self.constraint = ioXmlReadMeshKnifeTargetConstraintNode( xmlNode.getChild( 'constraint' ), self.constraint )

	def __writexml__(self, xmlNode):
		xmlNode.addChild( 'pick' )  <<  self.pick
		ioXmlWriteMeshKnifeTargetConstraintNode( xmlNode.addChild( 'constraint' ), self.constraint )


	def __copy__(self):
		return MeshKnifeTarget( self.pick, self.constraint )

ioObjectFactoryRegister( 'MeshKnifeTarget', MeshKnifeTarget )


class MeshKnifeTargetField (Field):
	valueClass = MeshKnifeTarget
	defaultValue = MeshKnifeTarget()

registerFieldType( MeshKnifeTarget, MeshKnifeTargetField )









class ProcMeshKnife (GSProcedure):
	startTarget = FieldColumn( MeshKnifeTarget, MeshKnifeTarget() )
	endTarget = FieldColumn( MeshKnifeTarget, MeshKnifeTarget() )
	faceCull = FieldColumn( BackfaceCull, None )
	bMarkedFacesOnly = FieldColumn( bool, False )
	bMarkCutEdges = FieldColumn( bool, True )
	bCullEnabled = FieldColumn( bool, False )
	bKnifeEnabled = FieldColumn( bool, False )

	description = _( 'Knife' )
	inputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		if self.bKnifeEnabled:
			cull = None
			if self.bCullEnabled:
				cull = self.faceCull
			inputObject.knife( self.startTarget.toCppMeshKnifeTarget(), self.endTarget.toCppMeshKnifeTarget(), self.bMarkedFacesOnly, cull, self.bMarkCutEdges )
		return inputObject





class MeshKnifeManipulator (Manipulator, ViewportEventHandler, Viewport3dPainter):
	def __init__(self):
		super( MeshKnifeManipulator, self ).__init__()

		self._startTargetVar = None
		self._endTargetVar = None
		self._knifeEnabledVar = None
		self._backfaceCullVar = None
		self._meshSource = None
		self._knifeViewport = None

		self._p1 = None
		self._p2 = None
		self._startPoint = None
		self._endPoint = None
		self._bButtonDown = False

		self._highlightVertexPosition = None
		self._highlightEdgeSegment = None
		self._highlightEdgeSplitPosition = None
		self._bHighlightEnabled = False
		self._highlightConstraint = None

		self._startTarget = None
		self._endTarget = None

		self._o_setUISensitivity( False )



	def manipAttachToView(self, view):
		assert self._view is None, 'already attached to a view'
		super( MeshKnifeManipulator, self ).manipAttachToView( view )
		self._view.addEventHandler( self, Viewport3d )
		self._view.addPainter( self, Viewport3d )
		self._view.postRedraw()

	def manipDetachFromView(self):
		assert self._view is not None, 'not attached to a view'
		self._view.removeEventHandler( self, Viewport3d )
		self._view.removePainter( self, Viewport3d )
		self._view.postRedraw()
		super( MeshKnifeManipulator, self ).manipDetachFromView()



	def attachVariables(self, startTarget, endTarget, knifeEnabled, backfaceCull):
		self._startTargetVar = startTarget
		self._endTargetVar = endTarget
		self._knifeEnabledVar = knifeEnabled
		self._backfaceCullVar = backfaceCull

	def detachVariables(self):
		self._startTargetVar = None
		self._endTargetVar = None
		self._knifeEnabledVar = None
		self._backfaceCullVar = None



	def attachMeshSource(self, source):
		assert self._meshSource is None, 'mesh source already attached'
		self._meshSource = source
		self._meshSource.changedSignal.connect( self._p_onMeshSourceValueChanged )

	def detachMeshSource(self):
		assert self._meshSource is not None, 'no mesh source attached'
		self._meshSource.changedSignal.disconnect( self._p_onMeshSourceValueChanged )
		self._meshSource = None


	def _o_setUISensitivity(self, bSensitive):
		pass



	def evPrimaryButtonDown(self, viewport):
		super( MeshKnifeManipulator, self ).evPrimaryButtonDown( viewport )

		# Note the viewport used for this operation
		self._knifeViewport = viewport

		# Backface culling
		if self._backfaceCullVar is not None:
			self._backfaceCullVar.value = viewport.camera.createBackfaceCull()

		# Set start target
		self._startTarget = self._p_createKnifeTarget( viewport )

		# Note the pointer position
		self._p1 = viewport.pointerStatus.positionF
		self._endPoint = copy( self._startPoint )

		# Button pressed
		self._bButtonDown = True


	def evPrimaryButtonUp(self, viewport):
		super( MeshKnifeManipulator, self ).evPrimaryButtonUp( viewport )

		# Button released
		self._bButtonDown = False
		self._bHighlightEnabled = False

		# Set end target
		self._endTarget = self._p_createKnifeTarget( viewport )

		# Refesh knife variables
		self._p_setKnife( viewport )

		# Finished
		self._p_commandHistoryBreak()
		self._endPoint = None

		# Redraw
		self._view.postRedraw()



	def evMotion(self, viewport, point):
		super( MeshKnifeManipulator, self ).evMotion( viewport, point )

		# Refresh the highlight
		if self._bButtonDown:
			self._endPoint = self._p_refreshHighlight( viewport )
		else:
			self._startPoint = self._p_refreshHighlight( viewport )

		# Need an immediate redraw
		viewport.forceRedraw()


	def evLeave(self, viewport, point):
		super( MeshKnifeManipulator, self ).evLeave( viewport, point )

		# Disable highlight
		self._bHighlightEnabled = False

		# Need a redraw
		viewport.postRedraw()



	def evPaint3d(self, viewport, paintLayer):
		# Paint the target highlight
		if self._bHighlightEnabled:
			if self._highlightVertexPosition is not None:
				ModelDraw.drawVertexPickHighlight( self._highlightVertexPosition, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )
			if self._highlightEdgeSegment is not None:
				ModelDraw.drawEdgePickHighlight( self._highlightEdgeSegment, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )
			if paintLayer == PaintLayer.PAINTLAYER_OVERLAY_NODEPTH  and  self._highlightEdgeSplitPosition is not None:
				ModelDraw.drawSplitPosition( self._highlightEdgeSplitPosition )

		if viewport is self._knifeViewport:
			if paintLayer == PaintLayer.PAINTLAYER_OVERLAY_NODEPTH:
				if self._bButtonDown:
					glDepthMask( False )

					glMatrixMode( GL_PROJECTION )
					glPushMatrix()
					viewport.glProjectionIdentity()

					glMatrixMode( GL_MODELVIEW )
					glPushMatrix()
					viewport.glModelViewIdentity()


					glPointSize( 7.0 )
					glColor4d( 0.0, 0.6, 1.0, 0.5 )
					glBegin( GL_POINTS )
					glVertex2d( self._startPoint.x, self._startPoint.y )
					glVertex2d( self._endPoint.x, self._endPoint.y )
					glEnd()

					glBegin( GL_LINES )
					glVertex2d( self._startPoint.x, self._startPoint.y )
					glVertex2d( self._endPoint.x, self._endPoint.y )
					glEnd()

					glPointSize( 1.0 )

					glPopMatrix()
					glMatrixMode( GL_PROJECTION )
					glPopMatrix()
					glMatrixMode( GL_MODELVIEW )

					glDepthMask( True )



	def _p_setKnife(self, viewport):
		# Set the knife target vars
		if self._startTargetVar is not None  and  self._startTarget is not None:
			self._startTargetVar.value = copy( self._startTarget )
		if self._endTargetVar is not None  and  self._endTarget is not None:
			self._endTargetVar.value = copy( self._endTarget )
		if self._knifeEnabledVar is not None  and  self._startTarget is not None  and  self._endTarget is not None:
			self._knifeEnabledVar.value = True



	def _p_refreshHighlight(self, viewport):
		self._highlightVertexPosition = None
		self._highlightEdgeSegment = None
		self._highlightEdgeSplitPosition = None
		self._bHighlightEnabled = False


		self._highlightConstraint = self._p_getConstraint( viewport )
		pick = viewport.createPick()

		# Get the screen position
		screenPoint = pick.screenPos

		mesh = self._p_getMesh()

		if self._highlightConstraint == MeshKnifeTargetConstraint.VERTEX:
			# Constrained to vertex
			vertexIndex, thruFaceIndex = mesh.pickVertex( pick.toMPick() )

			if vertexIndex != -1:
				# Highlight the vertex
				self._highlightVertexPosition = mesh.getVertexPosition( vertexIndex )
				self._bHighlightEnabled = True
				screenPoint = pick.projection.worldToScreenMatrix.transformHomogeneous2d( self._highlightVertexPosition )
			else:
				self._highlightVertexPosition = None
				self._bHighlightEnabled = False
		elif self._highlightConstraint == MeshKnifeTargetConstraint.EDGE:
			# Constrained to edge
			edgeIndex, closestPoint, thruFaceIndex = mesh.pickEdge( pick.toMPick() )

			if edgeIndex != -1:
				# Highlight the edge
				self._highlightEdgeSegment = mesh.getEdgeSegment( edgeIndex )
				self._highlightEdgeSplitPosition = closestPoint
				self._bHighlightEnabled = True
				screenPoint = pick.projection.worldToScreenMatrix.transformHomogeneous2d( self._highlightEdgeSplitPosition )
			else:
				self._highlightEdgeSegment = None
				self._highlightEdgeSplitPosition = None
				self._bHighlightEnabled = False
		elif self._highlightConstraint == MeshKnifeTargetConstraint.NOTHING:
			# No constraint; no highlight
			self._bHighlightEnabled = False
		else:
			assert False, 'invalid constraint'

		return screenPoint



	def _p_getConstraint(self, viewport):
		pstat = viewport.pointerStatus
		if pstat.bShiftKey:
			return MeshKnifeTargetConstraint.VERTEX
		elif pstat.bControlKey:
			return MeshKnifeTargetConstraint.EDGE
		else:
			return MeshKnifeTargetConstraint.NOTHING


	def _p_createKnifeTarget(self, viewport):
		pick = viewport.createPick()
		return MeshKnifeTarget( viewport.createPick(), self._p_getConstraint( viewport ) )


	def _p_onMeshSourceValueChanged(self):
		if self._bButtonDown:
			if self._knifeViewport is not None:
				self._endPoint = self._p_refreshHighlight( self._knifeViewport )
				self._knifeViewport.postRedraw()
		else:
			viewport = self._view.pointerViewport
			if viewport is not None:
				self._startPoint = self._p_refreshHighlight( viewport )
				viewport.postRedraw()


	def _p_getMesh(self):
		if self._meshSource is not None:
			return self._meshSource.getValueAsClass( GSProductMesh )
		else:
			return None




class MeshKnifeColumnManipulator (GSProcedureTool.GSProcedureToolColumnManipulator):
	def __init__(self, startTargetFieldColumn, endTargetFieldColumn, knifeEnabledFieldColumn, backfaceCullFieldColumn):
		super( MeshKnifeColumnManipulator, self ).__init__()
		self._startTargetFieldColumn = startTargetFieldColumn
		self._endTargetFieldColumn = endTargetFieldColumn
		self._knifeEnabledFieldColumn = knifeEnabledFieldColumn
		self._backfaceCullFieldColumn = backfaceCullFieldColumn


	def createElement(self, rowEditContext):
		return MeshKnifeManipulator()

	def initialiseElement(self, element, rowEditContext):
		startTarget = self._p_getVar( self._startTargetFieldColumn, rowEditContext )
		endTarget = self._p_getVar( self._endTargetFieldColumn, rowEditContext )
		knifeEnabled = self._p_getVar( self._knifeEnabledFieldColumn, rowEditContext )
		backfaceCull = self._p_getVar( self._backfaceCullFieldColumn, rowEditContext )
		element.attachVariables( startTarget, endTarget, knifeEnabled, backfaceCull )
		if rowEditContext.commandHistory is not None:
			element.commandHistory = rowEditContext.commandHistory

	def shutdownElement(self, element, rowEditContext):
		element.detachVariables()
		element.commandHistory = None


	def installElementEventHandlers(self, element, rowEditContext):
		element.manipAttachToView( rowEditContext.view )

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.manipDetachFromView()


	def attachInputProductSource(self, element, source):
		element.attachMeshSource( source )

	def detachInputProductSource(self, element, source):
		element.detachMeshSource()




class ToolMeshKnife (GSProcedureTool.GSProcedureTool):
	row = RowRefColumn( ProcMeshKnife )

	functionClass = ProcMeshKnife
	pageTitle = _( 'Knife' )
	titleText = _( 'Knife' )


	bMarkedFacesOnlyCheck = BoolColumnEditCheck( row.bMarkedFacesOnly, _( 'Marked faces only' ) )
	bMarkCutEdgesCheck = BoolColumnEditCheck( row.bMarkCutEdges, _( 'Mark cut edges' ) )
	bCullEnabledCheck = BoolColumnEditCheck( row.bCullEnabled, _( 'Cut front faces only' ) )

	knifeManip = MeshKnifeColumnManipulator( row.startTarget, row.endTarget, row.bKnifeEnabled, row.faceCull )
	knifeManip.enable()


	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Click to set start point,' )
	layout.row()  <<  _( 'hold, move to end point,' )
	layout.row()  <<  _( 'and release.' )
	layout.row()
	layout.row()  <<  _( 'Hold <shift> while pressing or' )
	layout.row()  <<  _( 'releasing to constrain to vertices' )
	layout.row()  <<  _( 'Hold <control> while pressing or' )
	layout.row()  <<  _( 'releasing to constrain to edges' )
	layout.row()
	layout.row()  <<  bMarkedFacesOnlyCheck.check
	layout.row()  <<  bMarkCutEdgesCheck.check
	layout.row()  <<  bCullEnabledCheck.check
	layout.row()
	layout.finalise()



