##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from OpenGL.GL import *

from copy import deepcopy

from gSculpt.Math.Math import Colour3f, Point3, Segment3

from gSculpt.FileIO.IOXml import ioObjectFactoryRegister

from gSculpt.Kernel.Variable import BoolVar
from gSculpt.Kernel.DB import *

from gSculpt.View.View import PaintLayer

from gSculpt.WorkArea.Viewport3d import Viewport3d
from gSculpt.Graphics.Graphics import glColor3Colour3f, glVertex3Point3

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.ProceduralCore.GSFunction import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *
from gSculpt.Kernel.ColumnEditors import *

from gSculpt.Mesh.Mesh import GSProductMesh
from gSculpt.Mesh.MeshPickHelper import MeshPickHelper

from gSculpt.Model import ModelDraw
from gSculpt.Model.TargetList import TargetList
from gSculpt.Model.TargetListColumnEditor import TargetListColumnEditor
from gSculpt.Model.Model import pyMPickList_to_MPickList, MPickList

from gSculpt.Manipulator.Manipulator import Manipulator






class ProcMeshConnectEdges (GSProcedure):
	targetList = FieldColumn( TargetList, TargetList() )
	terminalVertexTargetList = FieldColumn( TargetList, TargetList() )

	description = _( 'Connect edges' )
	inputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		mpicks = MPickList()
		terminalMPicks = MPickList()
		pyMPickList_to_MPickList( [ pick.toMPick()  for  pick in self.targetList ], mpicks )
		pyMPickList_to_MPickList( [ pick.toMPick()  for  pick in self.terminalVertexTargetList ], terminalMPicks )
		inputObject.connectMarkedEdges( mpicks, terminalMPicks, True )
		return inputObject




class MeshConnectEdgesManipulator (Manipulator, MeshPickHelper):
	TARGET_TERMINALVERTEX = 0
	TARGET_EDGE = 1

	def __init__(self):
		super( MeshConnectEdgesManipulator, self ).__init__()
		self._var = None
		self._terminalVar = None
		self.connectState = None

		self._seedEdgeIndex = -1
		self._seedEdgeSegment = None
		self._seedPickPoint = None
		self._bSeedValid = False

		self._terminalVertexPosition = None

		self._ringSegments = []
		self._cutPoints = []
		self._bClosed = False

		self._target = MeshConnectEdgesManipulator.TARGET_EDGE

		self._o_setUISensitivity( False )




	def manipAttachToView(self, view):
		super( MeshConnectEdgesManipulator, self ).manipAttachToView( view )
		self.attachView( view )
		self._view.addEventHandler( self, Viewport3d )		# HACK: need to handle other viewport types
		self._view.addPainter( self, Viewport3d )
		self._view.postRedraw()

	def manipDetachFromView(self):
		self._view.removeEventHandler( self, Viewport3d )
		self._view.removePainter( self, Viewport3d )
		self._view.postRedraw()
		self.detachView()
		super( MeshConnectEdgesManipulator, self ).manipDetachFromView()


	def attachVariables(self, var, terminalVar):
		assert self._var is None, 'already attached to a variable'
		self._var = var
		self._terminalVar = terminalVar
		self._o_watchVarValidity( self._var )
		self._o_watchVarValidity( self._terminalVar )
		self._o_refreshSensitivity()

	def detachVariables(self):
		assert self._var is not None, 'not attached to a variable'
		self._o_stopWatchingVarValidity( self._var )
		self._o_stopWatchingVarValidity( self._terminalVar )
		self._var = None
		self._terminalVar = None
		self._o_refreshSensitivity()



	def _o_setUISensitivity(self, bSensitive):
		self._o_setMeshPickHelperSensitivity( bSensitive )

	def _o_computeSensitivity(self):
		if self._var is not None  and  self._terminalVar is not None:
			return self._var.isValid()  and  self._terminalVar.isValid()
		else:
			return False



	def _p_refreshHit(self, viewport, mesh, pick):
		if viewport.pointerStatus.bShiftKey:
			self._target = MeshConnectEdgesManipulator.TARGET_TERMINALVERTEX
			pickedVertexIndex, thruFaceIndex = mesh.pickVertex( pick.toMPick() )
			if pickedVertexIndex != -1:
				self._terminalVertexPosition = mesh.getVertexPosition( pickedVertexIndex )
				return True
			else:
				self._terminalVertexPosition = None
				return False
		else:
			self._target = MeshConnectEdgesManipulator.TARGET_EDGE
			pickedEdgeIndex, self._seedPickPoint, pickedFaceIndex = mesh.pickEdge( pick.toMPick() )

			if pickedEdgeIndex != self._seedEdgeIndex:
				# The seed has changed
				self._seedEdgeIndex = pickedEdgeIndex

				if pickedEdgeIndex != -1:
					self._seedEdgeSegment = mesh.getEdgeSegment( pickedEdgeIndex )
					if mesh.isEdgeMarked( pickedEdgeIndex ):
						# An edge was chosed; discover the bandsaw edge ring
						self._ringSegments, self._bClosed = mesh.connectMarkedEdgesGetRingSegments( pickedEdgeIndex )
						self._bSeedValid = True
					else:
						self._ringSegments = []
						self._bClosed = False
						self._bSeedValid = False
						self._cutPoints = []
						return False

			if pickedEdgeIndex != -1:
				point, t = self._seedEdgeSegment.closestPointTo( self._seedPickPoint )
				self._cutPoints = [ segment.getPoint( t )  for segment in self._ringSegments ]
				return True
			else:
				self._ringSegments = []
				self._cutPoints = []
				return False



	def _p_paintHighlight(self, viewport, paintLayer):
		if self._target == MeshConnectEdgesManipulator.TARGET_TERMINALVERTEX:
			if self._terminalVertexPosition is not None:
				ModelDraw.drawVertexPickHighlight( self._terminalVertexPosition, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )
		elif self._target == MeshConnectEdgesManipulator.TARGET_EDGE:
			# Highlight the seed edge
			if self._bSeedValid:
				ModelDraw.drawEdgePickHighlight( self._seedEdgeSegment, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )

				# Draw the split position on the seed edge
				ModelDraw.drawSplitPosition( self._seedPickPoint )

				# Draw the bandsaw line
				if paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY  or  paintLayer == PaintLayer.PAINTLAYER_WIREFRAME:
					colour = ModelDraw.MODIFICATION1_HIGHLIGHT_COLOUR

					if paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY:
						colour = ModelDraw.computeOverlayColour( colour )

					glColor3Colour3f( colour )

					if self._bClosed:
						glBegin( GL_LINE_LOOP )
					else:
						glBegin( GL_LINE_STRIP )

					for point in self._cutPoints:
						glVertex3Point3( point )

					glEnd()
			else:
				ModelDraw.drawEdgePickHighlight( self._seedEdgeSegment, ModelDraw.INVALID_TARGET_HIGHLIGHT_COLOUR, paintLayer )


	def evPaint3d(self, viewport, paintLayer):
		super( MeshConnectEdgesManipulator, self ).evPaint3d( viewport, paintLayer )

		if self.connectState is not None:
			for point in self.connectState:
				ModelDraw.drawVertexReticle( point, paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY )


	def _p_pickConfirmed(self, viewport, pick):
		"Protected - call to commit a pick"
		assert self._var is not None  and  self._terminalVar is not None, 'no variable attached'
		if self._target == MeshConnectEdgesManipulator.TARGET_TERMINALVERTEX:
			self._terminalVar.value = self._terminalVar.value + [ pick ]
			self._p_commandHistoryBreak()
		elif self._target == MeshConnectEdgesManipulator.TARGET_EDGE:
			self._var.value = self._var.value + [ pick ]
			self._p_commandHistoryBreak()


	def _o_resetHighlight(self):
		self._seedEdgeIndex = -1
		self._seedEdgeSegment = None
		self._seedPickPoint = None

		self._ringSegments = []
		self._cutPoints = []
		self._bClosed = False


	def _p_onParamsChanged(self):
		self._o_resetHighlight()
		self._o_requestHighlightRefresh()





class MeshConnectEdgesTargetListColumnManipulator (GSProcedureTool.GSProcedureToolColumnManipulator):
	def __init__(self, cutTargetListFieldColumn, terminalVertexTargetListFieldColumn):
		super( MeshConnectEdgesTargetListColumnManipulator, self ).__init__()
		self._cutTargetListFieldColumn = cutTargetListFieldColumn
		self._terminalVertexTargetListFieldColumn = terminalVertexTargetListFieldColumn


	def createElement(self, rowEditContext):
		manip = MeshConnectEdgesManipulator()
		rowEditContext.rowEditor.targetListManip = manip
		return manip

	def destroyElement(self, element, rowEditContext):
		del rowEditContext.rowEditor.targetListManip

	def initialiseElement(self, element, rowEditContext):
		cutTargetListField = self._p_getVar( self._cutTargetListFieldColumn, rowEditContext )
		terminalVertexTargetListField = self._p_getVar( self._terminalVertexTargetListFieldColumn, rowEditContext )
		element.attachVariables( cutTargetListField, terminalVertexTargetListField )
		element.connectState = rowEditContext.rowEditor._adjustableInvoker.connectState
		if rowEditContext.commandHistory is not None:
			element.commandHistory = rowEditContext.commandHistory

	def shutdownElement(self, element, rowEditContext):
		element.detachVariables()
		element.commandHistory = None
		element.connectState = None


	def installElementEventHandlers(self, element, rowEditContext):
		element.manipAttachToView( rowEditContext.view )

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.manipDetachFromView()


	def attachInputProductSource(self, element, source):
		element.attachMeshSource( source )

	def detachInputProductSource(self, element, source):
		element.detachMeshSource()




class MeshConnectEdgesAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcMeshConnectEdges

	def __init__(self, *args, **kwargs):
		super( MeshConnectEdgesAdjustableInvoker, self ).__init__( *args, **kwargs )
		self.connectState = []


	def main(self):
		mpicks = MPickList()
		pyMPickList_to_MPickList( [ pick.toMPick()  for  pick in self._proc.targetList ], mpicks )

		terminalVertexMPicks = MPickList()
		pyMPickList_to_MPickList( [ pick.toMPick()  for  pick in self._proc.terminalVertexTargetList ], terminalVertexMPicks )

		terminalVertexIndices = self._inputObject.connectMarkedEdgesGetTerminalVertices( terminalVertexMPicks )
		self.connectState[:] = [ self._inputObject.getVertexPosition( vertexIndex )  for vertexIndex in terminalVertexIndices ]
		self.mesh = deepcopy( self._inputObject )
		self.mesh.connectMarkedEdges( mpicks, terminalVertexMPicks, False )

	def getResult(self):
		return self.mesh



class ToolMeshConnectEdges (GSProcedureTool.GSProcedureTool):
	row = RowRefColumn( ProcMeshConnectEdges )

	functionClass = ProcMeshConnectEdges
	adjustableInvoker = MeshConnectEdgesAdjustableInvoker
	pageTitle = _( 'Connect' )
	titleText = _( 'Connect Edges' )


	targetListEditor = TargetListColumnEditor( row.targetList )
	terminalVertexTargetListEditor = TargetListColumnEditor( row.terminalVertexTargetList )

	targetListManip = MeshConnectEdgesTargetListColumnManipulator( row.targetList, row.terminalVertexTargetList )
	targetListManip.enable()


	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Move the pointer onto the' )
	layout.row()  <<  _( 'mesh and highlight marked' )
	layout.row()  <<  _( 'edges to see a preview of' )
	layout.row()  <<  _( 'the cut that will be made.' )
	layout.row()  <<  _( 'Click to confirm the cut.' )
	layout.row()  <<  _( 'Shift-click to choose' )
	layout.row()  <<  _( 'terminal vertices, which' )
	layout.row()  <<  _( 'will be connected to the' )
	layout.row()  <<  _( 'ends of the cuts.' )
	layout.row()
	layout.row()  <<  _( 'Connect targets:' )
	layout.row()  <<  targetListEditor.numTargets
	layout.row()  <<  None  <<  ( targetListEditor.removeTarget, 2 )  <<  None
	layout.row()
	layout.row()  <<  _( 'Terminal vertex targets:' )
	layout.row()  <<  terminalVertexTargetListEditor.numTargets
	layout.row()  <<  None  <<  ( terminalVertexTargetListEditor.removeTarget, 2 )  <<  None
	layout.row()
	layout.finalise()



