##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from gSculpt.Math.Math import Vector3

from gSculpt.FileIO.IOXml import IOXmlEnum, ioObjectFactoryRegisterPrimitive

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.VarEdit.VarEdit import ChoiceVarEditRadioGroup

from gSculpt.Kernel.DB import *
from gSculpt.Kernel.DBEdit import BasicColumnWidgetEditor
from gSculpt.Kernel.ColumnEditors import *
from gSculpt.Kernel.Enum import Enum

from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool.GSFunctionTool import *
from gSculpt.ProceduralTool.GSProcedureTool import GSProcedureTool

from gSculpt.Mesh.Mesh import GSProductMesh




# MeshFlattenEdgesMode enumeration
class MeshFlattenEdgesMode (Enum):
	X = 0
	Y = 1
	Z = 2
	NORMAL = 3
	FLATTEN_RUNS = 4
	STRAIGHTEN_RUNS = 5




# XML IO for MeshFlattenEdgesMode

flattenModeIOXmlEnum = IOXmlEnum( MeshFlattenEdgesMode, 'MeshFlattenEdgesMode', MeshFlattenEdgesMode.NORMAL )
flattenModeIOXmlEnum.value( MeshFlattenEdgesMode.X, 'x' )
flattenModeIOXmlEnum.value( MeshFlattenEdgesMode.Y, 'y' )
flattenModeIOXmlEnum.value( MeshFlattenEdgesMode.Z, 'z' )
flattenModeIOXmlEnum.value( MeshFlattenEdgesMode.NORMAL, 'normal' )
flattenModeIOXmlEnum.value( MeshFlattenEdgesMode.FLATTEN_RUNS, 'flatten_runs' )
flattenModeIOXmlEnum.value( MeshFlattenEdgesMode.STRAIGHTEN_RUNS, 'straighten_runs' )

ioXmlReadMeshFlattenEdgesModeProp = flattenModeIOXmlEnum.propReader
ioXmlReadMeshFlattenEdgesModeNode = flattenModeIOXmlEnum.nodeReader
ioXmlWriteMeshFlattenEdgesModeProp = flattenModeIOXmlEnum.propWriter
ioXmlWriteMeshFlattenEdgesModeNode = flattenModeIOXmlEnum.nodeWriter

ioObjectFactoryRegisterPrimitive( 'MeshFlattenEdgesMode', MeshFlattenEdgesMode )


# Field for MeshFlattenEdgesMode

class MeshFlattenEdgesModeField (Field):
	valueClass = MeshFlattenEdgesMode
	defaultValue = MeshFlattenEdgesMode.NORMAL
	xmlNodeValueReader = ioXmlReadMeshFlattenEdgesModeNode
	xmlNodeValueWriter = ioXmlWriteMeshFlattenEdgesModeNode

registerFieldType( MeshFlattenEdgesMode, MeshFlattenEdgesModeField )


# Variable editor for MeshFlattenEdgesMode

class MeshFlattenEdgesModeVarEditRadioGroup (ChoiceVarEditRadioGroup):
	def __init__(self):
		super( MeshFlattenEdgesModeVarEditRadioGroup, self ).__init__()
		self.x = self.addChoice( MeshFlattenEdgesMode.X, _( 'X-axis' ) )
		self.y = self.addChoice( MeshFlattenEdgesMode.Y, _( 'Y-axis' ) )
		self.z = self.addChoice( MeshFlattenEdgesMode.Z, _( 'Z-axis' ) )
		self.normal = self.addChoice( MeshFlattenEdgesMode.NORMAL, _( 'Normal' ) )
		self.flattenRuns = self.addChoice( MeshFlattenEdgesMode.FLATTEN_RUNS, _( 'Flatten edge runs' ) )
		self.straightenRuns = self.addChoice( MeshFlattenEdgesMode.STRAIGHTEN_RUNS, _( 'Straighten edge runs' ) )


# Column editor for MeshFlattenEdgesMode

class MeshFlattenEdgesModeColumnEditRadioGroup (BasicColumnWidgetEditor):
	def __init__(self, fieldColumn):
		super( MeshFlattenEdgesModeColumnEditRadioGroup, self ).__init__( [ 'x', 'y', 'z', 'normal', 'flattenRuns', 'straightenRuns' ] )
		self._fieldColumn = fieldColumn


	def createElement(self, rowEditContext):
		return MeshFlattenEdgesModeVarEditRadioGroup()

	def initialiseElement(self, element, rowEditContext):
		super( MeshFlattenEdgesModeColumnEditRadioGroup, self ).initialiseElement( element, rowEditContext )
		field = self._p_getVar( self._fieldColumn, rowEditContext )
		element.attachVariable( field )

	def shutdownElement(self, element, rowEditContext):
		super( MeshFlattenEdgesModeColumnEditRadioGroup, self ).shutdownElement( element, rowEditContext )
		element.detachVariable()





class ProcMeshFlattenEdges (GSProcedure):
	description = _( 'Flatten edges' )

	bPerRegion = FieldColumn( bool, True )
	mode = FieldColumn( MeshFlattenEdgesMode, MeshFlattenEdgesMode.NORMAL )


	inputClass = GSProductMesh
	outputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		mode = self.mode
		if mode == MeshFlattenEdgesMode.X:
			inputObject.flattenEdges( Vector3( 1.0, 0.0, 0.0 ), self.bPerRegion )
		elif mode == MeshFlattenEdgesMode.Y:
			inputObject.flattenEdges( Vector3( 0.0, 1.0, 0.0 ), self.bPerRegion )
		elif mode == MeshFlattenEdgesMode.Z:
			inputObject.flattenEdges( Vector3( 0.0, 0.0, 1.0 ), self.bPerRegion )
		elif mode == MeshFlattenEdgesMode.NORMAL:
			inputObject.flattenEdgesNormal( self.bPerRegion )
		elif mode == MeshFlattenEdgesMode.FLATTEN_RUNS:
			inputObject.flattenEdgeRuns()
		elif mode == MeshFlattenEdgesMode.STRAIGHTEN_RUNS:
			inputObject.straightenEdgeRuns()
		return inputObject




class ToolMeshFlattenEdges (GSProcedureTool):
	row = RowRefColumn( ProcMeshFlattenEdges )

	functionClass = ProcMeshFlattenEdges
	pageTitle = _( 'Flatten' )
	titleText = _( 'Flatten edges' )

	bPerRegionCheck = BoolColumnEditCheck( row.bPerRegion, _( 'Per region' ) )
	modeRadio = MeshFlattenEdgesModeColumnEditRadioGroup( row.mode )

	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Flatten mode:' )
	layout.row()  <<  modeRadio.x
	layout.row()  <<  modeRadio.y
	layout.row()  <<  modeRadio.z
	layout.row()  <<  modeRadio.normal
	layout.row()  <<  modeRadio.flattenRuns
	layout.row()  <<  modeRadio.straightenRuns
	layout.row()
	layout.row()  <<  bPerRegionCheck.check
	layout.row()
	layout.finalise()
