##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from gSculpt.FileIO.IOXml import IOXmlEnum, ioObjectFactoryRegisterPrimitive

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.VarEdit.VarEdit import ChoiceVarEditRadioGroup

from gSculpt.Kernel.DB import *
from gSculpt.Kernel.DBEdit import BasicColumnWidgetEditor
from gSculpt.Kernel.ColumnEditors import *
from gSculpt.Kernel.Enum import Enum

from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool.GSFunctionTool import *
from gSculpt.ProceduralTool.GSProcedureTool import GSProcedureTool

from gSculpt.Model.Model import MarkPredicate
from gSculpt.Model.MarkPredicateColumnEditor import MarkPredicateColumnEditRadioGroup

from gSculpt.Mesh.Mesh import GSProductMesh



# MeshMarkEdgesTarget enumeration
class MeshMarkEdgesTarget (Enum):
	ALL = 0
	VERTEXMARKED = 1
	WHOLLYVERTEXMARKED = 2
	FACEMARKED = 3
	FACEMARKBOUNDARY = 4
	FACEMARKINTERNAL = 5
	BOUNDARY = 6
	INTERNAL = 7




# XML IO for MeshMarkEdgesTarget

markTargetIOXmlEnum = IOXmlEnum( MeshMarkEdgesTarget, 'MeshMarkEdgesTarget', MeshMarkEdgesTarget.ALL )
markTargetIOXmlEnum.value( MeshMarkEdgesTarget.ALL, 'all' )
markTargetIOXmlEnum.value( MeshMarkEdgesTarget.VERTEXMARKED, 'vertex_marked' )
markTargetIOXmlEnum.value( MeshMarkEdgesTarget.WHOLLYVERTEXMARKED, 'wholly_vertex_marked' )
markTargetIOXmlEnum.value( MeshMarkEdgesTarget.FACEMARKED, 'face_marked' )
markTargetIOXmlEnum.value( MeshMarkEdgesTarget.FACEMARKBOUNDARY, 'face_mark_boundary' )
markTargetIOXmlEnum.value( MeshMarkEdgesTarget.FACEMARKINTERNAL, 'face_mark_internal' )
markTargetIOXmlEnum.value( MeshMarkEdgesTarget.BOUNDARY, 'boundary' )
markTargetIOXmlEnum.value( MeshMarkEdgesTarget.INTERNAL, 'internal' )

ioXmlReadMeshMarkEdgesTargetProp = markTargetIOXmlEnum.propReader
ioXmlReadMeshMarkEdgesTargetNode = markTargetIOXmlEnum.nodeReader
ioXmlWriteMeshMarkEdgesTargetProp = markTargetIOXmlEnum.propWriter
ioXmlWriteMeshMarkEdgesTargetNode = markTargetIOXmlEnum.nodeWriter

ioObjectFactoryRegisterPrimitive( 'MeshMarkEdgesTarget', MeshMarkEdgesTarget )


# Field for MeshMarkEdgesTarget

class MeshMarkEdgesTargetField (Field):
	valueClass = MeshMarkEdgesTarget
	defaultValue = MeshMarkEdgesTarget.ALL
	xmlNodeValueReader = ioXmlReadMeshMarkEdgesTargetNode
	xmlNodeValueWriter = ioXmlWriteMeshMarkEdgesTargetNode

registerFieldType( MeshMarkEdgesTarget, MeshMarkEdgesTargetField )


# Variable editor for MeshMarkEdgesTarget

class MeshMarkEdgesTargetVarEditRadioGroup (ChoiceVarEditRadioGroup):
	def __init__(self):
		super( MeshMarkEdgesTargetVarEditRadioGroup, self ).__init__()
		self.all = self.addChoice( MeshMarkEdgesTarget.ALL, _( 'All' ) )
		self.vertexMarked = self.addChoice( MeshMarkEdgesTarget.VERTEXMARKED, _( 'With selected vertices' ) )
		self.whollyVertexMarked = self.addChoice( MeshMarkEdgesTarget.WHOLLYVERTEXMARKED, _( 'With all vertices selected' ) )
		self.faceMarked = self.addChoice( MeshMarkEdgesTarget.FACEMARKED, _( 'With selected faces' ) )
		self.faceMarkBoundary = self.addChoice( MeshMarkEdgesTarget.FACEMARKBOUNDARY, _( 'Face selection boundary' ) )
		self.faceMarkInternal = self.addChoice( MeshMarkEdgesTarget.FACEMARKINTERNAL, _( 'Both faces selected' ) )
		self.boundary = self.addChoice( MeshMarkEdgesTarget.BOUNDARY, _( 'Boundary' ) )
		self.internal = self.addChoice( MeshMarkEdgesTarget.INTERNAL, _( 'Internal' ) )


# Column editor for MeshMarkEdgesTarget

class MeshMarkEdgesTargetColumnEditRadioGroup (BasicColumnWidgetEditor):
	def __init__(self, fieldColumn):
		super( MeshMarkEdgesTargetColumnEditRadioGroup, self ).__init__( [ 'all', 'vertexMarked', 'whollyVertexMarked', 'faceMarked', 'faceMarkBoundary', 'faceMarkInternal', 'boundary', 'internal' ] )
		self._fieldColumn = fieldColumn


	def createElement(self, rowEditContext):
		return MeshMarkEdgesTargetVarEditRadioGroup()

	def initialiseElement(self, element, rowEditContext):
		super( MeshMarkEdgesTargetColumnEditRadioGroup, self ).initialiseElement( element, rowEditContext )
		field = self._p_getVar( self._fieldColumn, rowEditContext )
		element.attachVariable( field )

	def shutdownElement(self, element, rowEditContext):
		super( MeshMarkEdgesTargetColumnEditRadioGroup, self ).shutdownElement( element, rowEditContext )
		element.detachVariable()





class ProcMeshMarkEdges (GSProcedure):
	predicate = FieldColumn( MarkPredicate, MarkPredicate.MARK )
	target = FieldColumn( MeshMarkEdgesTarget, MeshMarkEdgesTarget.ALL )


	def getDescription(self):
		if self.predicate == MarkPredicate.MARK:
			return _( 'Select edges' )
		elif self.predicate == MarkPredicate.UNMARK:
			return _( 'Unselect edges' )
		elif self.predicate == MarkPredicate.ADD:
			return _( 'Select edges' )
		elif self.predicate == MarkPredicate.FILTER:
			return _( 'Filter edge selection' )
		elif self.predicate == MarkPredicate.INVERT:
			return _( 'Invert edge selection' )


	inputClass = GSProductMesh
	outputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		target = self.target
		if target == MeshMarkEdgesTarget.ALL:
			inputObject.markEdges_all( self.predicate )
		elif target == MeshMarkEdgesTarget.VERTEXMARKED:
			inputObject.markEdges_vertexMarked( self.predicate )
		elif target == MeshMarkEdgesTarget.WHOLLYVERTEXMARKED:
			inputObject.markEdges_whollyVertexMarked( self.predicate )
		elif target == MeshMarkEdgesTarget.FACEMARKED:
			inputObject.markEdges_faceMarked( self.predicate )
		elif target == MeshMarkEdgesTarget.FACEMARKBOUNDARY:
			inputObject.markEdges_faceMarkBoundary( self.predicate )
		elif target == MeshMarkEdgesTarget.FACEMARKINTERNAL:
			inputObject.markEdges_faceMarkInternal( self.predicate )
		elif target == MeshMarkEdgesTarget.BOUNDARY:
			inputObject.markEdges_boundary( self.predicate )
		elif target == MeshMarkEdgesTarget.INTERNAL:
			inputObject.markEdges_internal( self.predicate )
		return inputObject



class ToolMeshMarkEdges (GSProcedureTool):
	row = RowRefColumn( ProcMeshMarkEdges )

	functionClass = ProcMeshMarkEdges
	pageTitle = _( 'Select' )
	titleText = _( 'Select edges' )

	predicateRadio = MarkPredicateColumnEditRadioGroup( row.predicate )
	targetRadio = MeshMarkEdgesTargetColumnEditRadioGroup( row.target )

	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Select mode:' )
	layout.row()  <<  predicateRadio.mark
	layout.row()  <<  predicateRadio.unmark
	layout.row()  <<  predicateRadio.add
	layout.row()  <<  predicateRadio.filter
	layout.row()  <<  predicateRadio.invert
	layout.row()
	layout.row()  <<  _( 'Target:' )
	layout.row()  <<  targetRadio.all
	layout.row()  <<  targetRadio.vertexMarked
	layout.row()  <<  targetRadio.whollyVertexMarked
	layout.row()  <<  targetRadio.faceMarked
	layout.row()  <<  targetRadio.faceMarkBoundary
	layout.row()  <<  targetRadio.faceMarkInternal
	layout.row()  <<  targetRadio.boundary
	layout.row()  <<  targetRadio.internal
	layout.row()
	layout.finalise()

