##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import deepcopy

from gSculpt.Math.Math import Vector3

from gSculpt.FileIO.IOXml import IOXmlEnum, ioObjectFactoryRegisterPrimitive

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.VarEdit.VarEdit import ChoiceVarEditRadioGroup

from gSculpt.Kernel.DB import *
from gSculpt.Kernel.ColumnEditors import *

from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool.GSFunctionTool import *
from gSculpt.ProceduralTool.GSProcedureTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *

from gSculpt.Manipulator.RealManipulator import RealManipulator
from gSculpt.Manipulator.ColumnManipulators import *

from gSculpt.Mesh import Mesh
from gSculpt.Mesh.Mesh import MeshRiftEdgesAdjuster, MeshRiftEdgesStyle, GSProductMesh




# XML IO for MeshRiftEdgesStyle

riftStyleIOXmlEnum = IOXmlEnum( MeshRiftEdgesStyle, 'MeshRiftEdgesStyle', MeshRiftEdgesStyle.CENTRE )
riftStyleIOXmlEnum.value( MeshRiftEdgesStyle.A, 'a' )
riftStyleIOXmlEnum.value( MeshRiftEdgesStyle.B, 'b' )
riftStyleIOXmlEnum.value( MeshRiftEdgesStyle.CENTRE, 'centre' )

ioXmlReadMeshRiftEdgesStyleProp = riftStyleIOXmlEnum.propReader
ioXmlReadMeshRiftEdgesStyleNode = riftStyleIOXmlEnum.nodeReader
ioXmlWriteMeshRiftEdgesStyleProp = riftStyleIOXmlEnum.propWriter
ioXmlWriteMeshRiftEdgesStyleNode = riftStyleIOXmlEnum.nodeWriter

ioObjectFactoryRegisterPrimitive( 'MeshRiftEdgesStyle', MeshRiftEdgesStyle )


# Field for MeshRiftEdgesStyle

class MeshRiftEdgesStyleField (Field):
	valueClass = MeshRiftEdgesStyle
	defaultValue = MeshRiftEdgesStyle.CENTRE
	xmlNodeValueReader = ioXmlReadMeshRiftEdgesStyleNode
	xmlNodeValueWriter = ioXmlWriteMeshRiftEdgesStyleNode

registerFieldType( MeshRiftEdgesStyle, MeshRiftEdgesStyleField )


# Variable editor for MeshRiftEdgesStyle

class MeshRiftEdgesStyleVarEditRadioGroup (ChoiceVarEditRadioGroup):
	def __init__(self):
		super( MeshRiftEdgesStyleVarEditRadioGroup, self ).__init__()
		self.a = self.addChoice( MeshRiftEdgesStyle.A, 'A' )
		self.b = self.addChoice( MeshRiftEdgesStyle.B, 'B' )
		self.centre = self.addChoice( MeshRiftEdgesStyle.CENTRE, _( 'Centre' ) )


# Column editor for MeshRiftEdgesStyle

class MeshRiftEdgesStyleColumnEditRadioGroup (BasicColumnWidgetEditor):
	def __init__(self, fieldColumn):
		super( MeshRiftEdgesStyleColumnEditRadioGroup, self ).__init__( [ 'a', 'b', 'centre' ] )
		self._fieldColumn = fieldColumn


	def createElement(self, rowEditContext):
		return MeshRiftEdgesStyleVarEditRadioGroup()

	def initialiseElement(self, element, rowEditContext):
		super( MeshRiftEdgesStyleColumnEditRadioGroup, self ).initialiseElement( element, rowEditContext )
		field = self._p_getVar( self._fieldColumn, rowEditContext )
		element.attachVariable( field )

	def shutdownElement(self, element, rowEditContext):
		super( MeshRiftEdgesStyleColumnEditRadioGroup, self ).shutdownElement( element, rowEditContext )
		element.detachVariable()







class ProcMeshRiftEdges (GSProcedure):
	description = _( 'Rift edges' )

	rift = FieldColumn( Vector3, Vector3() )
	segments = FieldColumn( int, 1 )
	style = FieldColumn( MeshRiftEdgesStyle, MeshRiftEdgesStyle.CENTRE )
	bFill = FieldColumn( bool, True )
	bPinchAtMarkedVertices = FieldColumn( bool, False )

	inputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		adjuster = MeshRiftEdgesAdjuster()
		inputObject.riftMarkedEdges( adjuster, self.style, self.segments, self.bFill, self.bPinchAtMarkedVertices )
		adjuster.setParameters( self.rift )
		return inputObject



class RiftEdgesAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcMeshRiftEdges

	def main(self):
		self.mesh = deepcopy( self._inputObject )
		self.adjuster = MeshRiftEdgesAdjuster()
		self.mesh.riftMarkedEdges( self.adjuster, self._proc.style, self._proc.segments, self._proc.bFill, self._proc.bPinchAtMarkedVertices )

	def adjust(self):
		self.adjuster.setParameters( self._proc.rift )

	def getResult(self):
		return self.mesh




class ToolMeshRiftEdges (GSProcedureTool):
	row = RowRefColumn( ProcMeshRiftEdges )

	functionClass = ProcMeshRiftEdges
	adjustableInvoker = RiftEdgesAdjustableInvoker
	pageTitle = _( 'Rift' )
	titleText = _( 'Rift Edges' )


	def _p_focalPoint(self):
		return self.inputProduct.getMarkedEdgesCentre()

	focalPoint = FunctionColumn( _p_focalPoint )

	riftEntry = Vector3ColumnEditEntry( row.rift, -1.0e8, 1.0e8, 0.1, 10.0, 4 )
	segmentsEntry = IntColumnEditEntry( row.segments, 1, 1048576, 1, 10 )
	styleRadio = MeshRiftEdgesStyleColumnEditRadioGroup( row.style )
	bFillCheck = BoolColumnEditCheck( row.bFill, _( 'Fill' ) )
	bPinchCheck = BoolColumnEditCheck( row.bPinchAtMarkedVertices, _( 'Pinch at marked vertices' ) )

	riftManip = Translation3ColumnManipulator3d( row.rift, focalPoint, snapping=ColumnManipulatorSnappingDistance() )
	riftManip.enable()

	numSegmentsManip = IntColumnScrollManipulator( row.segments, 1, 1048576 )
	numSegmentsManip.enable()

	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Use the scroll wheel to' )
	layout.row()  <<  _( 'change the number of segments' )
	layout.row()
	layout.row()  <<  _( 'Rift:' )
	layout.row()  <<  'X:'  <<  riftEntry.x
	layout.row()  <<  'Y:'  <<  riftEntry.y
	layout.row()  <<  'Z:'  <<  riftEntry.z
	layout.row()
	layout.row()  <<  _( 'Segments:' )  <<  segmentsEntry.entry
	layout.row()
	layout.row()  <<  _( 'Style:' )
	layout.row()  <<  styleRadio.a  <<  styleRadio.b  <<  styleRadio.centre
	layout.row()
	layout.row()  <<  bFillCheck.check
	layout.row()  <<  bPinchCheck.check
	layout.row()
	layout.finalise()
