##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from OpenGL.GL import *

from copy import deepcopy

from gSculpt.Math.Math import Colour3f, Point3, Segment3

from gSculpt.Kernel.DB import *

from gSculpt.View.View import PaintLayer

from gSculpt.WorkArea.Viewport3d import Viewport3d
from gSculpt.Graphics.Graphics import glColor3Colour3f, glVertex3Point3

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.ProceduralCore.GSFunction import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *
from gSculpt.Kernel.ColumnEditors import *

from gSculpt.Mesh.Mesh import GSProductMesh
from gSculpt.Mesh.MeshPickHelper import MeshPickHelper

from gSculpt.Model import ModelDraw
from gSculpt.Model.Pick3d import Pick3d
from gSculpt.Model.Model import pyMPickList_to_MPickList, MPickList
from gSculpt.Model.TargetList import TargetList
from gSculpt.Model.TargetListColumnEditor import TargetListColumnEditor

from gSculpt.Manipulator.Manipulator import Manipulator




class ProcMeshSplitEdges (GSProcedure):
	targetList = FieldColumn( TargetList, TargetList() )

	description = _( 'Split edges' )
	inputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		mpicks = MPickList()
		pyMPickList_to_MPickList( [ pick.toMPick()  for  pick in self.targetList ], mpicks )
		inputObject.splitEdges( mpicks )
		return inputObject




class MeshSplitEdgesManipulator (Manipulator, MeshPickHelper):
	def __init__(self):
		super( MeshSplitEdgesManipulator, self ).__init__()
		self._var = None

		self._highlightedEdgeSegment = None
		self._splitPoint = None

		self._o_setUISensitivity( False )




	def manipAttachToView(self, view):
		super( MeshSplitEdgesManipulator, self ).manipAttachToView( view )
		self.attachView( view )
		self._view.addEventHandler( self, Viewport3d )		# HACK: need to handle other viewport types
		self._view.addPainter( self, Viewport3d )
		self._view.postRedraw()

	def manipDetachFromView(self):
		self._view.removeEventHandler( self, Viewport3d )
		self._view.removePainter( self, Viewport3d )
		self._view.postRedraw()
		self.detachView()
		super( MeshSplitEdgesManipulator, self ).manipDetachFromView()


	def attachVariable(self, var):
		assert self._var is None, 'already attached to a variable'
		self._var = var
		self._o_watchVarValidity( self._var )
		self._o_refreshSensitivity()

	def detachVariable(self):
		assert self._var is not None, 'not attached to a variable'
		self._o_stopWatchingVarValidity( self._var )
		self._var = None
		self._o_refreshSensitivity()



	def _o_setUISensitivity(self, bSensitive):
		self._o_setMeshPickHelperSensitivity( bSensitive )

	def _o_computeSensitivity(self):
		if self._var is not None:
			return self._var.isValid()
		else:
			return False



	def _p_refreshHit(self, viewport, mesh, pick):
		pickedEdgeIndex, self._splitPoint, pickedFaceIndex = mesh.pickEdge( pick.toMPick() )

		if pickedEdgeIndex != -1:
			self._highlightedEdgeSegment = mesh.getEdgeSegment( pickedEdgeIndex )
			return True
		else:
			return False



	def _p_paintHighlight(self, viewport, paintLayer):
		# Highlight the picked edge
		ModelDraw.drawEdgePickHighlight( self._highlightedEdgeSegment, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )

		# Draw the split position on the seed edge
		ModelDraw.drawSplitPosition( self._splitPoint )


	def _p_pickConfirmed(self, viewport, pick):
		"Protected - call to commit a pick"
		assert self._var is not None, 'no variable attached'
		self._var.value = self._var.value + [ deepcopy( pick ) ]
		self._p_commandHistoryBreak()


	def _o_resetHighlight(self):
		self._highlightedEdgeSegment = None
		self._splitPoint = None


	def _p_onParamsChanged(self):
		self._o_resetHighlight()
		self._o_requestHighlightRefresh()





class MeshSplitEdgesTargetListColumnManipulator (GSProcedureTool.GSProcedureToolColumnManipulator):
	def __init__(self, fieldColumn):
		super( MeshSplitEdgesTargetListColumnManipulator, self ).__init__()
		self._fieldColumn = fieldColumn


	def createElement(self, rowEditContext):
		manip = MeshSplitEdgesManipulator()
		rowEditContext.rowEditor.targetListManip = manip
		return manip

	def destroyElement(self, element, rowEditContext):
		del rowEditContext.rowEditor.targetListManip

	def initialiseElement(self, element, rowEditContext):
		field = self._p_getVar( self._fieldColumn, rowEditContext )
		element.attachVariable( field )
		if rowEditContext.commandHistory is not None:
			element.commandHistory = rowEditContext.commandHistory

	def shutdownElement(self, element, rowEditContext):
		element.detachVariable()
		element.commandHistory = None


	def installElementEventHandlers(self, element, rowEditContext):
		element.manipAttachToView( rowEditContext.view )

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.manipDetachFromView()


	def attachOutputProductSource(self, element, source):
		element.attachMeshSource( source )

	def detachOutputProductSource(self, element, source):
		element.detachMeshSource()




class MeshSplitEdgesAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcMeshSplitEdges

	def __init__(self, *args, **kwargs):
		super( MeshSplitEdgesAdjustableInvoker, self ).__init__( *args, **kwargs )


	def main(self):
		self.mesh = deepcopy( self._inputObject )
		mpicks = MPickList()
		pyMPickList_to_MPickList( [ pick.toMPick()  for  pick in self._proc.targetList ], mpicks )
		self.mesh.splitEdges( mpicks )

	def getResult(self):
		return self.mesh




class ToolMeshSplitEdges (GSProcedureTool.GSProcedureTool):
	row = RowRefColumn( ProcMeshSplitEdges )

	functionClass = ProcMeshSplitEdges
	adjustableInvoker = MeshSplitEdgesAdjustableInvoker
	pageTitle = _( 'Split' )
	titleText = _( 'Split edges' )


	targetListEditor = TargetListColumnEditor( row.targetList )

	targetListManip = MeshSplitEdgesTargetListColumnManipulator( row.targetList )
	targetListManip.enable()


	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Click to insert vertices into edges' )
	layout.row()
	layout.row()  <<  targetListEditor.numTargets
	layout.row()  <<  None  <<  ( targetListEditor.removeTarget, 2 )  <<  None
	layout.row()
	layout.finalise()



