##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from OpenGL.GL import *

from copy import deepcopy

from gSculpt.Math.Math import Colour3f, Point3, Segment3

from gSculpt.FileIO.IOXml import ioObjectFactoryRegister, ioXmlReadBoolNode, ioXmlWriteBoolNode

from gSculpt.Kernel.Variable import BoolVar
from gSculpt.Kernel.DB import *

from gSculpt.View.View import PaintLayer

from gSculpt.WorkArea.Viewport3d import Viewport3d
from gSculpt.Graphics.Graphics import glColor3Colour3f, glVertex3Point3

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.ProceduralCore.GSFunction import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *
from gSculpt.Kernel.ColumnEditors import *

from gSculpt.Mesh.Mesh import GSProductMesh, pyMeshBandsawTargetList_to_CppMeshBandsawTargetList, CppMeshBandsawTarget, CppMeshBandsawTargetList
from gSculpt.Mesh.MeshPickHelper import MeshPickHelper

from gSculpt.Model import ModelDraw
from gSculpt.Model.Pick3d import Pick3d
from gSculpt.Model.TargetList import TargetList
from gSculpt.Model.TargetListColumnEditor import TargetListColumnEditor

from gSculpt.Manipulator.Manipulator import Manipulator




class MeshBandsawTarget (object):
	def __init__(self, pick = None, bStopAtMarkedEdges = False):
		if pick is None:
			pick = Pick3d()
		self.pick = pick
		self.bStopAtMarkedEdges = bStopAtMarkedEdges


	def toCppMeshBandsawTarget(self):
		return CppMeshBandsawTarget( self.pick.toMPick(), self.bStopAtMarkedEdges )


	def __readxml__(self, xmlNode):
		xmlNode.getChild( 'pick' )  >>  self.pick
		self.bStopAtMarkedEdges = ioXmlReadBoolNode( xmlNode.getChild( 'b_stop_at_marked_edges' ), self.bStopAtMarkedEdges )

	def __writexml__(self, xmlNode):
		xmlNode.addChild( 'pick' )  <<  self.pick
		ioXmlWriteBoolNode( xmlNode.addChild( 'b_stop_at_marked_edges' ), self.bStopAtMarkedEdges )


ioObjectFactoryRegister( 'MeshBandsawTarget', MeshBandsawTarget )





class ProcMeshBandsaw (GSProcedure):
	targetList = FieldColumn( TargetList, TargetList() )

	description = _( 'Bandsaw' )
	inputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		cppTargets = CppMeshBandsawTargetList()
		pyMeshBandsawTargetList_to_CppMeshBandsawTargetList( [ target.toCppMeshBandsawTarget()  for  target in self.targetList ], cppTargets )
		inputObject.bandsaw( cppTargets )
		return inputObject




class MeshBandsawManipulator (Manipulator, MeshPickHelper):
	def __init__(self):
		super( MeshBandsawManipulator, self ).__init__()
		self._targetsVar = None
		self._bStopAtMarkedEdgesVar = None

		self._seedEdgeIndex = -1
		self._seedFaceIndex = -1
		self._seedEdgeSegment = None
		self._seedPickPoint = None

		self._bandsawSegments = []
		self._bandsawPoints = []
		self._bClosed = False

		self._o_setUISensitivity( False )




	def manipAttachToView(self, view):
		super( MeshBandsawManipulator, self ).manipAttachToView( view )
		self.attachView( view )
		self._view.addEventHandler( self, Viewport3d )		# HACK: need to handle other viewport types
		self._view.addPainter( self, Viewport3d )
		self._view.postRedraw()

	def manipDetachFromView(self):
		self._view.removeEventHandler( self, Viewport3d )
		self._view.removePainter( self, Viewport3d )
		self._view.postRedraw()
		self.detachView()
		super( MeshBandsawManipulator, self ).manipDetachFromView()


	def attachVariables(self, targets, bStopAtMarkedEdges):
		assert self._targetsVar is None, 'already attached to a variable'
		self._targetsVar = targets
		self._bStopAtMarkedEdgesVar = bStopAtMarkedEdges
		self._bStopAtMarkedEdgesVar.changedSignal.connect( self._p_onParamsChanged )
		self._o_watchVarValidity( self._targetsVar )
		self._o_refreshSensitivity()

	def detachVariables(self):
		assert self._targetsVar is not None, 'not attached to a variable'
		self._o_stopWatchingVarValidity( self._targetsVar )
		self._bStopAtMarkedEdgesVar.changedSignal.disconnect( self._p_onParamsChanged )
		self._targetsVar = None
		self._bStopAtMarkedEdgesVar = None
		self._o_refreshSensitivity()



	def _o_setUISensitivity(self, bSensitive):
		self._o_setMeshPickHelperSensitivity( bSensitive )

	def _o_computeSensitivity(self):
		if self._targetsVar is not None:
			return self._targetsVar.isValid()
		else:
			return False



	def _p_refreshHit(self, viewport, mesh, pick):
		pickedEdgeIndex, self._seedPickPoint, pickedFaceIndex = mesh.pickEdge( pick.toMPick() )

		if pickedEdgeIndex != self._seedEdgeIndex  or  pickedFaceIndex != self._seedFaceIndex:
			# The seed has changed
			self._seedEdgeIndex, self._seedFaceIndex = pickedEdgeIndex, pickedFaceIndex

			if pickedEdgeIndex != -1  and  pickedFaceIndex != -1:
				self._seedEdgeSegment = mesh.getEdgeSegment( pickedEdgeIndex )

				# An edge was chosed; discover the bandsaw edge ring
				self._bandsawSegments, self._bClosed = mesh.bandsawGetRingSegments( pickedEdgeIndex, pickedFaceIndex, self._bStopAtMarkedEdgesVar.value )

		if pickedEdgeIndex != -1  and  pickedFaceIndex != -1:
			point, t = self._seedEdgeSegment.closestPointTo( self._seedPickPoint )
			self._bandsawPoints = [ segment.getPoint( t )  for segment in self._bandsawSegments ]
			return True
		else:
			self._bandsawSegments = []
			self._bandsawPoints = []
			return False



	def _p_paintHighlight(self, viewport, paintLayer):
		# Highlight the seed edge
		ModelDraw.drawEdgePickHighlight( self._seedEdgeSegment, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )

		# Draw the split position on the seed edge
		ModelDraw.drawSplitPosition( self._seedPickPoint )

		# Draw the bandsaw line
		if paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY  or  paintLayer == PaintLayer.PAINTLAYER_WIREFRAME:
			colour = ModelDraw.MODIFICATION1_HIGHLIGHT_COLOUR

			if paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY:
				colour = ModelDraw.computeOverlayColour( colour )

			glColor3Colour3f( colour )

			if self._bClosed:
				glBegin( GL_LINE_LOOP )
			else:
				glBegin( GL_LINE_STRIP )

			for point in self._bandsawPoints:
				glVertex3Point3( point )

			glEnd()


	def _p_pickConfirmed(self, viewport, pick):
		"Protected - call to commit a pick"
		assert self._targetsVar is not None, 'no variable attached'
		target = MeshBandsawTarget( deepcopy( pick ), self._bStopAtMarkedEdgesVar.value )
		self._targetsVar.value = self._targetsVar.value + [ target ]
		self._p_commandHistoryBreak()


	def _o_resetHighlight(self):
		self._seedEdgeIndex = -1
		self._seedFaceIndex = -1
		self._seedEdgeSegment = None
		self._seedPickPoint = None

		self._bandsawSegments = []
		self._bandsawPoints = []
		self._bClosed = False


	def _p_onParamsChanged(self):
		self._o_resetHighlight()
		self._o_requestHighlightRefresh()





class MeshBandsawTargetListColumnManipulator (GSProcedureTool.GSProcedureToolColumnManipulator):
	def __init__(self, targetsColumn, bStopAtMarkedEdgesColumn):
		super( MeshBandsawTargetListColumnManipulator, self ).__init__()
		self._targetsColumn = targetsColumn
		self._bStopAtMarkedEdgesColumn = bStopAtMarkedEdgesColumn


	def createElement(self, rowEditContext):
		return MeshBandsawManipulator()

	def destroyElement(self, element, rowEditContext):
		del rowEditContext.rowEditor.targetListManip

	def initialiseElement(self, element, rowEditContext):
		targets = self._p_getVar( self._targetsColumn, rowEditContext )
		bStopAtMarkedEdges = self._p_getVar( self._bStopAtMarkedEdgesColumn, rowEditContext )
		element.attachVariables( targets, bStopAtMarkedEdges )
		if rowEditContext.commandHistory is not None:
			element.commandHistory = rowEditContext.commandHistory

	def shutdownElement(self, element, rowEditContext):
		element.detachVariables()
		element.commandHistory = None


	def installElementEventHandlers(self, element, rowEditContext):
		element.manipAttachToView( rowEditContext.view )

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.manipDetachFromView()


	def attachOutputProductSource(self, element, source):
		element.attachMeshSource( source )

	def detachOutputProductSource(self, element, source):
		element.detachMeshSource()




class MeshBandsawAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcMeshBandsaw

	def __init__(self, *args, **kwargs):
		super( MeshBandsawAdjustableInvoker, self ).__init__( *args, **kwargs )


	def main(self):
		self.mesh = deepcopy( self._inputObject )
		cppTargets = CppMeshBandsawTargetList()
		pyMeshBandsawTargetList_to_CppMeshBandsawTargetList( [ target.toCppMeshBandsawTarget()  for  target in self._proc.targetList ], cppTargets )
		self.mesh.bandsaw( cppTargets )

	def getResult(self):
		return self.mesh




class ToolMeshBandsaw (GSProcedureTool.GSProcedureTool):
	row = RowRefColumn( ProcMeshBandsaw )

	functionClass = ProcMeshBandsaw
	adjustableInvoker = MeshBandsawAdjustableInvoker
	pageTitle = _( 'Bandsaw' )
	titleText = _( 'Bandsaw' )


	bStopAtMarkedEdges = FieldColumn( bool, False )


	targetListEditor = TargetListColumnEditor( row.targetList )

	targetListManip = MeshBandsawTargetListColumnManipulator( row.targetList, bStopAtMarkedEdges )
	targetListManip.enable()

	bStopAtMarkedEdgesCheck = BoolColumnEditCheck( bStopAtMarkedEdges, _( 'Stop at marked edges' ) )


	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Click to place bandsaw cuts' )
	layout.row()
	layout.row()  <<  bStopAtMarkedEdgesCheck.check
	layout.row()
	layout.row()  <<  targetListEditor.numTargets
	layout.row()  <<  None  <<  ( targetListEditor.removeTarget, 2 )  <<  None
	layout.row()
	layout.finalise()



