##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import deepcopy

from gSculpt.Math.Math import Colour3f, Point3, Segment3

from gSculpt.View.View import PaintLayer

from gSculpt.WorkArea.Viewport3d import Viewport3d

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.Kernel.DB import *

from gSculpt.ProceduralCore.GSFunction import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *

from gSculpt.Model import ModelDraw
from gSculpt.Mesh.Mesh import GSProductMesh, pyMeshCutTargetList_to_CppMeshCutTargetList, CppMeshCutTargetList, MeshCutTargetType, MeshCutState
from gSculpt.Mesh.MeshPickHelper import MeshPickHelper

from gSculpt.Model.TargetList import TargetList
from gSculpt.Model.TargetListColumnEditor import TargetListColumnEditor

from MeshCutTarget import MeshCutTarget

from gSculpt.Manipulator.Manipulator import Manipulator




class ProcMeshCut (GSProcedure):
	targetList = FieldColumn( TargetList, TargetList() )

	description = _( 'Cut' )
	inputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		cppTargets = CppMeshCutTargetList()
		pyMeshCutTargetList_to_CppMeshCutTargetList( [ target.toCppMeshCutTarget()  for  target in self.targetList ], cppTargets )
		inputObject.cut( cppTargets, None )
		return inputObject




class MeshCutManipulator (Manipulator, MeshPickHelper):
	def __init__(self):
		super( MeshCutManipulator, self ).__init__()
		self._var = None
		self.cutState = None
		self._targetType = MeshCutTargetType.EDGE
		self._bTargetValid = False

		self._vertexPosition = None
		self._edgeSegment = None
		self._edgeSplitPosition = None
		self._faceCurrentPoint = None
		self._facePoints = []
		self._faceBValid = False

		self._o_setUISensitivity( False )



	def manipAttachToView(self, view):
		super( MeshCutManipulator, self ).manipAttachToView( view )
		self.attachView( view )
		self._view.addEventHandler( self, Viewport3d )		# HACK: need to handle other viewport types
		self._view.addPainter( self, Viewport3d )
		self._view.postRedraw()

	def manipDetachFromView(self):
		self._view.removeEventHandler( self, Viewport3d )
		self._view.removePainter( self, Viewport3d )
		self._view.postRedraw()
		self.detachView()
		super( MeshCutManipulator, self ).manipDetachFromView()


	def attachVariable(self, var):
		assert self._var is None, 'already attached to a variable'
		self._var = var
		self._var.changedSignal.connect( self._p_onVarChanged )
		self._o_watchVarValidity( self._var )
		self._o_refreshSensitivity()

	def detachVariable(self):
		assert self._var is not None, 'not attached to a variable'
		self._var.changedSignal.disconnect( self._p_onVarChanged )
		self._o_stopWatchingVarValidity( self._var )
		self._var = None
		self._o_refreshSensitivity()



	def _o_setUISensitivity(self, bSensitive):
		self._o_setMeshPickHelperSensitivity( bSensitive )

	def _o_computeSensitivity(self):
		if self._var is not None:
			return self._var.isValid()
		else:
			return False



	def _p_refreshPick(self, viewport):
		super( MeshCutManipulator, self )._p_refreshPick( viewport )
		pstat = viewport.pointerStatus
		if pstat.bShiftKey:
			# target vertices
			self._targetType = MeshCutTargetType.VERTEX
		elif pstat.bControlKey:
			# target faces
			self._targetType = MeshCutTargetType.FACE
		else:
			# target edges
			self._targetType = MeshCutTargetType.EDGE


	def _p_refreshHit(self, viewport, mesh, pick):
		self._bTargetValid = False

		target = MeshCutTarget( deepcopy( pick ), self._targetType )

		if self._targetType == MeshCutTargetType.VERTEX:
			# Target vertices
			vertexIndex, thruFaceIndex = mesh.pickVertex( pick.toMPick() )
			if vertexIndex != -1:
				self._vertexPosition = mesh.getVertexPosition( vertexIndex )
				self._bTargetValid = mesh.isValidCutTarget( target.toCppMeshCutTarget(), self.cutState )
				return True
			else:
				return False
		elif self._targetType == MeshCutTargetType.FACE:
			# Target faces
			faceIndex, intersectionPoint = mesh.pickFace( pick.toMPick() )
			if faceIndex != -1:
				self._faceCurrentPoint = intersectionPoint
				self._bTargetValid = mesh.isValidCutTarget( target.toCppMeshCutTarget(), self.cutState )
				return True
			else:
				return False
		elif self._targetType == MeshCutTargetType.EDGE:
			# Target edges
			edgeIndex, closestPoint, thruFaceIndex = mesh.pickEdge( pick.toMPick() )
			if edgeIndex != -1:
				self._edgeSegment = mesh.getEdgeSegment( edgeIndex )
				self._edgeSplitPosition = closestPoint
				self._bTargetValid = mesh.isValidCutTarget( target.toCppMeshCutTarget(), self.cutState )
				return True
			else:
				return False
		else:
			raise ValueError, 'invalid cut target type'



	def _p_paintHighlight(self, viewport, paintLayer):
		if self._targetType == MeshCutTargetType.VERTEX:
			if self._bTargetValid:
				ModelDraw.drawVertexPickHighlight( self._vertexPosition, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer );
			else:
				ModelDraw.drawVertexPickHighlight( self._vertexPosition, ModelDraw.INVALID_TARGET_HIGHLIGHT_COLOUR, paintLayer );

			if self._p_hasPreviousCutPoint()  and  self._bTargetValid:
				ModelDraw.drawSegment( Segment3( self._p_getPreviousCutPoint(), self._vertexPosition ), ModelDraw.MODIFICATION1_HIGHLIGHT_COLOUR, paintLayer );
		elif self._targetType == MeshCutTargetType.EDGE:
			if self._bTargetValid:
				ModelDraw.drawEdgePickHighlight( self._edgeSegment, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer );
			else:
				ModelDraw.drawEdgePickHighlight( self._edgeSegment, ModelDraw.INVALID_TARGET_HIGHLIGHT_COLOUR, paintLayer );

			if self._p_hasPreviousCutPoint()  and  self._bTargetValid:
				ModelDraw.drawSegment( Segment3( self._p_getPreviousCutPoint(), self._edgeSplitPosition ), ModelDraw.MODIFICATION1_HIGHLIGHT_COLOUR, paintLayer );
		elif self._targetType == MeshCutTargetType.FACE:
			if self._p_hasPreviousCutPoint()  and  self._bTargetValid:
				ModelDraw.drawSegment( Segment3( self._p_getPreviousCutPoint(), self._faceCurrentPoint ), ModelDraw.MODIFICATION1_HIGHLIGHT_COLOUR, paintLayer );

			if self._bTargetValid:
				ModelDraw.drawCircleHighlight( self._faceCurrentPoint, ModelDraw.MODIFICATION1_HIGHLIGHT_COLOUR, paintLayer );
			else:
				ModelDraw.drawCircleHighlight( self._faceCurrentPoint, ModelDraw.INVALID_TARGET_HIGHLIGHT_COLOUR, paintLayer );

		if paintLayer == PaintLayer.PAINTLAYER_OVERLAY_NODEPTH  and  self._targetType == MeshCutTargetType.EDGE  and  self._bTargetValid:
			ModelDraw.drawSplitPosition( self._edgeSplitPosition );


	def _p_pickConfirmed(self, viewport, pick):
		"Protected - call to commit a pick"
		assert self._var is not None, 'no variable attached'
		if self._bTargetValid:
			target = MeshCutTarget( deepcopy( pick ), self._targetType )
			self._var.value = self._var.value + [ target ]
			self._p_commandHistoryBreak()


	def evPaint3d(self, viewport, paintLayer):
		super( MeshCutManipulator, self ).evPaint3d( viewport, paintLayer )

		if paintLayer == PaintLayer.PAINTLAYER_WIREFRAME  or  paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY:
			if self.cutState.lastSplitVertexIndex != -1:
				lastSplitVertexPosition = self.cutState.lastSplitVertexPosition
				ModelDraw.drawVertexReticle( lastSplitVertexPosition, paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY )

				if len( self._facePoints ) > 0:
					ModelDraw.drawSegment( Segment3( lastSplitVertexPosition, self._facePoints[0] ), ModelDraw.MODIFICATION1_HIGHLIGHT_COLOUR, paintLayer )

			ModelDraw.drawLineStrip( self._facePoints, ModelDraw.MODIFICATION1_HIGHLIGHT_COLOUR, paintLayer )

			for point in self._facePoints:
				ModelDraw.drawCircleHighlight( point, ModelDraw.MODIFICATION1_HIGHLIGHT_COLOUR, paintLayer )


	def _p_hasPreviousCutPoint(self):
		return self.cutState.lastSplitVertexIndex != -1  and  self._faceBValid


	def _p_getPreviousCutPoint(self):
		if self._faceBValid  and  len( self._facePoints ) > 0:
			return self._facePoints[-1]
		else:
			return self.cutState.lastSplitVertexPosition


	def _p_onVarChanged(self):
		self._facePoints = []
		self._faceBValid = True

		targetList = self._var.value

		lastVertexOrEdgeTargetIndex = -1
		for i, target in reversed( list( enumerate( targetList ) ) ):
			if target.targetType == MeshCutTargetType.VERTEX  or  target.targetType == MeshCutTargetType.EDGE:
				lastVertexOrEdgeTargetIndex = i
				break

		if lastVertexOrEdgeTargetIndex != -1:
			mesh = self._p_getMesh()

			if mesh is not None:
				for i in xrange( lastVertexOrEdgeTargetIndex + 1, len( targetList ) ):
					faceIndex, pointOnFace = mesh.pickFace( targetList[i].pick.toMPick() )

					if faceIndex != -1:
						self._facePoints.append( pointOnFace )
						self._faceBValid = True
					else:
						self._facePoints = []
						self._faceBValid = False
						break





class MeshCutTargetListColumnManipulator (GSProcedureTool.GSProcedureToolColumnManipulator):
	def __init__(self, fieldColumn):
		super( MeshCutTargetListColumnManipulator, self ).__init__()
		self._fieldColumn = fieldColumn


	def createElement(self, rowEditContext):
		return MeshCutManipulator()

	def initialiseElement(self, element, rowEditContext):
		field = self._p_getVar( self._fieldColumn, rowEditContext )
		element.attachVariable( field )
		element.cutState = rowEditContext.rowEditor._adjustableInvoker.cutState
		if rowEditContext.commandHistory is not None:
			element.commandHistory = rowEditContext.commandHistory

	def shutdownElement(self, element, rowEditContext):
		element.detachVariable()
		element.commandHistory = None
		element.cutState = None


	def installElementEventHandlers(self, element, rowEditContext):
		element.manipAttachToView( rowEditContext.view )

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.manipDetachFromView()


	def attachOutputProductSource(self, element, source):
		element.attachMeshSource( source )

	def detachOutputProductSource(self, element, source):
		element.detachMeshSource()




class MeshCutAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcMeshCut

	def __init__(self, *args, **kwargs):
		super( MeshCutAdjustableInvoker, self ).__init__( *args, **kwargs )
		self.cutState = MeshCutState()


	def main(self):
		self.mesh = deepcopy( self._inputObject )
		cppTargets = CppMeshCutTargetList()
		pyMeshCutTargetList_to_CppMeshCutTargetList( [ target.toCppMeshCutTarget()  for  target in self._proc.targetList ], cppTargets )
		self.mesh.cut( cppTargets, self.cutState )

	def getResult(self):
		return self.mesh




class ToolMeshCut (GSProcedureTool.GSProcedureTool):
	row = RowRefColumn( ProcMeshCut )

	functionClass = ProcMeshCut
	adjustableInvoker = MeshCutAdjustableInvoker
	pageTitle = _( 'Cut' )
	titleText = _( 'Cut' )


	targetListEditor = TargetListColumnEditor( row.targetList )

	targetListManip = MeshCutTargetListColumnManipulator( row.targetList )
	targetListManip.enable()


	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Cut shapes into a mesh' )
	layout.row()
	layout.row()  <<  _( 'Click to add points to the' )
	layout.row()  <<  _( 'cut path.' )
	layout.row()  <<  _( 'Edges will be targeted, or:' )
	layout.row()  <<  _( 'Hold <shift> to target vertices' )
	layout.row()  <<  _( 'Hold <control> to place' )
	layout.row()  <<  _( 'points in faces.' )
	layout.row()
	layout.row()  <<  _( 'Note that points in faces' )
	layout.row()  <<  _( 'must be preceeded and' )
	layout.row()  <<  _( 'followed by vertex or edge' )
	layout.row()  <<  _( 'points' )
	layout.row()
	layout.row()  <<  targetListEditor.numTargets
	layout.row()  <<  None  <<  ( targetListEditor.removeTarget, 2 )  <<  None
	layout.row()
	layout.finalise()



