##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from OpenGL.GL import *

from copy import deepcopy

from gSculpt.Math.Math import Colour3f, Point3, Segment3

from gSculpt.FileIO.IOXml import *

from gSculpt.Kernel.Variable import BoolVar, IntVar
from gSculpt.Kernel.DB import *

from gSculpt.View.View import PaintLayer

from gSculpt.WorkArea.Viewport3d import Viewport3d
from gSculpt.Graphics.Graphics import glColor3Colour3f, glVertex3Point3

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.ProceduralCore.GSFunction import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *
from gSculpt.Kernel.ColumnEditors import *

from gSculpt.Mesh.Mesh import GSProductMesh, pyMeshNBandsawTargetList_to_CppMeshNBandsawTargetList, CppMeshNBandsawTarget, CppMeshNBandsawTargetList
from gSculpt.Mesh.MeshPickHelper import MeshPickHelper

from gSculpt.Model import ModelDraw
from gSculpt.Model.Pick3d import Pick3d
from gSculpt.Model.TargetList import TargetList
from gSculpt.Model.TargetListColumnEditor import TargetListColumnEditor

from gSculpt.Manipulator.ColumnManipulators import IntColumnScrollManipulator
from gSculpt.Manipulator.Manipulator import Manipulator




class MeshNBandsawTarget (object):
	def __init__(self, pick = None, bStopAtMarkedEdges = False, numCuts = 1):
		if pick is None:
			pick = Pick3d()
		self.pick = pick
		self.bStopAtMarkedEdges = bStopAtMarkedEdges
		self.numCuts = numCuts


	def toCppMeshNBandsawTarget(self):
		return CppMeshNBandsawTarget( self.pick.toMPick(), self.bStopAtMarkedEdges, self.numCuts )


	def __readxml__(self, xmlNode):
		xmlNode.getChild( 'pick' )  >>  self.pick
		self.bStopAtMarkedEdges = ioXmlReadBoolNode( xmlNode.getChild( 'b_stop_at_marked_edges' ), self.bStopAtMarkedEdges )
		self.numCuts = ioXmlReadIntNode( xmlNode.getChild( 'num_cuts' ), self.numCuts )

	def __writexml__(self, xmlNode):
		xmlNode.addChild( 'pick' )  <<  self.pick
		ioXmlWriteBoolNode( xmlNode.addChild( 'b_stop_at_marked_edges' ), self.bStopAtMarkedEdges )
		ioXmlWriteIntNode( xmlNode.addChild( 'num_cuts' ), self.numCuts )


ioObjectFactoryRegister( 'MeshNBandsawTarget', MeshNBandsawTarget )





class ProcMeshNBandsaw (GSProcedure):
	targetList = FieldColumn( TargetList, TargetList() )

	description = _( 'N-Bandsaw' )
	inputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		cppTargets = CppMeshNBandsawTargetList()
		pyMeshNBandsawTargetList_to_CppMeshNBandsawTargetList( [ target.toCppMeshNBandsawTarget()  for  target in self.targetList ], cppTargets )
		inputObject.nBandsaw( cppTargets )
		return inputObject




class MeshNBandsawManipulator (Manipulator, MeshPickHelper):
	def __init__(self):
		super( MeshNBandsawManipulator, self ).__init__()
		self._targetsVar = None
		self._bStopAtMarkedEdgesVar = None
		self._numCutsVar = None

		self._seedEdgeIndex = -1
		self._seedFaceIndex = -1
		self._seedEdgeSegment = None
		self._seedPickPoint = None

		self._bandsawSegments = []
		self._bandsawPoints = []
		self._bClosed = False

		self._o_setUISensitivity( False )




	def manipAttachToView(self, view):
		super( MeshNBandsawManipulator, self ).manipAttachToView( view )
		self.attachView( view )
		self._view.addEventHandler( self, Viewport3d )		# HACK: need to handle other viewport types
		self._view.addPainter( self, Viewport3d )
		self._view.postRedraw()

	def manipDetachFromView(self):
		self._view.removeEventHandler( self, Viewport3d )
		self._view.removePainter( self, Viewport3d )
		self._view.postRedraw()
		self.detachView()
		super( MeshNBandsawManipulator, self ).manipDetachFromView()


	def attachVariables(self, targets, bStopAtMarkedEdges, numCuts):
		assert self._targetsVar is None, 'already attached to a variable'
		self._targetsVar = targets
		self._bStopAtMarkedEdgesVar = bStopAtMarkedEdges
		self._numCutsVar = numCuts
		self._o_watchVarValidity( self._targetsVar )
		self._bStopAtMarkedEdgesVar.changedSignal.connect( self._p_onParamsChanged )
		self._numCutsVar.changedSignal.connect( self._p_onParamsChanged )
		self._o_refreshSensitivity()

	def detachVariables(self):
		assert self._targetsVar is not None, 'not attached to a variable'
		self._o_stopWatchingVarValidity( self._targetsVar )
		self._bStopAtMarkedEdgesVar.changedSignal.disconnect( self._p_onParamsChanged )
		self._numCutsVar.changedSignal.disconnect( self._p_onParamsChanged )
		self._targetsVar = None
		self._bStopAtMarkedEdgesVar = None
		self._numCutsVar = None
		self._o_refreshSensitivity()



	def _o_setUISensitivity(self, bSensitive):
		self._o_setMeshPickHelperSensitivity( bSensitive )

	def _o_computeSensitivity(self):
		if self._targetsVar is not None:
			return self._targetsVar.isValid()
		else:
			return False



	def _p_refreshHit(self, viewport, mesh, pick):
		pickedEdgeIndex, self._seedPickPoint, pickedFaceIndex = mesh.pickEdge( pick.toMPick() )

		if pickedEdgeIndex != self._seedEdgeIndex  or  pickedFaceIndex != self._seedFaceIndex:
			# The seed has changed
			self._seedEdgeIndex, self._seedFaceIndex = pickedEdgeIndex, pickedFaceIndex

			if pickedEdgeIndex != -1  and  pickedFaceIndex != -1:
				self._seedEdgeSegment = mesh.getEdgeSegment( pickedEdgeIndex )

				# An edge was chosed; discover the bandsaw edge ring
				self._bandsawSegments, self._bClosed = mesh.bandsawGetRingSegments( pickedEdgeIndex, pickedFaceIndex, self._bStopAtMarkedEdgesVar.value )

		if pickedEdgeIndex != -1  and  pickedFaceIndex != -1:
			numCuts = self._numCutsVar.value
			numSegments = numCuts + 1
			deltaT = 1.0  /  float( numSegments )
			splitFractions = [ deltaT * i  for i in xrange( 1, numSegments ) ]
			self._bandsawPoints = [ [ segment.getPoint( t )  for segment in self._bandsawSegments ]    for t in splitFractions ]
			return True
		else:
			self._bandsawSegments = []
			self._bandsawPoints = []
			return False



	def _p_paintHighlight(self, viewport, paintLayer):
		# Highlight the seed edge
		ModelDraw.drawEdgePickHighlight( self._seedEdgeSegment, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )

		# Draw the bandsaw lines
		if paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY  or  paintLayer == PaintLayer.PAINTLAYER_WIREFRAME:
			colour = ModelDraw.MODIFICATION1_HIGHLIGHT_COLOUR

			if paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY:
				colour = ModelDraw.computeOverlayColour( colour )

			glColor3Colour3f( colour )

			for pointList in self._bandsawPoints:
				if self._bClosed:
					glBegin( GL_LINE_LOOP )
				else:
					glBegin( GL_LINE_STRIP )

				for point in pointList:
					glVertex3Point3( point )

				glEnd()


	def _p_pickConfirmed(self, viewport, pick):
		"Protected - call to commit a pick"
		assert self._targetsVar is not None, 'no variable attached'
		target = MeshNBandsawTarget( deepcopy( pick ), self._bStopAtMarkedEdgesVar.value, self._numCutsVar.value + 1 )
		self._targetsVar.value = self._targetsVar.value + [ target ]
		self._p_commandHistoryBreak()


	def _o_resetHighlight(self):
		self._seedEdgeIndex = -1
		self._seedFaceIndex = -1
		self._seedEdgeSegment = None
		self._seedPickPoint = None

		self._bandsawSegments = []
		self._bandsawPoints = []
		self._bClosed = False


	def _p_onParamsChanged(self):
		self._o_resetHighlight()
		self._o_requestHighlightRefresh()





class MeshNBandsawTargetListColumnManipulator (GSProcedureTool.GSProcedureToolColumnManipulator):
	def __init__(self, targetsColumn, bStopAtMarkedEdgesColumn, numCutsColumn):
		super( MeshNBandsawTargetListColumnManipulator, self ).__init__()
		self._targetsColumn = targetsColumn
		self._bStopAtMarkedEdgesColumn = bStopAtMarkedEdgesColumn
		self._numCutsColumn = numCutsColumn


	def createElement(self, rowEditContext):
		return MeshNBandsawManipulator()

	def destroyElement(self, element, rowEditContext):
		del rowEditContext.rowEditor.targetListManip

	def initialiseElement(self, element, rowEditContext):
		targets = self._p_getVar( self._targetsColumn, rowEditContext )
		bStopAtMarkedEdges = self._p_getVar( self._bStopAtMarkedEdgesColumn, rowEditContext )
		numCuts = self._p_getVar( self._numCutsColumn, rowEditContext )
		element.attachVariables( targets, bStopAtMarkedEdges, numCuts )
		if rowEditContext.commandHistory is not None:
			element.commandHistory = rowEditContext.commandHistory

	def shutdownElement(self, element, rowEditContext):
		element.detachVariables()
		element.commandHistory = None


	def installElementEventHandlers(self, element, rowEditContext):
		element.manipAttachToView( rowEditContext.view )

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.manipDetachFromView()


	def attachOutputProductSource(self, element, source):
		element.attachMeshSource( source )

	def detachOutputProductSource(self, element, source):
		element.detachMeshSource()




class MeshNBandsawAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcMeshNBandsaw

	def __init__(self, *args, **kwargs):
		super( MeshNBandsawAdjustableInvoker, self ).__init__( *args, **kwargs )


	def main(self):
		self.mesh = deepcopy( self._inputObject )
		cppTargets = CppMeshNBandsawTargetList()
		pyMeshNBandsawTargetList_to_CppMeshNBandsawTargetList( [ target.toCppMeshNBandsawTarget()  for  target in self._proc.targetList ], cppTargets )
		self.mesh.nBandsaw( cppTargets )

	def getResult(self):
		return self.mesh




class ToolMeshNBandsaw (GSProcedureTool.GSProcedureTool):
	row = RowRefColumn( ProcMeshNBandsaw )

	functionClass = ProcMeshNBandsaw
	adjustableInvoker = MeshNBandsawAdjustableInvoker
	pageTitle = _( 'N-Bandsaw' )
	titleText = _( 'N-Bandsaw' )


	bStopAtMarkedEdges = FieldColumn( bool, False )
	numCuts = FieldColumn( int, 1 )


	targetListEditor = TargetListColumnEditor( row.targetList )

	targetListManip = MeshNBandsawTargetListColumnManipulator( row.targetList, bStopAtMarkedEdges, numCuts )
	targetListManip.enable()

	bStopAtMarkedEdgesCheck = BoolColumnEditCheck( bStopAtMarkedEdges, _( 'Stop at marked edges' ) )
	numCutsEntry = IntColumnEditEntry( numCuts, 1, 1048576, 1, 10 )

	numCutsManip = IntColumnScrollManipulator( numCuts, 1, 1048576 )
	numCutsManip.enable()


	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Click to place bandsaw cuts' )
	layout.row()
	layout.row()  <<  _( 'Use the scroll wheel to' )
	layout.row()  <<  _( 'change the number of cuts' )
	layout.row()
	layout.row()  <<  bStopAtMarkedEdgesCheck.check
	layout.row()  <<  _( 'Num. cuts:' )  <<  numCutsEntry.entry
	layout.row()
	layout.row()  <<  targetListEditor.numTargets
	layout.row()  <<  None  <<  ( targetListEditor.removeTarget, 2 )  <<  None
	layout.row()
	layout.finalise()



