##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import deepcopy

from gSculpt.Math.Math import Colour3f, Point3, Segment3

from gSculpt.View.View import PaintLayer

from gSculpt.WorkArea.Viewport3d import Viewport3d

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.Kernel.DB import *
from gSculpt.Kernel.ColumnEditors import IntColumnEditEntry

from gSculpt.ProceduralCore.GSFunction import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *

from gSculpt.Model import ModelDraw
from gSculpt.Mesh.Mesh import GSProductMesh, pyMeshCutTargetList_to_CppMeshCutTargetList, CppMeshCutTargetList, MeshCutTargetType, MeshNCutState
from gSculpt.Mesh.MeshPickHelper import MeshPickHelper

from gSculpt.Model.TargetList import TargetList
from gSculpt.Model.TargetListColumnEditor import TargetListColumnEditor

from MeshCutTarget import MeshCutTarget

from gSculpt.Manipulator.ColumnManipulators import IntColumnScrollManipulator
from gSculpt.Manipulator.Manipulator import Manipulator





class ProcMeshNCut (GSProcedure):
	numCuts = FieldColumn( int, 1 )
	targetList = FieldColumn( TargetList, TargetList() )

	description = _( 'N-Cut' )
	inputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		cppTargets = CppMeshCutTargetList()
		pyMeshCutTargetList_to_CppMeshCutTargetList( [ target.toCppMeshCutTarget()  for  target in self.targetList ], cppTargets )
		inputObject.nCut( self.numCuts + 1, cppTargets, None )
		return inputObject




class MeshNCutManipulator (Manipulator, MeshPickHelper):
	def __init__(self):
		super( MeshNCutManipulator, self ).__init__()
		self._var = None
		self.cutState = None
		self._targetType = MeshCutTargetType.EDGE
		self._bTargetValid = False

		self._vertexPosition = None
		self._edgeSegment = None
		self._edgeSplitPosition = None
		self._faceCurrentPoint = None
		self._facePoints = []
		self._faceBValid = False

		self._o_setUISensitivity( False )



	def manipAttachToView(self, view):
		super( MeshNCutManipulator, self ).manipAttachToView( view )
		self.attachView( view )
		self._view.addEventHandler( self, Viewport3d )		# HACK: need to handle other viewport types
		self._view.addPainter( self, Viewport3d )
		self._view.postRedraw()

	def manipDetachFromView(self):
		self._view.removeEventHandler( self, Viewport3d )
		self._view.removePainter( self, Viewport3d )
		self._view.postRedraw()
		self.detachView()
		super( MeshNCutManipulator, self ).manipDetachFromView()


	def attachVariable(self, var):
		assert self._var is None, 'already attached to a variable'
		self._var = var
		self._o_watchVarValidity( self._var )
		self._o_refreshSensitivity()

	def detachVariable(self):
		assert self._var is not None, 'not attached to a variable'
		self._o_stopWatchingVarValidity( self._var )
		self._var = None
		self._o_refreshSensitivity()



	def _o_setUISensitivity(self, bSensitive):
		self._o_setMeshPickHelperSensitivity( bSensitive )

	def _o_computeSensitivity(self):
		if self._var is not None:
			return self._var.isValid()
		else:
			return False



	def _p_refreshPick(self, viewport):
		super( MeshNCutManipulator, self )._p_refreshPick( viewport )
		pstat = viewport.pointerStatus
		if pstat.bShiftKey:
			# target vertices
			self._targetType = MeshCutTargetType.VERTEX
		else:
			# target edges
			self._targetType = MeshCutTargetType.EDGE


	def _p_refreshHit(self, viewport, mesh, pick):
		self._bTargetValid = False

		target = MeshCutTarget( deepcopy( pick ), self._targetType )

		if self._targetType == MeshCutTargetType.VERTEX:
			# Target vertices
			vertexIndex, thruFaceIndex = mesh.pickVertex( pick.toMPick() )
			if vertexIndex != -1:
				self._vertexPosition = mesh.getVertexPosition( vertexIndex )
				self._bTargetValid = mesh.isValidNCutTarget( target.toCppMeshCutTarget(), self.cutState )
				return True
			else:
				return False
		elif self._targetType == MeshCutTargetType.EDGE:
			# Target edges
			edgeIndex, closestPoint, thruFaceIndex = mesh.pickEdge( pick.toMPick() )
			if edgeIndex != -1:
				self._edgeSegment = mesh.getEdgeSegment( edgeIndex )
				self._edgeSplitPosition = closestPoint
				self._bTargetValid = mesh.isValidNCutTarget( target.toCppMeshCutTarget(), self.cutState )
				return True
			else:
				return False
		else:
			raise ValueError, 'invalid cut target type'



	def _p_paintHighlight(self, viewport, paintLayer):
		if self._targetType == MeshCutTargetType.VERTEX:
			if self._bTargetValid:
				ModelDraw.drawVertexPickHighlight( self._vertexPosition, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer );
			else:
				ModelDraw.drawVertexPickHighlight( self._vertexPosition, ModelDraw.INVALID_TARGET_HIGHLIGHT_COLOUR, paintLayer );
		elif self._targetType == MeshCutTargetType.EDGE:
			if self._bTargetValid:
				ModelDraw.drawEdgePickHighlight( self._edgeSegment, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer );
			else:
				ModelDraw.drawEdgePickHighlight( self._edgeSegment, ModelDraw.INVALID_TARGET_HIGHLIGHT_COLOUR, paintLayer );


	def _p_pickConfirmed(self, viewport, pick):
		"Protected - call to commit a pick"
		assert self._var is not None, 'no variable attached'
		if self._bTargetValid:
			target = MeshCutTarget( deepcopy( pick ), self._targetType )
			self._var.value = self._var.value + [ target ]
			self._p_commandHistoryBreak()


	def evPaint3d(self, viewport, paintLayer):
		super( MeshNCutManipulator, self ).evPaint3d( viewport, paintLayer )

		if paintLayer == PaintLayer.PAINTLAYER_WIREFRAME  or  paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY:
			for position in self.cutState.lastSplitVertexPositions():
				ModelDraw.drawVertexReticle( position, paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY )







class MeshNCutTargetListColumnManipulator (GSProcedureTool.GSProcedureToolColumnManipulator):
	def __init__(self, fieldColumn):
		super( MeshNCutTargetListColumnManipulator, self ).__init__()
		self._fieldColumn = fieldColumn


	def createElement(self, rowEditContext):
		return MeshNCutManipulator()

	def initialiseElement(self, element, rowEditContext):
		field = self._p_getVar( self._fieldColumn, rowEditContext )
		element.attachVariable( field )
		element.cutState = rowEditContext.rowEditor._adjustableInvoker.cutState
		if rowEditContext.commandHistory is not None:
			element.commandHistory = rowEditContext.commandHistory

	def shutdownElement(self, element, rowEditContext):
		element.detachVariable()
		element.commandHistory = None
		element.cutState = None


	def installElementEventHandlers(self, element, rowEditContext):
		element.manipAttachToView( rowEditContext.view )

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.manipDetachFromView()


	def attachOutputProductSource(self, element, source):
		element.attachMeshSource( source )

	def detachOutputProductSource(self, element, source):
		element.detachMeshSource()




class MeshNCutAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcMeshNCut

	def __init__(self, *args, **kwargs):
		super( MeshNCutAdjustableInvoker, self ).__init__( *args, **kwargs )
		self.cutState = MeshNCutState()


	def main(self):
		self.mesh = deepcopy( self._inputObject )
		cppTargets = CppMeshCutTargetList()
		pyMeshCutTargetList_to_CppMeshCutTargetList( [ target.toCppMeshCutTarget()  for  target in self._proc.targetList ], cppTargets )
		self.mesh.nCut( self._proc.numCuts + 1, cppTargets, self.cutState )

	def getResult(self):
		return self.mesh




class ToolMeshNCut (GSProcedureTool.GSProcedureTool):
	row = RowRefColumn( ProcMeshNCut )

	functionClass = ProcMeshNCut
	adjustableInvoker = MeshNCutAdjustableInvoker
	pageTitle = _( 'N-Cut' )
	titleText = _( 'N-Cut' )


	numCutsEntry = IntColumnEditEntry( row.numCuts, 1, 1048576, 1, 10 )
	targetListEditor = TargetListColumnEditor( row.targetList )

	targetListManip = MeshNCutTargetListColumnManipulator( row.targetList )
	targetListManip.enable()

	numCutsManip = IntColumnScrollManipulator( row.numCuts, 1, 1048576 )
	numCutsManip.enable()


	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Cut a path into a mesh' )
	layout.row()
	layout.row()  <<  _( 'Click to add points to the' )
	layout.row()  <<  _( 'cut path.' )
	layout.row()  <<  _( 'Edges will be targeted, or:' )
	layout.row()  <<  _( 'Hold <shift> to target vertices' )
	layout.row()
	layout.row()  <<  _( 'Use the scroll wheel to' )
	layout.row()  <<  _( 'change the number of cuts' )
	layout.row()
	layout.row()  <<  _( 'Num. cuts:' )  <<  numCutsEntry.entry
	layout.row()
	layout.row()  <<  targetListEditor.numTargets
	layout.row()  <<  None  <<  ( targetListEditor.removeTarget, 2 )  <<  None
	layout.row()
	layout.finalise()



