##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import deepcopy

from gSculpt.Math.Math import Colour3f, Point3, Segment3

from gSculpt.Util.CoEventHandler import CoEventHandler

from gSculpt.FileIO.IOXml import IOXmlEnum, ioObjectFactoryRegisterPrimitive, ioObjectFactoryRegister

from gSculpt.View.View import PaintLayer

from gSculpt.WorkArea.Viewport import ViewportContextButtonListener
from gSculpt.WorkArea.Viewport3d import Viewport3d, Viewport3dPainter

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.Kernel.DB import *

from gSculpt.ProceduralCore.GSFunction import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *

from gSculpt.Model import ModelDraw
from gSculpt.Mesh.Mesh import GSProductMesh, pyMeshWeldTargetList_to_CppMeshWeldTargetList, CppMeshWeldTarget, CppMeshWeldTargetList, MeshTargetWeldVerticesTargetType
from gSculpt.Mesh.MeshManipulatorHelper import MeshManipulatorHelper

from gSculpt.Model.Pick3d import Pick3d
from gSculpt.Model.TargetList import TargetList
from gSculpt.Model.TargetListColumnEditor import TargetListColumnEditor

from gSculpt.Manipulator.Manipulator import Manipulator






weldTargetTypeIOXmlEnum = IOXmlEnum( MeshTargetWeldVerticesTargetType, 'MeshTargetWeldVerticesTargetType', MeshTargetWeldVerticesTargetType.EDGE )
weldTargetTypeIOXmlEnum.value( MeshTargetWeldVerticesTargetType.VERTEX, 'vertex' )
weldTargetTypeIOXmlEnum.value( MeshTargetWeldVerticesTargetType.EDGE_CLOSEST_POINT_TO_SOURCE, 'edge_closest_point_to_source' )
weldTargetTypeIOXmlEnum.value( MeshTargetWeldVerticesTargetType.EDGE, 'edge' )

ioXmlReadMeshWeldTargetTypeProp = weldTargetTypeIOXmlEnum.propReader
ioXmlReadMeshWeldTargetTypeNode = weldTargetTypeIOXmlEnum.nodeReader
ioXmlWriteMeshWeldTargetTypeProp = weldTargetTypeIOXmlEnum.propWriter
ioXmlWriteMeshWeldTargetTypeNode = weldTargetTypeIOXmlEnum.nodeWriter

ioObjectFactoryRegisterPrimitive( 'MeshTargetWeldVerticesTargetType', MeshTargetWeldVerticesTargetType )





class MeshWeldVerticesTarget (object):
	def __init__(self, sourcePick = None, targetPick = None, targetType = MeshTargetWeldVerticesTargetType.VERTEX):
		if sourcePick is None:
			sourcePick = Pick3d()
		if targetPick is None:
			targetPick = Pick3d()
		self.sourcePick = sourcePick
		self.targetPick = targetPick
		self.targetType = targetType


	def toCppMeshWeldTarget(self):
		return CppMeshWeldTarget( self.sourcePick.toMPick(), self.targetPick.toMPick(), self.targetType )


	def __readxml__(self, xmlNode):
		xmlNode.getChild( 'source' )  >>  self.sourcePick
		xmlNode.getChild( 'target' )  >>  self.targetPick
		self.targetType = ioXmlReadMeshWeldTargetTypeNode( xmlNode.getChild( 'target_type' ), self.targetType )

	def __writexml__(self, xmlNode):
		xmlNode.addChild( 'source' )  <<  self.sourcePick
		xmlNode.addChild( 'target' )  <<  self.targetPick
		ioXmlWriteMeshWeldTargetTypeNode( xmlNode.addChild( 'target_type' ), self.targetType )


ioObjectFactoryRegister( 'MeshWeldVerticesTarget', MeshWeldVerticesTarget )







class ProcMeshTargetWeldVertices (GSProcedure):
	targetList = FieldColumn( TargetList, TargetList() )

	description = _( 'Target weld vertices' )
	inputClass = GSProductMesh

	def procedureInvoke(self, inputObject, errorHandler):
		cppTargets = CppMeshWeldTargetList()
		pyMeshWeldTargetList_to_CppMeshWeldTargetList( [ target.toCppMeshWeldTarget()  for  target in self.targetList ], cppTargets )
		inputObject.targetWeld( cppTargets )
		return inputObject






class MeshTargetWeldVerticesManipulator (Manipulator, MeshManipulatorHelper, ViewportContextButtonListener, Viewport3dPainter):
	EVENT_BUTTONDOWN = 0
	EVENT_BUTTONUP = 1
	EVENT_MOTION = 2
	EVENT_LEAVE = 3
	EVENT_CONTEXT = 4
	EVENT_RESET = 5


	class Event (object):
		def __init__(self, eventType, **kwargs):
			self.eventType = eventType
			self.__dict__.update( kwargs )

	class ContextException:
		pass

	class ResetException:
		pass




	def __init__(self):
		super( MeshTargetWeldVerticesManipulator, self ).__init__()
		self._var = None

		self._coEventHandler = None
		self._storedEvents = []

		self._vertexHighlightPosition = None
		self._edgeHighlightSegment = None
		self._edgeHighlightSplitPosition = None
		self._sourceVertexHighlightPosition = None
		self._bTargetValid = True

		self._o_setUISensitivity( False )




	def manipAttachToView(self, view):
		super( MeshTargetWeldVerticesManipulator, self ).manipAttachToView( view )
		self._coEventHandler = CoEventHandler( self._p_targetWeldLoop )
		self._coEventHandler.start()
		self.attachView( view )
		self._view.addEventHandler( self, Viewport3d )		# HACK: need to handle other viewport types
		self._view.addPainter( self, Viewport3d )
		self._view.postRedraw()

	def manipDetachFromView(self):
		self._view.removeEventHandler( self, Viewport3d )
		# Very important that we remove the context button listener here, as if the manipulator is removed before completing
		# the editing cycle (by undo/redo for example), then the context listener can remain, blocking other listeners with a lower priority
		self._view.removeContextButtonListener( self, Viewport3d )
		self._view.removePainter( self, Viewport3d )
		self._view.postRedraw()
		self.detachView()
		self._coEventHandler.terminate()
		self._coEventHandler = None
		super( MeshTargetWeldVerticesManipulator, self ).manipDetachFromView()


	def attachVariable(self, var):
		assert self._var is None, 'already attached to a variable'
		self._var = var
		self._var.changedSignal.connect( self._p_onVarChanged )
		self._o_watchVarValidity( self._var )
		self._o_refreshSensitivity()

	def detachVariable(self):
		assert self._var is not None, 'not attached to a variable'
		self._var.changedSignal.disconnect( self._p_onVarChanged )
		self._o_stopWatchingVarValidity( self._var )
		self._var = None
		self._o_refreshSensitivity()



	def _o_setUISensitivity(self, bSensitive):
		self._o_setMeshManipulatorHelperSensitivity( bSensitive )

	def _o_computeSensitivity(self):
		if self._var is not None:
			return self._var.isValid()
		else:
			return False




	def _p_getSourcePick(self, event, sourceVertexIndex):
		mesh = self._o_getMesh()
		sourcePick = event.viewport.createPick()
		vertexIndex, thruFaceIndex = mesh.pickVertex( sourcePick.toMPick() )

		if vertexIndex != sourceVertexIndex:
			if vertexIndex != -1:
				self._vertexHighlightPosition = mesh.getVertexPosition( vertexIndex )
			else:
				self._vertexHighlightPosition = None
			self._view.postRedraw()
			sourceVertexIndex = vertexIndex

		return sourcePick, sourceVertexIndex




	def _p_getTargetPick(self, event, sourceVertexIndex, targetVertexIndex, targetEdgeIndex, targetType):
		mesh = self._o_getMesh()
		targetPick = event.viewport.createPick()

		if event.viewport.pointerStatus.bShiftKey:
			targetVertexIndex = -1
			self._vertexHighlightPosition = None

			# Target is the closest point on the picked edge
			edgeIndex, closestPoint, thruFaceIndex = mesh.pickEdge( targetPick.toMPick() )

			if edgeIndex != targetEdgeIndex  or  targetType != MeshTargetWeldVerticesTargetType.EDGE_CLOSEST_POINT_TO_SOURCE:
				targetType = MeshTargetWeldVerticesTargetType.EDGE_CLOSEST_POINT_TO_SOURCE
				if edgeIndex != -1:
					self._edgeHighlightSegment = mesh.getEdgeSegment( edgeIndex )
					self._edgeHighlightSplitPosition, t = self._edgeHighlightSegment.closestPointTo( self._sourceVertexHighlightPosition )
					self._bTargetValid = mesh.isValidWeldTargetEdge( sourceVertexIndex, edgeIndex, targetType )
				else:
					self._edgeHighlightSegment = None
					self._edgeHighlightSplitPosition = None
					self._bTargetValid = True
				self._view.postRedraw()
				targetEdgeIndex = edgeIndex

		elif event.viewport.pointerStatus.bControlKey:
			targetVertexIndex = -1
			self._vertexHighlightPosition = None

			# Target is the chosen point on the picked edge
			edgeIndex, closestPoint, thruFaceIndex = mesh.pickEdge( targetPick.toMPick() )

			targetType = MeshTargetWeldVerticesTargetType.EDGE
			if edgeIndex != -1:
				self._edgeHighlightSegment = mesh.getEdgeSegment( edgeIndex )
				self._edgeHighlightSplitPosition = closestPoint
				self._bTargetValid = mesh.isValidWeldTargetEdge( sourceVertexIndex, edgeIndex, targetType )
			else:
				self._edgeHighlightSegment = None
				self._edgeHighlightSplitPosition = None
				self._bTargetValid = True
			self._view.postRedraw()
			targetEdgeIndex = edgeIndex

		else:
			targetEdgeIndex = -1
			self._edgeHighlightSegment = None
			self._edgeHighlightSplitPosition = None

			# Target is the picked vertex
			vertexIndex, thruFaceIndex = mesh.pickVertex( targetPick.toMPick() )

			if vertexIndex != targetVertexIndex  or  targetType != MeshTargetWeldVerticesTargetType.VERTEX:
				targetType = MeshTargetWeldVerticesTargetType.VERTEX
				if vertexIndex != -1:
					self._vertexHighlightPosition = mesh.getVertexPosition( vertexIndex )
					self._bTargetValid = mesh.isValidWeldTargetVertex( sourceVertexIndex, vertexIndex )
				else:
					self._vertexHighlightPosition = None
					self._bTargetValid = True
				self._view.postRedraw()
				targetVertexIndex = vertexIndex

		return targetPick, targetVertexIndex, targetEdgeIndex, targetType



	def _p_getWeldTargetFromUser(self):
		self._vertexHighlightPosition = None
		self._edgeHighlightSegment = None
		self._edgeHighlightSplitPosition = None
		self._sourceVertexHighlightPosition = None
		self._bTargetValid = True
		sourceVertexIndex = -1
		targetVertexIndex = -1
		targetEdgeIndex = -1
		targetType = None
		targetPick = None


		event = self._coEventHandler.waitForNextEvent()

		# Get the source vertex
		while not ( event.eventType == MeshTargetWeldVerticesManipulator.EVENT_BUTTONDOWN  and  sourceVertexIndex != -1 ):
			if event.eventType == MeshTargetWeldVerticesManipulator.EVENT_RESET:
				self._view.postRedraw()
				raise MeshTargetWeldVerticesManipulator.ResetException

			if event.eventType != MeshTargetWeldVerticesManipulator.EVENT_LEAVE:
				sourcePick, sourceVertexIndex = self._p_getSourcePick( event, sourceVertexIndex )

			event = self._coEventHandler.waitForNextEvent()


		# We have a source vertex
		self._sourceVertexHighlightPosition = self._vertexHighlightPosition

		# The context button can be pressed by the user to cancel a weld
		self._view.addContextButtonListener( self, Viewport3d )

		# Now, get the target
		try:
			while not ( event.eventType == MeshTargetWeldVerticesManipulator.EVENT_BUTTONUP  and  ( targetVertexIndex != -1  or  targetEdgeIndex != -1 )  and  self._bTargetValid ):
				# Handle a reset event
				if event.eventType == MeshTargetWeldVerticesManipulator.EVENT_RESET:
					self._view.postRedraw()
					raise MeshTargetWeldVerticesManipulator.ResetException

				# Handle a context button event
				if event.eventType == MeshTargetWeldVerticesManipulator.EVENT_CONTEXT:
					self._view.postRedraw()
					raise MeshTargetWeldVerticesManipulator.ContextException

				# Get the target pick
				if event.eventType != MeshTargetWeldVerticesManipulator.EVENT_LEAVE:
					targetPick, targetVertexIndex, targetEdgeIndex, targetType = self._p_getTargetPick( event, sourceVertexIndex, targetVertexIndex, targetEdgeIndex, targetType )

				event = self._coEventHandler.waitForNextEvent()


			self._view.removeContextButtonListener( self, Viewport3d )
			weldTarget = MeshWeldVerticesTarget( deepcopy( sourcePick ), deepcopy( targetPick ), targetType )
			self._var.value = self._var.value + [ weldTarget ]
			self._p_commandHistoryBreak()
		except MeshTargetWeldVerticesManipulator.ContextException:
			self._view.removeContextButtonListener( self, Viewport3d )
			self._view.postRedraw()




	def _p_targetWeldLoop(self):
		while True:
			try:
				self._p_getWeldTargetFromUser()
			except MeshTargetWeldVerticesManipulator.ResetException:
				self._view.postRedraw()





	def evPaint3d(self, viewport, paintLayer):
		super( MeshTargetWeldVerticesManipulator, self ).evPaint3d( viewport, paintLayer )

		if self._vertexHighlightPosition is not None:
			if self._bTargetValid:
				ModelDraw.drawVertexPickHighlight( self._vertexHighlightPosition, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )
			else:
				ModelDraw.drawVertexPickHighlight( self._vertexHighlightPosition, ModelDraw.INVALID_TARGET_HIGHLIGHT_COLOUR, paintLayer )

		if self._edgeHighlightSegment is not None:
			if self._bTargetValid:
				ModelDraw.drawEdgePickHighlight( self._edgeHighlightSegment, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )
			else:
				ModelDraw.drawEdgePickHighlight( self._edgeHighlightSegment, ModelDraw.INVALID_TARGET_HIGHLIGHT_COLOUR, paintLayer )

		if self._edgeHighlightSplitPosition is not None:
			if paintLayer == PaintLayer.PAINTLAYER_OVERLAY_NODEPTH:
				ModelDraw.drawSplitPosition( self._edgeHighlightSplitPosition )

		if self._sourceVertexHighlightPosition is not None:
			if paintLayer == PaintLayer.PAINTLAYER_WIREFRAME  or  paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY:
				ModelDraw.drawVertexReticle( self._sourceVertexHighlightPosition, paintLayer == PaintLayer.PAINTLAYER_TRANSPARENCY )
			if self._bTargetValid:
				if self._vertexHighlightPosition is not None:
					ModelDraw.drawSegment( Segment3( self._sourceVertexHighlightPosition, self._vertexHighlightPosition ), ModelDraw.MODIFICATION1_HIGHLIGHT_COLOUR, paintLayer )
				if self._edgeHighlightSplitPosition is not None:
					ModelDraw.drawSegment( Segment3( self._sourceVertexHighlightPosition, self._edgeHighlightSplitPosition ), ModelDraw.MODIFICATION1_HIGHLIGHT_COLOUR, paintLayer )





	def evPrimaryButtonDown(self, viewport):
		super( MeshTargetWeldVerticesManipulator, self ).evPrimaryButtonDown( viewport )
		if self._bSensitive:
			self._coEventHandler.sendEvent( MeshTargetWeldVerticesManipulator.Event( MeshTargetWeldVerticesManipulator.EVENT_BUTTONDOWN, viewport=viewport ) )

	def evPrimaryButtonUp(self, viewport):
		super( MeshTargetWeldVerticesManipulator, self ).evPrimaryButtonDown( viewport )
		if self._bSensitive:
			self._coEventHandler.sendEvent( MeshTargetWeldVerticesManipulator.Event( MeshTargetWeldVerticesManipulator.EVENT_BUTTONUP, viewport=viewport ) )


	def evMotion(self, viewport, point):
		super( MeshTargetWeldVerticesManipulator, self ).evMotion( viewport, point )
		if self._bSensitive:
			self._coEventHandler.sendEvent( MeshTargetWeldVerticesManipulator.Event( MeshTargetWeldVerticesManipulator.EVENT_MOTION, viewport=viewport, point=point ) )

	def evEnter(self, viewport, point):
		super( MeshTargetWeldVerticesManipulator, self ).evEnter( viewport, point )
		if self._bSensitive:
			self._coEventHandler.sendEvent( MeshTargetWeldVerticesManipulator.Event( MeshTargetWeldVerticesManipulator.EVENT_MOTION, viewport=viewport, point=point ) )

	def evLeave(self, viewport, point):
		super( MeshTargetWeldVerticesManipulator, self ).evLeave( viewport, point )
		if self._bSensitive:
			self._coEventHandler.sendEvent( MeshTargetWeldVerticesManipulator.Event( MeshTargetWeldVerticesManipulator.EVENT_LEAVE, viewport=viewport, point=point ) )

	def evContextButton(self, viewport):
		if self._bSensitive:
			self._coEventHandler.sendEvent( MeshTargetWeldVerticesManipulator.Event( MeshTargetWeldVerticesManipulator.EVENT_CONTEXT, viewport=viewport ) )


	def _p_onVarChanged(self):
		self._o_reset()

	def _o_reset(self):
		self._coEventHandler.sendEvent( MeshTargetWeldVerticesManipulator.Event( MeshTargetWeldVerticesManipulator.EVENT_RESET ) )





class MeshTargetWeldVerticesTargetListColumnManipulator (GSProcedureTool.GSProcedureToolColumnManipulator):
	def __init__(self, fieldColumn):
		super( MeshTargetWeldVerticesTargetListColumnManipulator, self ).__init__()
		self._fieldColumn = fieldColumn


	def createElement(self, rowEditContext):
		return MeshTargetWeldVerticesManipulator()

	def initialiseElement(self, element, rowEditContext):
		field = self._p_getVar( self._fieldColumn, rowEditContext )
		element.attachVariable( field )
		if rowEditContext.commandHistory is not None:
			element.commandHistory = rowEditContext.commandHistory

	def shutdownElement(self, element, rowEditContext):
		element.detachVariable()
		element.commandHistory = None


	def installElementEventHandlers(self, element, rowEditContext):
		element.manipAttachToView( rowEditContext.view )

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.manipDetachFromView()


	def attachOutputProductSource(self, element, source):
		element.attachMeshSource( source )

	def detachOutputProductSource(self, element, source):
		element.detachMeshSource()




class ToolMeshTargetWeldVertices (GSProcedureTool.GSProcedureTool):
	row = RowRefColumn( ProcMeshTargetWeldVertices )

	functionClass = ProcMeshTargetWeldVertices
	pageTitle = _( 'Tgt.Weld' )
	titleText = _( 'Target Weld' )


	targetListEditor = TargetListColumnEditor( row.targetList )

	targetListManip = MeshTargetWeldVerticesTargetListColumnManipulator( row.targetList )
	targetListManip.enable()


	layout = FormLayoutFactory()
	layout.row()  <<  _( '1. Click to choose a source vertex' )
	layout.row()
	layout.row()  <<  _( '2. Then choose a target:' )
	layout.row()  <<  _( 'click to weld to another vertex' )
	layout.row()  <<  _( '<shift>-click to weld to the' )
	layout.row()  <<  _( 'closest point on an edge' )
	layout.row()  <<  _( '<control>-click to weld to a' )
	layout.row()  <<  _( 'chosen point on an edge.' )
	layout.row()
	layout.row()  <<  _( 'Repeat the above steps' )
	layout.row()  <<  _( 'for each weld.' )
	layout.row()
	layout.row()  <<  targetListEditor.numTargets
	layout.row()  <<  None  <<  ( targetListEditor.removeTarget, 2 )  <<  None
	layout.row()
	layout.finalise()



