##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import copy, deepcopy

from OpenGL.GL import *
from OpenGL.GLU import *

from gSculpt.Math.Math import Colour3f, Vector3, Point3

from gSculpt.Graphics.Graphics import glColor4Colour3fA, glVertex3Point3

from gSculpt.FileIO.IOXml import IOXmlEnum, ioXmlReadBoolNode, ioXmlWriteBoolNode, ioObjectFactoryRegister, ioObjectFactoryRegisterPrimitive

from gSculpt.View.View import PaintLayer

from gSculpt.WorkArea.Viewport import ViewportEventHandler, ViewportScrollListener
from gSculpt.WorkArea.Viewport3d import Viewport3d, Viewport3dPainter

from gSculpt.gSculptConfig import gSculptConfig

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.Util.CoEventHandler import CoEventHandler

from gSculpt.Kernel.DB import *
from gSculpt.Kernel.Enum import Enum
from gSculpt.Kernel.Variable import VarInterface, Variable
from gSculpt.Kernel.ColumnEditors import BasicColumnWidgetEditor

from gSculpt.VarEdit.VarEdit import ChoiceVarEditRadioGroup

from gSculpt.ProceduralCore.GSFunction import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *

from gSculpt.Model import ModelDraw
from gSculpt.Model.TargetList import TargetList
from gSculpt.Model.TargetListColumnEditor import TargetListColumnEditor
from gSculpt.Model.Pick3d import Pick3d

from gSculpt.Mesh.Mesh import GSProductMesh, CppMeshVertexReposition, CppMeshVertexRepositionList, pyMeshVertexRepositionList_to_CppMeshVertexRepositionList

from gSculpt.Manipulator.Manipulator import Manipulator





#
# MeshTemplateTweakVertex class
#

class MeshTemplateTweakVertex (object):
	def __init__(self, pick=None, position=None):
		super( MeshTemplateTweakVertex, self ).__init__()
		if pick is None:
			pick = Pick3d()
		if position is None:
			position = Point3()
		self.pick = pick
		self.position = position


	def toCppMeshVertexReposition(self):
		return CppMeshVertexReposition( self.pick.toMPick(), self.position )


	def __readxml__(self, xmlNode):
		xmlNode.getChild( 'pick' )  >>  self.pick
		xmlNode.getChild( 'position' )  >>  self.position

	def __writexml__(self, xmlNode):
		xmlNode.addChild( 'pick' )  <<  self.pick
		xmlNode.addChild( 'position' )  <<  self.position


	def __copy__(self):
		return MeshTemplateTweak( copy( self.pick ), copy( self.position ) )

ioObjectFactoryRegister( 'MeshTemplateTweakVertex', MeshTemplateTweakVertex )










class ProcMeshTemplateTweakVertices (GSProcedure):
	description = _( 'Template tweak vertices' )

	tweaks = FieldColumn( TargetList, TargetList() )

	inputClass = GSProductMesh



	def procedureInvoke(self, inputObject, errorHandler):
		cppTweaks = CppMeshVertexRepositionList()
		pyMeshVertexRepositionList_to_CppMeshVertexRepositionList( [ tweak.toCppMeshVertexReposition()   for tweak in self.tweaks ], cppTweaks )
		inputObject.repositionVertices( cppTweaks )
		return inputObject







class MeshTemplateTweakVerticesManipulator (Manipulator, ViewportEventHandler, ViewportScrollListener, Viewport3dPainter):
	EV_BUTTON_DOWN = 0
	EV_BUTTON_UP = 1
	EV_MOTION = 2
	EV_LEAVE = 3
	EV_RESET = 4
	EV_OUTPUTCHANGED = 5


	class Event (object):
		def __init__(self, eventType, **kwargs):
			self.eventType = eventType
			self.__dict__.update( kwargs )


	class ResetException:
		pass



	def __init__(self):
		super( MeshTemplateTweakVerticesManipulator, self ).__init__()

		self._tweaksVar = None

		self._coEventHandler = None

		self._highlightVertexPosition = None
		self._initialVertexPosition = None

		self._f_onSetTargetVertex = None
		self._f_onAdjustPosition = None

		self._meshSource = None

		self._o_setUISensitivity( False )




	def manipAttachToView(self, view):
		assert self._view is None, 'already attached to a view'
		super( MeshTemplateTweakVerticesManipulator, self ).manipAttachToView( view )
		self._coEventHandler = CoEventHandler( self._p_tweakLoop )
		self._coEventHandler.start()
		self._view.addEventHandler( self, Viewport3d )
		self._view.addPainter( self, Viewport3d )
		self._view.postRedraw()

	def manipDetachFromView(self):
		assert self._view is not None, 'not attached to a view'
		self._view.removeEventHandler( self, Viewport3d )
		self._view.removePainter( self, Viewport3d )
		self._view.postRedraw()
		self._coEventHandler.terminate()
		self._coEventHandler = None
		super( MeshTemplateTweakVerticesManipulator, self ).manipDetachFromView()



	def attachVariables(self, tweaks):
		self._tweaksVar = tweaks

	def detachVariables(self):
		self._tweaksVar = None



	def _o_setUISensitivity(self, bSensitive):
		pass



	def attachMeshSource(self, source):
		assert self._meshSource is None, 'mesh source already attached'
		self._meshSource = source
		self._meshSource.changedSignal.connect( self._p_onOutputMeshSourceValueChanged )

	def detachMeshSource(self):
		assert self._meshSource is not None, 'no mesh source attached'
		self._meshSource.changedSignal.disconnect( self._p_onOutputMeshSourceValueChanged )
		self._meshSource = None


	def attachInputMeshSource(self, source):
		source.changedSignal.connect( self._p_onMeshSourceValueChanged )

	def detachInputMeshSource(self, source):
		source.changedSignal.disconnect( self._p_onMeshSourceValueChanged )



	def _p_getTarget(self, viewport, targetVertexIndex):
		mesh = self._p_getMesh()
		targetPick = viewport.createPick()

		# Target vertices
		vertexIndex, thruFaceIndex = mesh.pickVertex( targetPick.toMPick() )

		if vertexIndex != targetVertexIndex:
			if vertexIndex != -1:
				# Highlight the vertex
				self._highlightVertexPosition = mesh.getVertexPosition( vertexIndex )
				self._initialVertexPosition = self._highlightVertexPosition
			else:
				self._highlightVertexPosition = None
				self._initialVertexPosition = None
			self._view.postRedraw()

		targetVertexIndex = vertexIndex

		return targetPick, targetVertexIndex



	def _p_getTweakFromUser(self):
		self._highlightVertexPosition = None
		self._initialVertexPosition = None
		targetVertexIndex = -1
		targetPick = None
		position = Point3()
		targetDistance = None
		viewport = None
		startPoint = None
		startPointer = None

		event = self._coEventHandler.waitForNextEvent()

		# Get the target component
		while not ( event.eventType == MeshTemplateTweakVerticesManipulator.EV_BUTTON_DOWN  and  targetVertexIndex != -1 ):
			if event.eventType == MeshTemplateTweakVerticesManipulator.EV_RESET:
				self._view.postRedraw()
				raise MeshTemplateTweakVerticesManipulator.ResetException

			if event.eventType == MeshTemplateTweakVerticesManipulator.EV_OUTPUTCHANGED:
				self._highlightVertexPosition = None
				targetVertexIndex = -1
			else:
				viewport = event.viewport

			if event.eventType != MeshTemplateTweakVerticesManipulator.EV_LEAVE  and  viewport is not None:
				targetPick, targetVertexIndex = self._p_getTarget( viewport, targetVertexIndex )

			event = self._coEventHandler.waitForNextEvent()


		# We have a target component
		# Remove the highlight
		self._highlightVertexPosition = None

		# If the selection is valid
		if self._initialVertexPosition is not None:
			targetDistance = event.viewport.camera.transformWorldToCameraZ( self._initialVertexPosition )
			startPointer = event.viewport.pointerStatus.positionF
			startPoint = event.viewport.camera.transformScreenToWorldCameraZ( startPointer, targetDistance )
			bScaledMotion = event.bScaledMotion


			# Add the tweak object to the tweak list
			self._tweaksVar.value = self._tweaksVar.value + [ MeshTemplateTweakVertex( deepcopy( targetPick ), deepcopy( self._initialVertexPosition ) ) ]
			position = self._initialVertexPosition

			if self._f_onSetTargetVertex is not None  and  targetVertexIndex != -1:
				self._f_onSetTargetVertex( targetVertexIndex, self._initialVertexPosition )
		else:
			return



		event = self._coEventHandler.waitForNextEvent()

		# Now, track the motion
		while not( event.eventType == MeshTemplateTweakVerticesManipulator.EV_BUTTON_UP  and  event.bScaledMotion == bScaledMotion ):
			if event.eventType == MeshTemplateTweakVerticesManipulator.EV_RESET:
				self._view.postRedraw()
				raise MeshTemplateTweakVerticesManipulator.ResetException

			if event.eventType != MeshTemplateTweakVerticesManipulator.EV_OUTPUTCHANGED:
				endPointer = event.viewport.pointerStatus.positionF
				endPoint = event.viewport.camera.transformScreenToWorldCameraZ( endPointer, targetDistance )

				screenDelta = endPointer - startPointer
				translation = endPoint - startPoint


				if bScaledMotion:
					translation *= gSculptConfig.manipulatorFineTuneMotionScale
					screenDelta *= gSculptConfig.manipulatorFineTuneMotionScale


				vertexScreen = event.viewport.transformWorldToScreen( self._initialVertexPosition ).toPoint2()
				vertexScreen += screenDelta
				newPosition = event.viewport.computeDrawPoint( screenPoint=vertexScreen, bNoneIfNoHit=True )
				if newPosition is not None:
					position = newPosition
				else:
					position = self._initialVertexPosition + translation


				if self._f_onAdjustPosition is not None:
					self._f_onAdjustPosition( position )
				self._view.forceRedraw()

			event = self._coEventHandler.waitForNextEvent()


		assert len( self._tweaksVar.value ) > 0, 'no tweaks'
		topTweak = self._tweaksVar.value[-1]
		newTopTweak = MeshTemplateTweakVertex( topTweak.pick, position )
		tweaks = self._tweaksVar.value
		tweaks[-1] = newTopTweak
		self._tweaksVar.value = tweaks
		self._p_commandHistoryBreak()



	def _p_tweakLoop(self):
		while True:
			try:
				self._p_getTweakFromUser()
			except MeshTemplateTweakVerticesManipulator.ResetException:
				self._view.postRedraw()



	def evPrimaryButtonDown(self, viewport):
		super( MeshTemplateTweakVerticesManipulator, self ).evPrimaryButtonDown( viewport )
		self._coEventHandler.sendEvent( MeshTemplateTweakVerticesManipulator.Event( MeshTemplateTweakVerticesManipulator.EV_BUTTON_DOWN, viewport=viewport, bScaledMotion=False ) )

	def evPrimaryButtonUp(self, viewport):
		super( MeshTemplateTweakVerticesManipulator, self ).evPrimaryButtonUp( viewport )
		self._coEventHandler.sendEvent( MeshTemplateTweakVerticesManipulator.Event( MeshTemplateTweakVerticesManipulator.EV_BUTTON_UP, viewport=viewport, bScaledMotion=False ) )


	def evSecondaryButtonDown(self, viewport):
		super( MeshTemplateTweakVerticesManipulator, self ).evPrimaryButtonDown( viewport )
		self._coEventHandler.sendEvent( MeshTemplateTweakVerticesManipulator.Event( MeshTemplateTweakVerticesManipulator.EV_BUTTON_DOWN, viewport=viewport, bScaledMotion=True ) )

	def evSecondaryButtonUp(self, viewport):
		super( MeshTemplateTweakVerticesManipulator, self ).evPrimaryButtonUp( viewport )
		self._coEventHandler.sendEvent( MeshTemplateTweakVerticesManipulator.Event( MeshTemplateTweakVerticesManipulator.EV_BUTTON_UP, viewport=viewport, bScaledMotion=True ) )


	def evMotion(self, viewport, point):
		super( MeshTemplateTweakVerticesManipulator, self ).evMotion( viewport, point )
		self._coEventHandler.sendEvent( MeshTemplateTweakVerticesManipulator.Event( MeshTemplateTweakVerticesManipulator.EV_MOTION, viewport=viewport ) )

	def evEnter(self, viewport, point):
		super( MeshTemplateTweakVerticesManipulator, self ).evMotion( viewport, point )
		self._coEventHandler.sendEvent( MeshTemplateTweakVerticesManipulator.Event( MeshTemplateTweakVerticesManipulator.EV_MOTION, viewport=viewport ) )

	def evLeave(self, viewport, point):
		super( MeshTemplateTweakVerticesManipulator, self ).evMotion( viewport, point )
		self._coEventHandler.sendEvent( MeshTemplateTweakVerticesManipulator.Event( MeshTemplateTweakVerticesManipulator.EV_LEAVE, viewport=viewport ) )


	def evPaint3d(self, viewport, paintLayer):
		# Paint the target highlight
		if self._highlightVertexPosition is not None:
			ModelDraw.drawVertexPickHighlight( self._highlightVertexPosition, ModelDraw.TARGET_HIGHLIGHT_COLOUR, paintLayer )



	def _p_onMeshSourceValueChanged(self):
		self._coEventHandler.sendEvent( MeshTemplateTweakVerticesManipulator.Event( MeshTemplateTweakVerticesManipulator.EV_RESET ) )

	def _p_onOutputMeshSourceValueChanged(self):
		self._coEventHandler.sendEvent( MeshTemplateTweakVerticesManipulator.Event( MeshTemplateTweakVerticesManipulator.EV_OUTPUTCHANGED ) )

	def _p_getMesh(self):
		if self._meshSource is not None:
			return self._meshSource.getValueAsClass( GSProductMesh )
		else:
			return None






class MeshTemplateTweakVerticesColumnManipulator (GSProcedureTool.GSProcedureToolColumnManipulator):
	def __init__(self, tweaksColumn):
		super( MeshTemplateTweakVerticesColumnManipulator, self ).__init__()
		self._tweaksColumn = tweaksColumn


	def createElement(self, rowEditContext):
		return MeshTemplateTweakVerticesManipulator()

	def destroyElement(self, element, rowEditContext):
		del rowEditContext.rowEditor.tweakManip

	def initialiseElement(self, element, rowEditContext):
		tweaks = self._p_getVar( self._tweaksColumn, rowEditContext )
		element.attachVariables( tweaks )
		if rowEditContext.commandHistory is not None:
			element.commandHistory = rowEditContext.commandHistory


	def shutdownElement(self, element, rowEditContext):
		element.detachVariables()
		element.commandHistory = None


	def installElementEventHandlers(self, element, rowEditContext):
		element.manipAttachToView( rowEditContext.view )

	def uninstallElementEventHandlers(self, element, rowEditContext):
		element.manipDetachFromView()


	def attachInputProductSource(self, element, source):
		element.attachInputMeshSource( source )

	def detachInputProductSource(self, element, source):
		element.detachInputMeshSource( source )

	def attachOutputProductSource(self, element, source):
		element.attachMeshSource( source )

	def detachOutputProductSource(self, element, source):
		element.detachMeshSource()




class TemplateTweakVerticesAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcMeshTemplateTweakVertices


	def __init__(self, function, inputObject, errorHandler):
		super( TemplateTweakVerticesAdjustableInvoker, self ).__init__( function, inputObject, errorHandler )
		self._vertexIndex = None
		self._initialVertexPosition = None



	def init(self):
		# Initialise with full pre-compute. This will remove the cost of the full pre-compute which would normally occur on each modification of the tweak list
		self._inputObject.preComputeFull()

	def main(self):
		self.mesh = deepcopy( self._inputObject )
		cppTweaks = CppMeshVertexRepositionList()
		pyMeshVertexRepositionList_to_CppMeshVertexRepositionList( [ tweak.toCppMeshVertexReposition()   for tweak in self._proc.tweaks ], cppTweaks )
		self.mesh.repositionVertices( cppTweaks )



	def setTargetVertex(self, vertexIndex, initialPosition):
		self._vertexIndex = vertexIndex
		self._initialVertexPosition = initialPosition


	def adjustPosition(self, position):
		if self._vertexIndex is not None:
			self.mesh.setVertexPosition( self._vertexIndex, position )


	def getResult(self):
		return self.mesh






class ToolMeshTemplateTweakVertices (GSProcedureTool.GSProcedureTool):
	row = RowRefColumn( ProcMeshTemplateTweakVertices )

	functionClass = ProcMeshTemplateTweakVertices
	adjustableInvoker = TemplateTweakVerticesAdjustableInvoker
	pageTitle = _( 'TmpTwk' )
	titleText = _( 'Tempate Tweak' )


	tweakListEditor = TargetListColumnEditor( row.tweaks,
									targetLabelText=_( '%d tweak' ), pluralTargetLabelText=_( '%d tweaks' ), removeTargetButtonText=_( 'Remove tweak' ) )

	tweakManip = MeshTemplateTweakVerticesColumnManipulator( row.tweaks )
	tweakManip.enable()


	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Click and drag to move.' )
	layout.row()  <<  _( 'Vertices will be constrained to' )
	layout.row()  <<  _( 'the surfaces of any drawing' )
	layout.row()  <<  _( ' templates that are enabled.' )
	layout.row()
	layout.row()  <<  tweakListEditor.numTargets
	layout.row()  <<  None  <<  ( tweakListEditor.removeTarget, 2 )  <<  None
	layout.row()
	layout.finalise()



	def initialise(self):
		super( ToolMeshTemplateTweakVertices, self ).initialise()
		self.tweakManip._f_onSetTargetVertex = self._onSetTargetVertex
		self.tweakManip._f_onAdjustPosition = self._onAdjustPosition


	def shutdown(self):
		self.tweakManip._f_onSetTargetVertex = None
		self.tweakManip._f_onAdjustPosition = None
		super( ToolMeshTemplateTweakVertices, self ).shutdown()



	def _onSetTargetVertex(self, vertexIndex, initialPosition):
		self._adjustableInvoker.setTargetVertex( vertexIndex, initialPosition )

	def _onAdjustPosition(self, position):
		self._adjustableInvoker.adjustPosition( position )




