##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from gSculpt.Math.Math import Point3, Matrix4, Axis
from gSculpt.Math.MathField import Point3Field, AxisField


from gSculpt.Model.Model import GSProductModelBlank

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.Kernel.DB import *
from gSculpt.Kernel.ColumnEditors import *

from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool.GSProcedureTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *

from gSculpt.Manipulator.ColumnManipulators import *

from gSculpt.Mesh import Mesh
from gSculpt.Mesh.Mesh import CylinderPrimitive, MeshVertexList, GSProductMesh



class ProcCylinderPrimitive (GSProcedure):
	description = _( 'Cylinder' )

	axis = FieldColumn( Axis, Axis.AXIS_Y )
	angularSegments = FieldColumn( int, 12 )
	lengthSegments = FieldColumn( int, 1 )
	radialSegments = FieldColumn( int, 1 )
	radius = FieldColumn( float, 20.0 )
	length = FieldColumn( float, 100.0 )
	bCapStart = FieldColumn( bool, True )
	bCapEnd = FieldColumn( bool, True )
	position = FieldColumn( Point3, Point3() )


	inputClass = GSProductModelBlank
	outputClass = GSProductMesh


	def procedureInvoke(self, inputObject, errorHandler):
		prim = CylinderPrimitive( self.axis, self.angularSegments, self.lengthSegments, self.radialSegments, self.radius, self.radius, self.length,
							self.bCapStart, self.bCapEnd )
		mesh = GSProductMesh()
		verts = MeshVertexList()
		prim.createCylinder( mesh, verts )
		verts.transform( Matrix4.translate( self.position.toVector3() ), mesh )
		return mesh



class CylinderAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcCylinderPrimitive

	def initInvoke(self):
		self.verts = MeshVertexList()

	def main(self):
		prim = CylinderPrimitive( self._proc.axis, self._proc.angularSegments, self._proc.lengthSegments, self._proc.radialSegments, self._proc.radius, self._proc.radius, self._proc.length, self._proc.bCapStart, self._proc.bCapEnd )
		self.verts.clear()
		self.mesh = GSProductMesh()
		prim.createCylinder( self.mesh, self.verts )
		return self.mesh

	def save(self):
		self.verts.savePositions( self.mesh )

	def restore(self):
		self.verts.restorePositions( self.mesh )

	def adjust(self):
		self.verts.transform( Matrix4.translate( self._proc.position.toVector3() ), self.mesh )

	def getResult(self):
		return self.mesh




class ToolCylinderPrimitive (GSProcedureTool):
	row = RowRefColumn( ProcCylinderPrimitive )

	functionClass = ProcCylinderPrimitive
	adjustableInvoker = CylinderAdjustableInvoker
	pageTitle = _( 'Cylinder' )
	titleText = _( 'Create Cylinder' )

	axisRadio = AxisColumnEditRadioGroup( row.axis )
	angularSegsEntry = IntColumnEditEntry( row.angularSegments, 3, 1048576, 1, 10 )
	lengthSegsEntry = IntColumnEditEntry( row.lengthSegments, 1, 1048576, 1, 10 )
	radialSegsEntry = IntColumnEditEntry( row.radialSegments, 1, 1048576, 1, 10 )
	radiusEntry = FloatColumnEditEntry( row.radius, 0.0, 1.0e8, 0.1, 10.0, 4 )
	lengthEntry = FloatColumnEditEntry( row.length, 0.0, 1.0e8, 1.0, 10.0, 4 )
	bCapStartCheck = BoolColumnEditCheck( row.bCapStart, _( 'Cap start' ) )
	bCapEndCheck = BoolColumnEditCheck( row.bCapEnd, _( 'Cap end' ) )
	positionEntry = Point3ColumnEditEntry( row.position, -1.0e8, 1.0e8, 1.0, 10.0, 4 )

	positionManip = Point3ColumnManipulator( row.position, snapping=ColumnManipulatorSnappingDistance() )
	positionManip.enable()

	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Cylinder axis:' )
	layout.row()  <<  axisRadio.x  <<  axisRadio.y  <<  axisRadio.z
	layout.row()
	layout.row()  <<  _( 'Angular segs.:' )  <<  angularSegsEntry.entry
	layout.row()  <<  _( 'Length segs.:' )  <<  lengthSegsEntry.entry
	layout.row()  <<  _( 'Radial segs.:' )  <<  radialSegsEntry.entry
	layout.row()
	layout.row()  <<  _( 'Radius:' )  <<  radiusEntry.entry
	layout.row()  <<  _( 'Length:' )  <<  lengthEntry.entry
	layout.row()
	layout.row()  <<  bCapStartCheck.check  <<  bCapEndCheck.check
	layout.row()
	layout.row()  <<  _( 'Position' )
	layout.row()  <<  'X:'  <<  positionEntry.x
	layout.row()  <<  'Y:'  <<  positionEntry.y
	layout.row()  <<  'Z:'  <<  positionEntry.z
	layout.row()
	layout.finalise()
