##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from gSculpt.Math.Math import Point3, Matrix4, Axis
from gSculpt.Math.MathField import Point3Field, AxisField

from gSculpt.FileIO.IOXml import IOXmlEnum, ioObjectFactoryRegisterPrimitive

from gSculpt.Model.Model import GSProductModelBlank

from gSculpt.UI.FormLayout import FormLayoutFactory

from gSculpt.VarEdit.VarEdit import ChoiceVarEditRadioGroup

from gSculpt.Kernel.DB import *
from gSculpt.Kernel.DBEdit import BasicColumnWidgetEditor
from gSculpt.Kernel.ColumnEditors import *

from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool.GSFunctionTool import *
from gSculpt.ProceduralTool.GSProcedureTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *

from gSculpt.Manipulator.ColumnManipulators import *

from gSculpt.Mesh import Mesh
from gSculpt.Mesh.Mesh import TubePrimitive, TubePrimitiveCapMaterialMapping, MeshVertexList, GSProductMesh




# XML IO for TubePrimitiveCapMaterialMapping

capMappingIOXmlEnum = IOXmlEnum( TubePrimitiveCapMaterialMapping, 'TubePrimitiveCapMaterialMapping', TubePrimitiveCapMaterialMapping.PLANAR )
capMappingIOXmlEnum.value( TubePrimitiveCapMaterialMapping.PLANAR, 'planar' )
capMappingIOXmlEnum.value( TubePrimitiveCapMaterialMapping.POLAR, 'polar' )

ioXmlReadTubePrimitiveCapMaterialMappingProp = capMappingIOXmlEnum.propReader
ioXmlReadTubePrimitiveCapMaterialMappingNode = capMappingIOXmlEnum.nodeReader
ioXmlWriteTubePrimitiveCapMaterialMappingProp = capMappingIOXmlEnum.propWriter
ioXmlWriteTubePrimitiveCapMaterialMappingNode = capMappingIOXmlEnum.nodeWriter

ioObjectFactoryRegisterPrimitive( 'TubePrimitiveCapMaterialMapping', TubePrimitiveCapMaterialMapping )


# Field for TubePrimitiveCapMaterialMapping

class TubePrimitiveCapMaterialMappingField (Field):
	valueClass = TubePrimitiveCapMaterialMapping
	defaultValue = TubePrimitiveCapMaterialMapping.PLANAR
	xmlNodeValueReader = ioXmlReadTubePrimitiveCapMaterialMappingNode
	xmlNodeValueWriter = ioXmlWriteTubePrimitiveCapMaterialMappingNode

registerFieldType( TubePrimitiveCapMaterialMapping, TubePrimitiveCapMaterialMappingField )


# Variable editor for TubePrimitiveCapMaterialMapping

class TubePrimitiveCapMaterialMappingVarEditRadioGroup (ChoiceVarEditRadioGroup):
	def __init__(self):
		super( TubePrimitiveCapMaterialMappingVarEditRadioGroup, self ).__init__()
		self.planar = self.addChoice( TubePrimitiveCapMaterialMapping.PLANAR, _( 'Planar' ) )
		self.polar = self.addChoice( TubePrimitiveCapMaterialMapping.POLAR, _( 'Polar' ) )


# Column editor for TubePrimitiveCapMaterialMapping

class TubePrimitiveCapMaterialMappingColumnEditRadioGroup (BasicColumnWidgetEditor):
	def __init__(self, fieldColumn):
		super( TubePrimitiveCapMaterialMappingColumnEditRadioGroup, self ).__init__( [ 'planar', 'polar' ] )
		self._fieldColumn = fieldColumn


	def createElement(self, rowEditContext):
		return TubePrimitiveCapMaterialMappingVarEditRadioGroup()

	def initialiseElement(self, element, rowEditContext):
		super( TubePrimitiveCapMaterialMappingColumnEditRadioGroup, self ).initialiseElement( element, rowEditContext )
		field = self._p_getVar( self._fieldColumn, rowEditContext )
		element.attachVariable( field )

	def shutdownElement(self, element, rowEditContext):
		super( TubePrimitiveCapMaterialMappingColumnEditRadioGroup, self ).shutdownElement( element, rowEditContext )
		element.detachVariable()





class ProcTubePrimitive (GSProcedure):
	description = _( 'Tube' )

	axis = FieldColumn( Axis, Axis.AXIS_Y )
	slices = FieldColumn( int, 12 )
	lengthSegments = FieldColumn( int, 1 )
	radialSegments = FieldColumn( int, 1 )
	innerRadius = FieldColumn( float, 40.0 )
	outerRadius = FieldColumn( float, 50.0 )
	length = FieldColumn( float, 100.0 )
	bInnerSection = FieldColumn( bool, True )
	bOuterSection = FieldColumn( bool, True )
	bCapStart = FieldColumn( bool, True )
	bCapEnd = FieldColumn( bool, True )
	capMapping = FieldColumn( TubePrimitiveCapMaterialMapping, TubePrimitiveCapMaterialMapping.PLANAR )
	position = FieldColumn( Point3, Point3() )


	inputClass = GSProductModelBlank
	outputClass = GSProductMesh


	def procedureInvoke(self, inputObject, errorHandler):
		prim = TubePrimitive( self.axis, self.radialSegments, self.lengthSegments, self.slices, self.innerRadius, self.outerRadius, self.length, self.bInnerSection, self.bOuterSection, self.bCapStart, self.bCapEnd, self.capMapping )
		mesh = GSProductMesh()
		verts = MeshVertexList()
		prim.createTube( mesh, verts )
		verts.transform( Matrix4.translate( self.position.toVector3() ), mesh )
		return mesh



class TubeAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcTubePrimitive

	def initInvoke(self):
		self.verts = MeshVertexList()

	def main(self):
		prim = TubePrimitive( self._proc.axis, self._proc.radialSegments, self._proc.lengthSegments, self._proc.slices, self._proc.innerRadius, self._proc.outerRadius, self._proc.length, self._proc.bInnerSection, self._proc.bOuterSection, self._proc.bCapStart, self._proc.bCapEnd, self._proc.capMapping )
		self.verts.clear()
		self.mesh = GSProductMesh()
		prim.createTube( self.mesh, self.verts )
		return self.mesh

	def save(self):
		self.verts.savePositions( self.mesh )

	def restore(self):
		self.verts.restorePositions( self.mesh )

	def adjust(self):
		self.verts.transform( Matrix4.translate( self._proc.position.toVector3() ), self.mesh )

	def getResult(self):
		return self.mesh




class ToolTubePrimitive (GSProcedureTool):
	row = RowRefColumn( ProcTubePrimitive )

	functionClass = ProcTubePrimitive
	adjustableInvoker = TubeAdjustableInvoker
	pageTitle = _( 'Tube' )
	titleText = _( 'Create Tube' )

	axisRadio = AxisColumnEditRadioGroup( row.axis )
	slicesEntry = IntColumnEditEntry( row.slices, 3, 1048576, 1, 10 )
	lengthSegsEntry = IntColumnEditEntry( row.lengthSegments, 1, 1048576, 1, 10 )
	radialSegsEntry = IntColumnEditEntry( row.radialSegments, 1, 1048576, 1, 10 )
	innerRadiusEntry = FloatColumnEditEntry( row.innerRadius, 0.0, 1.0e8, 0.1, 10.0, 4 )
	outerRadiusEntry = FloatColumnEditEntry( row.outerRadius, 0.0, 1.0e8, 0.1, 10.0, 4 )
	lengthEntry = FloatColumnEditEntry( row.length, 0.0, 1.0e8, 1.0, 10.0, 4 )
	bInnerSectionCheck = BoolColumnEditCheck( row.bInnerSection, _( 'Inner section' ) )
	bOuterSectionCheck = BoolColumnEditCheck( row.bOuterSection, _( 'Outer section' ) )
	bCapStartCheck = BoolColumnEditCheck( row.bCapStart, _( 'Cap start' ) )
	bCapEndCheck = BoolColumnEditCheck( row.bCapEnd, _( 'Cap end' ) )
	capMappingRadio = TubePrimitiveCapMaterialMappingColumnEditRadioGroup( row.capMapping )
	positionEntry = Point3ColumnEditEntry( row.position, -1.0e8, 1.0e8, 1.0, 10.0, 4 )

	positionManip = Point3ColumnManipulator( row.position, snapping=ColumnManipulatorSnappingDistance() )
	positionManip.enable()

	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Tube axis:' )
	layout.row()  <<  axisRadio.x  <<  axisRadio.y  <<  axisRadio.z
	layout.row()
	layout.row()  <<  _( 'Slices.:' )  <<  slicesEntry.entry
	layout.row()  <<  _( 'Length segs.:' )  <<  lengthSegsEntry.entry
	layout.row()  <<  _( 'Radial segs.:' )  <<  radialSegsEntry.entry
	layout.row()
	layout.row()  <<  _( 'Inner radius:' )  <<  innerRadiusEntry.entry
	layout.row()  <<  _( 'Outer radius:' )  <<  outerRadiusEntry.entry
	layout.row()  <<  _( 'Length:' )  <<  lengthEntry.entry
	layout.row()
	layout.row()  <<  bInnerSectionCheck.check  <<  bOuterSectionCheck.check
	layout.row()  <<  bCapStartCheck.check  <<  bCapEndCheck.check
	layout.row()
	layout.row()  <<  _( 'Cap material mapping:' )
	layout.row()  <<  capMappingRadio.planar  <<  capMappingRadio.polar
	layout.row()
	layout.row()  <<  _( 'Position' )
	layout.row()  <<  'X:'  <<  positionEntry.x
	layout.row()  <<  'Y:'  <<  positionEntry.y
	layout.row()  <<  'Z:'  <<  positionEntry.z
	layout.row()
	layout.finalise()
