##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import copy

from gSculpt.Util.SignalSlot import *

from gSculpt.FileIO import IOXml

from gSculpt.Kernel.KMeta import *
from gSculpt.Kernel.DB import FieldColumn, CompositeColumn

from gSculpt.GSGraph import GSGraph
from gSculpt.GSGraph.GSGraph import *

from gSculpt.ProceduralCore.GSFunction import GSFunction, GSFunctionClass
from gSculpt.ProceduralCore.GSPlan import GSPlan, GSPlanInitialProductSource
from gSculpt.ProceduralCore import GSFunctionGraphCommandTracker






class GSFunctionGraphNodePin (object):
	"""Funciton graph node pin"""
	def __init__(self, functionGraphNode=None, name=''):
		self._functionGraphNode = functionGraphNode
		self._name = name
		self._graphPin = self._o_createGraphPin()
		self._graphPin._functionGraphNodePin = self


	def _p_setFunctionGraphNode(self, node):
		self._functionGraphNode = node


	def __readxml__(self, xmlNode):
		pass

	def __writexml__(self, xmlNode):
		IOXml.ioXmlWriteStringNode( xmlNode.addChild( 'name' ), self._name )


	def _o_createGraphPin(self):
		assert False, abstract


	def _f_setName(self, name):
		self._name = name
		self._graphPin.name = self._name


	def _getGraphPin(self):
		return self._graphPin

	def _getName(self):
		return self._graphPin.name


	name = property( _getName, doc='Name of the pin' )




class GSFunctionGraphNodeSource (GSFunctionGraphNodePin):
	"""Function graph node source (output) pin"""
	def _o_createGraphPin(self):
		assert isinstance( self._name, str )
		return GSGraphSource( self._name )

	graphSource = property( GSFunctionGraphNodePin._getGraphPin, 'graph source' )


IOXml.ioObjectFactoryRegister( 'GSFunctionGraphNodeSource', GSFunctionGraphNodeSource )




class GSFunctionGraphNodeSink (GSFunctionGraphNodePin):
	"""Function graph node sink pin"""
	def __contains__(self, source):
 		return source._graphPin in self._graphPin

	def __len__(self):
		return len( self._graphPin )

	def maxLen(self):
		"""Get the maximum number of inputs that this sink can take"""
		return self._graphPin.maxLen()

	def __getitem__(self, index):
		if isinstance( index, slice ):
			return [ self._getSourceFromGraphLink( link )   for link in self._graphPin[index] ]
		else:
			return self._getSourceFromGraphLink( self._graphPin[index] )

	def __iter__(self):
		for link in self._graphPin:
			yield self._getSourceFromGraphLink( link )


	def getInputProduct(self, index):
		"""Get the input product from the input indexed by @index"""
		if isinstance( index, slice ):
			return [ self._getResultProductFromSource( source )   for source in self[index] ]
		else:
			return self._getResultProductFromSource( self[index] )

	def inputProducts(self):
		"""The the input products from the inputs into this sink pin (returns a list)"""
		for source in self:
			yield self._getResultProductFromSource( source )

	def getInputProductClass(self, index):
		"""Get the class of the input product from the input indexed by @index"""
		if isinstance( index, slice ):
			return [ self._getResultProductClassFromSource( source )   for source in self[index] ]
		else:
			return self._getResultProductClassFromSource( self[index] )

	def inputProductClasses(self):
		"""The the class of the input products from the inputs into this sink pin (returns a list)"""
		for source in self:
			yield self._getResultProductClassFromSource( source )


	def copyLinksFrom(self, sink):
		"""Copy the input links from @sink"""
		self._graphPin.copyLinskFrom( sink._graphPin )


	@staticmethod
	def _getSourceFromGraphLink(link):
		if link is not None:
			return link.source._functionGraphNodePin
		else:
			return None


	@staticmethod
	def _getResultProductFromSource(source):
		if source is not None:
			return source.getResultProduct()
		else:
			return None

	@staticmethod
	def _getResultProductClassFromSource(source):
		if source is not None:
			return source.getResultProductClass()
		else:
			return None


	graphSink = property( GSFunctionGraphNodePin._getGraphPin, doc='graph sink' )



class GSFunctionGraphNodeSinkSingle (GSFunctionGraphNodeSink):
	"""Funciton graph node sink that can take a single input"""
	def _o_createGraphPin(self):
		assert isinstance( self._name, str )
		return GSGraphSinkSingle( self._name )

IOXml.ioObjectFactoryRegister( 'GSFunctionGraphNodeSinkSingle', GSFunctionGraphNodeSinkSingle )



class GSFunctionGraphNodeSinkMultiple (GSFunctionGraphNodeSink):
	"""Function graph node sink that can take multiple inputs"""
	def _o_createGraphPin(self):
		assert isinstance( self._name, str )
		return GSGraphSinkMultiple( self._name )

IOXml.ioObjectFactoryRegister( 'GSFunctionGraphNodeSinkMultiple', GSFunctionGraphNodeSinkMultiple )





class NodePinColumn (KMetaMember):
	"""Node pin column"""
	def __init__(self, name, doc = ''):
		super( NodePinColumn, self ).__init__( doc )
		self._attrName = None
		self._pinName = name


	def _f_metaMember_initMetaMember(self, cls, name):
		super( NodePinColumn, self )._f_metaMember_initMetaMember( cls, name )
		self._attrName = intern( NodePinColumn._p_computeAttrName( name ) )
		self._xmlName = self._o_computeXmlName( name )


	def _f_readInstancePinXml(self, instance, xmlNode):
		pin = self._f_getPinFromInstance( instance )
		xmlNode.getChild( self._xmlName )  >>  pin
		pin._f_setName( self._name )

	def _f_writeInstancePinXml(self, instance, xmlNode):
		xmlNode.addChild( self._xmlName )  <<  self._f_getPinFromInstance( instance )



	def __get__(self, obj, objtype = None):
		if obj is not None:
			return self._f_getPinFromInstance( obj )
		else:
			return self

	def __set__(self, obj, value):
		raise AttributeError, 'cannot set pin'

	def __delete__(self, obj):
		raise AttributeError, 'cannot delete pin'


	def _f_getPinFromInstance(self, instance):
		return getattr( instance, self._attrName )


	@staticmethod
	def _p_computeAttrName(name):
		return '_pin_' + name





class NodeSinkColumn (NodePinColumn):
	"""Node sink column"""
	def _f_initClass(self, cls):
		pass

	def _f_metaMember_initInstance(self, instance, srcInstance = None):
		sink = self._o_createSink( instance, self._pinName )
		instance._f_addColumnSink( sink )
		# Copy links from the source instance
		if srcInstance is not None:
			sink.copyLinksFrom( self._f_getPinFromInstance( srcInstance ) )
		setattr( instance, self._attrName, sink )


	def _o_createSink(self, instance, name):
		assert False, 'abstract'


class NodeSinkSingleColumn (NodeSinkColumn):
	"""Node single input sink column"""
	def _o_createSink(self, instance, name):
		return GSFunctionGraphNodeSinkSingle( instance, name )


class NodeSinkMultipleColumn (NodeSinkColumn):
	"""Node multiple input sink column"""
	def _o_createSink(self, instance, name):
		return GSFunctionGraphNodeSinkMultiple( instance, name )







class NodeSourceColumn (NodePinColumn):
	"""Node source column"""
	def _f_initClass(self, cls):
		pass

	def _f_metaMember_initInstance(self, instance, srcInstance = None):
		source = self._o_createSource( instance, self._pinName )
		instance._f_addColumnSource( source )
		setattr( instance, self._attrName, source )


	def _o_createSource(self, instance, name):
		return GSFunctionGraphNodeSource( instance, name )






class GSFunctionGraphNodeClass (GSFunctionClass):
	"""GSFunctionGraphNode metaclass"""
	def __init__(cls, clsName, clsBases, clsDict):
		cls._p_processPins( clsName, clsBases, clsDict, NodeSinkColumn, intern( 's_GSFunctionGraphNode_sinkColumns' ) )
		cls._p_processPins( clsName, clsBases, clsDict, NodeSourceColumn, intern( 's_GSFunctionGraphNode_sourceColumns' ) )

		super( GSFunctionGraphNodeClass, cls ).__init__( clsName, clsBases, clsDict )



	def getSinkColumnByName(cls, sinkName):
		return cls.s_GSFunctionGraphNode_sinkColumns[sinkName]

	def getSourceColumnByName(cls, sourceName):
		return cls.s_GSFunctionGraphNode_sourceColumns[sourceName]


	def _p_processPins(cls, clsName, clsBases, clsDict, pinClass, pinColumnDictAttrName):
		pinColumns = cls._o_gatherDictFromBases( clsBases, pinColumnDictAttrName )

		for name, value in clsDict.items():
			if isinstance( value, pinClass ):
				pinColumns[name] = value


		# Combine the gathered lists of fields with the lists from base classes
		setattr( cls, pinColumnDictAttrName, pinColumns )





class GSFunctionGraphGSGraphNode (GSGraphNode):
	"""Graph node for GSFunctionGraphNode"""
	def __init__(self, functionGraphNode):
		super( GSFunctionGraphGSGraphNode, self ).__init__()
		self._functionGraphNode = functionGraphNode





class GSFunctionGraphNode (GSFunction):
	"""Function graph node base class"""
	__metaclass__ = GSFunctionGraphNodeClass
	trackerClass = GSFunctionGraphCommandTracker.GSFunctionGraphNodeCommandTracker
	description = 'Abstract function graph node'


	def __init__(self, name = None, src=None):
		"""Constructor:
		@name - name of node
		@src - source object to copy"""
		self._columnSinks = []
		self._columnSources = []
		self._sinks = []
		self._sources = []
		self._graphNode = None

		GSFunction.__init__( self, src )

		self._graphNode = self._o_createGSGraphNode()
		for sink in self._sinks:
			self._graphNode.addSink( sink.graphSink )
		for source in self._sources:
			self._graphNode.addSource( source.graphSource )




	def addSink(self, sink):
		self._sinks.append( sink )
		if self._graphNode is not None:
			self._graphNode.addSink( sink.graphSink )

	def removeSink(self, sink):
		self._sinks.remove( sink )
		if self._graphNode is not None:
			self._graphNode.removeSink( sink.graphSink )


	def addSource(self, source):
		self._sources.append( source )
		if self._graphNode is not None:
			self._graphNode.addSource( source.graphSource )

	def removeSource(self, source):
		self._sources.remove( source )
		if self._graphNode is not None:
			self._graphNode.removeSource( source.graphSource )


	def clearSinksAndSources(self):
		for sink in self._sinks:
			if sink not in self._columnSinks:
				self._graphNode.removeSink( sink.graphSink )
		self._sinks = copy( self._columnSinks )

		for source in self._sources:
			if source not in self._columnSources:
				self._graphNode.removeSource( source.graphSource )
		self._sources = copy( self._columnSources )



	def _f_addColumnSink(self, sink):
		self._columnSinks.append( sink )
		self.addSink( sink )

	def _f_addColumnSource(self, source):
		self._columnSources.append( source )
		self.addSource( source )


	def __readxml__(self, xmlNode):
		super( GSFunctionGraphNode, self ).__readxml__( xmlNode )

		self.clearSinksAndSources()

		xmlNode.getChild( 'graph_node' )  >>  self._graphNode

		for sinkColumn in self.s_GSFunctionGraphNode_sinkColumns.values():
			sinkColumn._f_readInstancePinXml( self, xmlNode )

		for sourceColumn in self.s_GSFunctionGraphNode_sourceColumns.values():
			sourceColumn._f_readInstancePinXml( self, xmlNode )

		sinksXmlNode = xmlNode.getChild( 'sinks' )
		for sinkXmlNode in sinksXmlNode.childrenNamed( 'sink' ):
			sink = sinkXmlNode.readObject()
			if sink not in self._sinks:
				# Need to given newly created pins a reference to the graph node that owns them
				sink._p_setFunctionGraphNode( self )
				self.addSink( sink )

		sourcesXmlNode = xmlNode.getChild( 'sources' )
		for sourceXmlNode in sourcesXmlNode.childrenNamed( 'source' ):
			source = sourceXmlNode.readObject()
			if source not in self._sources:
				# Need to given newly created pins a reference to the graph node that owns them
				source._p_setFunctionGraphNode( self )
				self.addSource( source )




	def __writexml__(self, xmlNode):
		super( GSFunctionGraphNode, self ).__writexml__( xmlNode )

		xmlNode.addChild( 'graph_node' )  <<  self._graphNode

		for sinkColumn in self.s_GSFunctionGraphNode_sinkColumns.values():
			sinkColumn._f_writeInstancePinXml( self, xmlNode )

		for sourceColumn in self.s_GSFunctionGraphNode_sourceColumns.values():
			sourceColumn._f_writeInstancePinXml( self, xmlNode )

		sinksXmlNode = xmlNode.addChild( 'sinks' )
		for sink in self._sinks:
			sinksXmlNode.addChild( 'sink' ).writeObject( sink )

		sourcesXmlNode = xmlNode.addChild( 'sources' )
		for source in self._sources:
			sourcesXmlNode.addChild( 'source' ).writeObject( source )



	def _o_createGSGraphNode(self):
		return GSFunctionGraphGSGraphNode( self )



	@staticmethod
	def getFunctionGraphNodeForGSGraphNode(graphNode):
		"""Get the function graph node for a given gs-graph node"""
		if graphNode is not None:
			return graphNode._functionGraphNode
		else:
			return None


	def _p_getSinks(self):
		return self._sinks

	def _p_getSources(self):
		return self._sources

	def _p_getGraphNode(self):
		return self._graphNode


	sinks = property( _p_getSinks )
	sources = property( _p_getSources )
	graphNode = property( _p_getGraphNode )







class GSFunctionGraph (object):
	"""Function graph - a function graph
	Contains the list of nodes, and the graph which controls their connections"""
	trackerClass = GSFunctionGraphCommandTracker.GSFunctionGraphCommandTracker


	class _NodeList (object):
	 	def __init__(self, functionGraph):
			super( GSFunctionGraph._NodeList, self ).__init__()
			self._nodes = []
			self._functionGraph = functionGraph


		def __iter__(self):
			return iter( self._nodes )

		def __len__(self):
			return len( self._nodes )

		def __contains__(self, node):
			return node in self._nodes

		def __getitem__(self, index):
			return self._nodes[index]

		def __iadd__(self, nodes):
			self.extend( nodes )

		def index(self, node):
			return self._nodes.index( node )


		def append(self, node):
			assert node not in self._nodes, 'function graph node already in list'
			self._nodes.append( node )
			self._functionGraph._p_onNodeAppended( node )

		def extend(self, nodes):
			for node in nodes:
				append( node )

		def remove(self, node):
			assert node in self._nodes, 'function graph node not in list'
			self._functionGraph._p_onRemoveNode( node )
			self._nodes.remove( node )
			self._functionGraph._p_onNodeRemoved( node )



	nodeAddedSignal = ClassSignal()
	removeNodeSignal = ClassSignal()
	nodeRemovedSignal = ClassSignal()


	def __init__(self):
		super( GSFunctionGraph, self ).__init__()
		self._graph = GSGraph()
		self._nodes = GSFunctionGraph._NodeList( self )

		self._commandTracker_ = None



	def __readxml__(self, xmlNode):
		self._p_clear()

		for n in xmlNode.getChild( 'nodes' ).childrenNamed( 'node' ):
			node = n.readObject()
			if node is not None  and  isinstance( node, GSFunctionGraphNode ):
				self._nodes.append( node )

		for linkNode in xmlNode.getChild( 'links' ).childrenNamed( 'link' ):
			sink = linkNode.getChild( 'sink' ).readObject()
			source = linkNode.getChild( 'source' ).readObject()
			sink.graphSink.addInput( source.graphSource )


	def __writexml__(self, xmlNode):
		nodeListNode = xmlNode.addChild( 'nodes' )
		for node in self._nodes:
			nodeListNode.addChild( 'node' ).writeObject( node )

		links = xmlNode.addChild( 'links' )
		for link in self._graph.links:
			linkNode = links.addChild( 'link' )
			linkNode.addChild( 'source' ).writeObject( link.source._functionGraphNodePin )
			linkNode.addChild( 'sink' ).writeObject( link.sink._functionGraphNodePin )



	def _p_clear(self):
		while len( self._nodes ) > 0:
			self._nodes.remove( self._nodes[-1] )



	def _p_onNodeAppended(self, node):
		self._graph.addNode( node.graphNode )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onFunctionGraphNodeAppended( self, node )
		self.nodeAddedSignal.emit( self, node )

	def _p_onRemoveNode(self, node):
		self.removeNodeSignal.emit( self, node )
		self._graph.removeNode( node.graphNode )

	def _p_onNodeRemoved(self, node):
		self.nodeRemovedSignal.emit( self, node )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onFunctionGraphNodeRemoved( self, node )



	def _getGraph(self):
		return self._graph

	def _getNodes(self):
		return self._nodes



	graph = property( _getGraph, doc='Graph' )
	nodes = property( _getNodes, doc='List of nodes' )


IOXml.ioObjectFactoryRegister( 'GSFunctionGraph', GSFunctionGraph )


