##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from gSculpt.FileIO import IOXml

from gSculpt.GSGraph.GSGraph import GSGraph

from gSculpt.View.BackgroundImage import BackgroundImageList
from gSculpt.View.ModelLayer import ModelLayerTable

from gSculpt.ProceduralCore.GSFunctionGraph import GSFunctionGraph
from gSculpt.ProceduralCore.GSSceneNode import GSSceneNode
from gSculpt.ProceduralCore import GSSceneCommandTracker





class GSScene (GSFunctionGraph):
	"""Scene - a gSculpt scene
	Contains the list of scene nodes, the graph which controls their connections, the layer table, and the background images table"""
	trackerClass = GSSceneCommandTracker.GSSceneCommandTracker


	class _NameToNodeTable (object):
	 	def __init__(self, scene):
			super( GSScene._NameToNodeTable, self ).__init__()
			self._nameToNode = {}
			self._nodeToName = {}
			self._nodeToResponseFunction = {}
			self._scene = scene


		def __iter__(self):
			return iter( self._nameToNode )

		def __len__(self):
			return len( self._nameToNode )

		def __contains__(self, node):
			return node in self._nameToNode

		def __getitem__(self, key):
			return self._nameToNode[key]

		def keys(self):
			return self._nameToNode.keys()

		def values(self):
			return self._nameToNode.values()

		def items(self):
			return self._nameToNode.items()


		def _f_onAddNode(self, node):
			def onNodeName(name):
				oldName = self._nodeToName[node]
				self._p_unmapNameNode( node, oldName )
				self._p_mapNameNode( node, name )
			self._p_mapNameNode( node, node.name )
			node.fields.name.valueSignal.connect( onNodeName )
			self._nodeToResponseFunction[node] = onNodeName

		def _f_onRemoveNode(self, node):
			oldName = self._nodeToName[node]
			self._p_unmapNameNode( node, oldName )
			onNodeName = self._nodeToResponseFunction[node]
			node.fields.name.valueSignal.disconnect( onNodeName )
			del self._nodeToResponseFunction[node]



		def _p_mapNameNode(self, node, name):
			try:
				nodeList = self._nameToNode[name]
			except KeyError:
				nodeList = []
				self._nameToNode[name] = nodeList
			nodeList.append( node )
			self._nodeToName[node] = name

		def _p_unmapNameNode(self, node, name):
			nodeList = self._nameToNode[name]
			nodeList.remove( node )
			if len( nodeList ) == 0:
				del self._nameToNode[name]
			del self._nodeToName[node]




	def __init__(self):
		super( GSScene, self ).__init__()

		self._nodesByName = GSScene._NameToNodeTable( self )
		self._layerTable = None
		self._backgroundImages = BackgroundImageList()
		self._layerTable = ModelLayerTable( self )



	def __readxml__(self, xmlNode):
		super( GSScene, self ).__readxml__( xmlNode )

		xmlNode.getChild( 'background_images' )  >>  self._backgroundImages
		xmlNode.getChild( 'layer_table' )  >>  self._layerTable


	def __writexml__(self, xmlNode):
		super( GSScene, self ).__writexml__( xmlNode )

		xmlNode.addChild( 'background_images' )  <<  self._backgroundImages
		xmlNode.addChild( 'layer_table' )  <<  self._layerTable



	def _p_onNodeAppended(self, node):
		self._nodesByName._f_onAddNode( node )
		super( GSScene, self )._p_onNodeAppended( node )

	def _p_onNodeRemoved(self, node):
		super( GSScene, self )._p_onNodeRemoved( node )
		self._nodesByName._f_onRemoveNode( node )



	def _getNodesByName(self):
		return self._nodesByName

	def _getLayerTable(self):
		return self._layerTable

	def _getBackgroundImages(self):
		return self._backgroundImages


	nodesByName = property( _getNodesByName, doc='Name to node mapping' )
	layerTable = property( _getLayerTable, doc='Layer table' )
	backgroundImages = property( _getBackgroundImages, doc='Background image list' )


IOXml.ioObjectFactoryRegister( 'GSScene', GSScene )




if __name__ == '__main__':
	class MySceneNode (GSSceneNode):
		in1 = NodeSinkSingleColumn()
		in2 = NodeSinkMultipleColumn()


	scene = GSScene()


	node1 = MySceneNode()
	scene.nodes.append( node1 )
	node2 = MySceneNode()
	scene.nodes.append( node2 )

	node1.in1.graphSink.addInput( node2.out.graphSource )


