##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from gSculpt.Util.SignalSlot import *

from gSculpt.FileIO import IOXml

from gSculpt.Kernel.DB import FieldColumn, CompositeColumn

from gSculpt.GSGraph import GSGraph
from gSculpt.GSGraph.GSGraph import *

from gSculpt.ProceduralCore.GSFunctionGraph import *
from gSculpt.ProceduralCore.GSPlan import GSPlan, GSPlanInitialProductSource
from gSculpt.ProceduralCore import GSSceneNodeCommandTracker







class GSSceneNodeSource (GSFunctionGraphNodeSource):
	"""Scene node source (output) pin"""
	productModifiedSignal = ClassSignal()



	def getResultProduct(self):
		"""Get the result product from this pin"""
		assert False, 'abstract'

	def getResultProductClass(self):
		"""Get the class of the result product from this pin"""
		assert False, 'abstract'

IOXml.ioObjectFactoryRegister( 'GSSceneNodeSource', GSSceneNodeSource )




class GSSceneNodeSourcePlanResult (GSSceneNodeSource):
	"""The scene node plan result source pin"""
	def getResultProduct(self):
		"""Get the result product from this pin"""
		return self._functionGraphNode.getResultProduct()

	def getResultProductClass(self):
		"""Get the class of the result product from this pin"""
		return self._functionGraphNode.getResultProductClass()

IOXml.ioObjectFactoryRegister( 'GSSceneNodeSourcePlanResult', GSSceneNodeSourcePlanResult )



class GSSceneNodeSinkBase (object):
	"""Scene node sink pin base class"""
	def __init__(self, sceneNode, name):
		super( GSSceneNodeSinkBase, self ).__init__( sceneNode, name )
		self._graphPin.addInputSignal.connect( self._onAddInput )
		self._graphPin.removeInputSignal.connect( self._onRemoveInput )


	def getInputProduct(self, index):
		"""Get the input product from the input indexed by @index"""
		if isinstance( index, slice ):
			return [ self._getResultProductFromSource( source )   for source in self[index] ]
		else:
			return self._getResultProductFromSource( self[index] )

	def inputProducts(self):
		"""The the input products from the inputs into this sink pin (returns a list)"""
		for source in self:
			yield self._getResultProductFromSource( source )

	def getInputProductClass(self, index):
		"""Get the class of the input product from the input indexed by @index"""
		if isinstance( index, slice ):
			return [ self._getResultProductClassFromSource( source )   for source in self[index] ]
		else:
			return self._getResultProductClassFromSource( self[index] )

	def inputProductClasses(self):
		"""The the class of the input products from the inputs into this sink pin (returns a list)"""
		for source in self:
			yield self._getResultProductClassFromSource( source )


	def _onAddInput(self, graphSink, graphSource, graphLink):
		self._functionGraphNode._p_onInputLinkModified()

	def _onRemoveInput(self, graphSink, graphSource, graphLink):
		self._functionGraphNode._p_onInputLinkModified()


	@staticmethod
	def _getResultProductFromSource(source):
		if source is not None:
			return source.getResultProduct()
		else:
			return None

	@staticmethod
	def _getResultProductClassFromSource(source):
		if source is not None:
			return source.getResultProductClass()
		else:
			return None



class GSSceneNodeSinkSingle (GSFunctionGraphNodeSinkSingle, GSSceneNodeSinkBase):
	"""Scene node sink that can take a single input"""
	pass

IOXml.ioObjectFactoryRegister( 'GSSceneNodeSinkSingle', GSSceneNodeSinkSingle )



class GSSceneNodeSinkMultiple (GSFunctionGraphNodeSinkMultiple, GSSceneNodeSinkBase):
	"""Scene node sink that can take multiple inputs"""
	pass

IOXml.ioObjectFactoryRegister( 'GSSceneNodeSinkMultiple', GSSceneNodeSinkMultiple )





class SceneNodeSinkSingleColumn (NodeSinkColumn):
	"""Node single input sink column"""
	def _o_createSink(self, instance, name):
		return GSSceneNodeSinkSingle( instance, name )


class SceneNodeSinkMultipleColumn (NodeSinkColumn):
	"""Node multiple input sink column"""
	def _o_createSink(self, instance, name):
		return GSSceneNodeSinkMultiple( instance, name )







class SceneNodeSourceColumn (NodeSourceColumn):
	def _o_createSource(self, instance, name):
		return GSSceneNodeSource( instance, name )







class GSSceneGSGraphNode (GSFunctionGraphGSGraphNode):
	"""GSGraph node for GSSceneNode"""
	nameSignal = ClassSignal()


	def __init__(self, functionGraphNode):
		super( GSSceneGSGraphNode, self ).__init__( functionGraphNode )
		self._name = functionGraphNode.name
		functionGraphNode.fields.name.valueSignal.connect( self._p_onName )


	def _p_onName(self, name):
		self._name = name
		self.nameSignal.emit( name )


	def _getName(self):
		return self._name

	name = property( _getName )





class GSSceneNode (GSFunctionGraphNode, GSPlanInitialProductSource):
	"""Scene Node base class"""
	class NodePlanResultSourceColumn (NodeSourceColumn):
		"""Scene node plan result source column"""
		def _o_createSource(self, instance, name):
			return GSSceneNodeSourcePlanResult( instance, name )


	trackerClass = GSSceneNodeCommandTracker.GSSceneNodeCommandTracker
	description = _( 'Abstract scene node' )

	out = NodePlanResultSourceColumn( 'out', _( 'Default output source' ) )

	name = FieldColumn( str, '' )
	bVisible = FieldColumn( bool, True )


	nodeViewSettingsAddedSignal = ClassSignal()
	removeNodeViewSettingsSignal = ClassSignal()
	nodeViewSettingsModifiedSignal = ClassSignal()


	def __init__(self, name = None, src=None):
		"""Constructor:
		@name - name of node
		@src - source object to copy"""
		GSFunctionGraphNode.__init__( self, src )
		GSPlanInitialProductSource.__init__( self )

		self._graphNode.inputSources.appendSignal.connect( self._p_onAddInputSource )
		self._graphNode.inputSources.removeSignal.connect( self._p_onRemoveInputSource )


		self._plan = GSPlan( self )

		self._initialProduct = None
		self._initialProductClass = None
		self._bInitialProductModified = True
		self._bInitialProductClassModified = True

		self._plan.productModifiedSignal.chainConnect( self.out.productModifiedSignal )
		self._plan.productModifiedSignal.connect( self._p_onPlanProductModified )


		self._resultProduct = None
		self._resultProductClass = None
		self._bResultProductModified = True
		self._bResultProductClassModified = True

		self._nodeViewSettings = []





	def getInitialProduct(self, errorHandler):
		"""Get the initial product (the product that is processed by the plan)"""
		if self._bInitialProductModified:
			self._initialProduct = self.computeInitialProduct( None )

			if self._initialProduct is not None:
				self._initialProductClass = type( self._initialProduct )
			else:
				self._initialProductClass = None

			self._bInitialProductModified = False
			self._bInitialProductClassModified = False
			unblockSlot( self._p_onInputModified )
		return self._initialProduct


	def getInitialProductClass(self):
		"""Get the class of the initial product (the product that is processed by the plan)"""
		if self._bInitialProductClassModified:
			self._initialProductClass = self.computeInitialProductClass()

			self._bInitialProductClassModified = False
			unblockSlot( self._p_onInputModified )
		return self._initialProductClass


	def computeInitialProduct(self, errorHandler):
		assert False, 'abstract'

	def computeInitialProductClass(self):
		assert False, 'abstract'




	def getResultProduct(self):
		"""Get the result product"""
		self._p_updateResult()
		return self._resultProduct


	def getResultProductClass(self):
		"""Get the class of the result product"""
		self._p_updateResultClass()
		return self._resultProductClass



	def getNodeViewSettingsForClass(self, settingsClass):
		"""Get the node view settings whose class is @settingsClass"""
		for settings in self._nodeViewSettings:
			if settings.__class__ is settingsClass:
				return settings
		return None

	def getCompatibleNodeViewSettings(self, settingsClass):
		"""Get a list of node view settings objects that are instances of @settingsClass"""
		return [ settings   for settings in self._nodeViewSettings   if isinstance( settings, settingsClass ) ]

	def getNodeViewSettings(self):
		"""Get a list of node view settings"""
		return self._nodeViewSettings

	def setNodeViewSettings(self, settings):
		"""Set node view settings; replaces and existing entry that is of the same class"""
		for s in self._nodeViewSettings:
			if s.__class__ is settings.__class__:
				self._p_removeNodeViewSettings( s )
		self._p_addNodeViewSettings( settings )



	def _p_addNodeViewSettings(self, settings):
		self._nodeViewSettings.append( settings )
		settings.fieldModifiedSignal.connect( self._p_onNodeViewSettingsFieldModified )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onSceneNodeViewSettingsAdded( self, settings )
		self.nodeViewSettingsAddedSignal.emit( self, settings )
		self.nodeViewSettingsModifiedSignal.emit( self )


	def _p_removeNodeViewSettings(self, settings):
		self.removeNodeViewSettingsSignal.emit( self, settings )
		settings.fieldModifiedSignal.disconnect( self._p_onNodeViewSettingsFieldModified )
		self._nodeViewSettings.remove( settings )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onSceneNodeViewSettingsRemoved( self, settings )




	def __readxml__(self, xmlNode):
		super( GSSceneNode, self ).__readxml__( xmlNode )

		xmlNode.getChild( 'plan' )  >>  self._plan

		viewSettingsListXmlNode = xmlNode.getChild( 'node_view_settings' )
		if viewSettingsListXmlNode.isValid():
			for settingsXmlNode  in viewSettingsListXmlNode.childrenNamed( 'view_settings' ):
				settings = settingsXmlNode.readObject()
				if settings is not None:
					self._nodeViewSettings.append( settings )
					settings.fieldModifiedSignal.connect( self._p_onNodeViewSettingsFieldModified )
					self.nodeViewSettingsAddedSignal.emit( self, settings )
		self.nodeViewSettingsModifiedSignal.emit( self )


	def __writexml__(self, xmlNode):
		super( GSSceneNode, self ).__writexml__( xmlNode )

		xmlNode.addChild( 'plan' )  <<  self._plan

		viewSettingsListXmlNode = xmlNode.addChild( 'node_view_settings' )
		for settings in self._nodeViewSettings:
			viewSettingsListXmlNode.addChild( 'view_settings' ).writeObject( settings )




	def _o_createGSGraphNode(self):
		return GSSceneGSGraphNode( self )


	def _p_updateResult(self):
		if self._bResultProductModified:
			if self._plan.isConsistent():
				self._resultProduct = self._plan.computeProduct()
				self._resultProductClass = type( self._resultProduct )
			else:
				self._resultProduct = None
				self._resultProductClass = None

			self._bResultProductModified = False
			self._bResultProductClassModified = False
			unblockSlot( self._p_onPlanProductModified )

	def _p_updateResultClass(self):
		if self._bResultProductClassModified:
			if self._plan.isConsistent():
				self._resultProductClass = self._plan.computeProductClass()
			else:
				self._resultProductClass = None

			self._bResultProductClassModified = False
			unblockSlot( self._p_onPlanProductModified )




	def _f_onFieldModify(self, fieldName):
		"Notify of field value change"
		super( GSSceneNode, self )._f_onFieldModify( fieldName )
		self._initialProduct = None
		self._initialProductClass = None
		self._bInitialProductModified = True
		self._bInitialProductClassModified = True
		self.initialProductModifiedSignal.emit()


	def _p_onPlanProductModified(self):
		"Notify of plan product change"
		self._resultProduct = None
		self._resultProductClass = None
		self._bResultProductModified = True
		self._bResultProductClassModified = True
		blockSlot( self._p_onPlanProductModified )



	def _p_onInputModified(self):
		"Private - Notify of input product change"
		# Don't need to emit @self.initialProductModifiedSignal as this is done by a chain connection
		self._initialProduct = None
		self._initialProductClass = None
		self._bInitialProductModified = True
		self._bInitialProductClassModified = True
		blockSlot( self._p_onInputModified )


	def _p_onAddInputSource(self, graphSource):
		source = graphSource._functionGraphNodePin
		source.productModifiedSignal.connect( self._p_onInputModified )
		source.productModifiedSignal.chainConnect( self.initialProductModifiedSignal )
		self._p_onInputModified()
		self.initialProductModifiedSignal.emit()


	def _p_onRemoveInputSource(self, graphSource):
		source = graphSource._functionGraphNodePin
		source.productModifiedSignal.disconnect( self._p_onInputModified )
		source.productModifiedSignal.chainDisconnect( self.initialProductModifiedSignal )
		self._p_onInputModified()
		self.initialProductModifiedSignal.emit()


	def _p_onInputLinkModified(self):
		self._p_onInputModified()
		self.initialProductModifiedSignal.emit()


	def _p_onNodeViewSettingsFieldModified(self, viewSettings, fieldName):
		self.nodeViewSettingsModifiedSignal.emit( self )




	def _getPlan(self):
		return self._plan


	plan = property( _getPlan, doc='node plan' )




