##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import deepcopy

from gSculpt.Util.SignalSlot import ClassSignal, blockSlot, unblockSlot

from gSculpt.Painter import Painter
from gSculpt.Painter.RequestPainter import RequestPainterObjectSource
from gSculpt.Painter.RequestViewportPainter import RequestViewportPainter

from gSculpt.Kernel.CachedObjectSource import CachedObjectSource

from gSculpt.ProceduralCore.GSProcedure import GSProcedure

from gSculpt.ProceduralTool.GSFunctionTool import GSFunctionTool, FunctionToolProductView

from gSculpt.Kernel.DBEdit import ColumnEditor

from gSculpt.Manipulator.ColumnManipulators import ColumnManipulator










class GSProcedureToolColumnManipulator (ColumnManipulator):
	"UI element factory interface"
	def attachInputProductSource(self, element, source):
		"""Attach the input product source
		@element - the element
		@source - the product source"""
		pass

	def detachInputProductSource(self, element, source):
		"""Attach the input product source
		@element - the element
		@source - the product source"""
		pass


	def attachOutputProductSource(self, element, source):
		"""Attach the displayed product source
		@element - the element
		@source - the product source"""
		pass

	def detachOutputProductSource(self, element, source):
		"""Attach the displayed product source
		@element - the element
		@source - the product source"""
		pass




class PlanProcedureInputProductSource (CachedObjectSource):
	def __init__(self):
		super( PlanProcedureInputProductSource, self ).__init__()
		self._plan = None
		self._proc = None



	def attachPlanAndProcedure(self, plan, proc):
		assert self._plan is None, 'plan and procedure already attached'
		self._plan = plan
		self._proc = proc
		self._plan.invalidateProductSignal.connect( self._p_onInvalidateProduct )


	def detachPlanAndProcedure(self):
		assert self._plan is not None, 'no plan and procedure attached'
		self._plan.invalidateProductSignal.disconnect( self._p_onInvalidateProduct )
		self._plan = None
		self._proc = None



	def _p_evaluateSource(self):
		result = None

		if self._proc is not None:
			procIndex = self._plan.index( self._proc )

			inputProduct = self._plan.computePartialProduct( procIndex - 1 )

			if inputProduct is not None:
				# Check types
				if self._proc.canOperateOn( inputProduct ):
					result = inputProduct

		unblockSlot( self._p_onInvalidateProduct )

		return result


	def _p_onInvalidateProduct(self, plan, index):
		assert plan is self._plan, '@plan not attached to this'

		if self._p_isCached():
			# May need to invalidate the cache
			if self._proc is None:
				self._p_invalidateCache()
			else:
				procIndex = self._plan.index( self._proc )

				if index <= ( procIndex - 1 ):
					blockSlot( self._p_onInvalidateProduct )
					self._p_invalidateCache()





class ProcedureToolProductView (FunctionToolProductView):
	def __init__(self):
		super( ProcedureToolProductView, self ).__init__()

		self._inputProductSource = PlanProcedureInputProductSource()
		self._view = None
		self._plan = None
		self._proc = None



	def initialise(self, view, plan, proc, tool):
		self._view = view
		self._plan = plan
		self._proc = proc
		self._inputProductSource.attachPlanAndProcedure( plan, proc )


	def shutdown(self, tool):
		self._inputProductSource.detachPlanAndProcedure()
		self._view = None
		self._plan = None
		self._proc = None



	def _p_getInputProductSource(self):
		return self._inputProductSource

	inputProductSource = property( _p_getInputProductSource )





class BasicProcedureToolProductView (ProcedureToolProductView, RequestPainterObjectSource):
	class ProcInvokeProductSource (CachedObjectSource):
		def __init__(self):
			super( BasicProcedureToolProductView.ProcInvokeProductSource, self ).__init__()
			self._inputSource = None
			self._proc = None
			self._adjustableInvoker = None
			self._bInputObjectNeedsRefresh = False
			self.errorHandler = None



		def attachInputSourceAndProcedure(self, inputSource, proc, tool):
			assert self._inputSource is None
			self._inputSource = inputSource
			self._proc = proc
			self._inputSource.changedSignal.connect( self._p_onInputValueChanged )
			self._proc.fieldModifiedSignal.connect( self._p_onFieldModified )
			adjustableInvokerClass = tool._f_getAdjustableInvokerClass()
			if adjustableInvokerClass is not None:
				self._adjustableInvoker = adjustableInvokerClass( proc, inputSource.getValue(), self.errorHandler )
				self._adjustableInvoker.attach()
				tool._f_setAdjustableInvoker( self._adjustableInvoker )

		def detachInputSourceAndProcedure(self, tool):
			assert self._inputSource is not None
			self._inputSource.changedSignal.disconnect( self._p_onInputValueChanged )
			self._proc.fieldModifiedSignal.disconnect( self._p_onFieldModified )
			self._inputSource = None
			self._proc = None
			if self._adjustableInvoker is not None:
				self._adjustableInvoker.detach()
				self._adjustableInvoker = None
				tool._f_setAdjustableInvoker( None )


		def _p_evaluateSource(self):
			assert self._inputSource is not None
			inputProduct = self._inputSource.getValue()

			unblockSlot( self._p_onFieldModified )
			unblockSlot( self._p_onInputValueChanged )

			if inputProduct is not None:
				if self._adjustableInvoker is not None:
					if self._bInputObjectNeedsRefresh:
						self._adjustableInvoker.setInputObject( deepcopy( inputProduct ) )
						self._bInputObjectNeedsRefresh = False

					return self._adjustableInvoker.requestResult()
				else:
					return self._proc.invoke( deepcopy( inputProduct ), self.errorHandler )
			else:
				return None


		def _p_onFieldModified(self, func, fieldName):
			assert func is self._proc
			if self._adjustableInvoker is not None:
				self._adjustableInvoker.onFunctionArgumentModify( fieldName )
			blockSlot( self._p_onFieldModified )
			self._p_invalidateCache()


		def _p_onInputValueChanged(self):
			if self._adjustableInvoker is not None:
				self._bInputObjectNeedsRefresh = True
			blockSlot( self._p_onInputValueChanged )
			self._p_invalidateCache()



	def __init__(self, painterFactory):
		super( BasicProcedureToolProductView, self ).__init__()
		self._procSource = BasicProcedureToolProductView.ProcInvokeProductSource()
		self._painter = RequestViewportPainter( painterFactory, False )



	def initialise(self, view, plan, proc, tool):
		super( BasicProcedureToolProductView, self ).initialise( view, plan, proc, tool )

		self._painter.painterObjectSource = self
		self._procSource.attachInputSourceAndProcedure( self._inputProductSource, proc, tool )
		self._procSource.changedSignal.connect( self._p_onProcSourceValueChanged )


	def shutdown(self, tool):
		super( BasicProcedureToolProductView, self ).shutdown( tool )

		self._painter.painterObjectSource = None
		self._procSource.detachInputSourceAndProcedure( tool )
		self._procSource.changedSignal.disconnect( self._p_onProcSourceValueChanged )



	def installPainters(self, view):
		self._painter.addToView( view )


	def uninstallPainters(self, view):
		self._painter.removeFromView( view )


	def requestObjectToPaint(self):
		return self.objectToPaint


	def _p_onProcSourceValueChanged(self):
		self._view.postRedraw()



	def _p_setErrorHandler(self, errorHandler):
		self._procSource.errorHandler = errorHandler

	def _p_getOutputProductSource(self):
		return self._procSource

	def _p_getDisplayedProduct(self):
		return self._procSource.getValue()

	def _p_getObjectToPaint(self):
		return self._procSource.getValue()

	errorHandler = property( None, _p_setErrorHandler )
	outputProductSource = property( _p_getOutputProductSource )
	displayedProduct = property( _p_getDisplayedProduct )
	objectToPaint = property( _p_getObjectToPaint )









class GSProcedureTool (GSFunctionTool):
	functionClass = GSProcedure

	def __init__(self):
		super( GSProcedureTool, self ).__init__()
		self._plan = None


	def attachFunction(self, func, view, editorSettings, parentWindow, commandHistory, painterFactory):
		"""Attach a GSFunction object to this tool
		@func - the function object to attach"""
		super( GSProcedureTool, self ).attachFunction( func, view, editorSettings, parentWindow, commandHistory, painterFactory )
		self._plan = func.r_plan

	def detachFunction(self):
		"""Detach the currently attach GSFunction object"""
		super( GSProcedureTool, self ).detachFunction()
		self._plan = None




	def initialise(self):
		for columnEditor in self._o_getColumnEditors():
			if isinstance( columnEditor, GSProcedureToolColumnManipulator ):
				element = self._p_getUIElementForColumnEditor( columnEditor )
				columnEditor.attachInputProductSource( element, self._productView.inputProductSource )
				columnEditor.attachOutputProductSource( element, self._productView.outputProductSource )

		super( GSProcedureTool, self ).initialise()


	def shutdown(self):
		super( GSProcedureTool, self ).shutdown()

		for columnEditor in self._o_getColumnEditors():
			if isinstance( columnEditor, GSProcedureToolColumnManipulator ):
				element = self._p_getUIElementForColumnEditor( columnEditor )
				columnEditor.detachOutputProductSource( element, self._productView.outputProductSource )
				columnEditor.detachInputProductSource( element, self._productView.inputProductSource )





	def _p_createProductView(self):
		v = BasicProcedureToolProductView( self.painterFactory )
		v.initialise( self.view, self.plan, self.procedure, self )
		return v


	def _p_shutdownProductView(self, productView):
		productView.shutdown( self )




	def _p_getPlan(self):
		return self._plan

	def _p_getProcedure(self):
		return self._func

	def _p_getInputProduct(self):
		return self._productView.inputProductSource.getValue()

	plan = property( _p_getPlan )
	procedure = property( _p_getProcedure )
	inputProduct = property( _p_getInputProduct )






if __name__ == '__main__':
	from gSculpt.Kernel.DB import FieldColumn
	from gSculpt.ProceduralCore.GSProcedure import GSProcedure

	from gSculpt.Math import Math
	from gSculpt.Math.Math import Vector3
	from gSculpt.Math import MathVariable

	from gSculpt.UI.FormLayout import FormLayoutFactory

	from gSculpt.Kernel.ColumnEditors import *
	from gSculpt.ProceduralTool.GSFunctionTool import createFunctionTool

	from gSculpt.Painter.ObjectPainterFactory import ObjectPainterFactory


	# Make the procedure class
	class TestProcedure (GSProcedure):
		segments = FieldColumn( int, 0 )
		extrusionVector = FieldColumn( Vector3, Vector3() )


	class TestTool (GSProcedureTool):
		functionClass = TestProcedure
		pageTitle = 'Test'
		titleText = 'Test'
		segmentsEdit = IntColumnEditEntry( TestProcedure.segments, 0, 1048576, 1, 10 )
		extrusionVectorEdit = Vector3ColumnEditEntry( TestProcedure.extrusionVector, 0.0, 1.0e16, 0.1, 1.0, 4 )

		layout = FormLayoutFactory()
		layout.row()  <<  'Segments:'  <<  segmentsEdit
		layout.row()
		layout.row()  <<  'Extrusion:'
		layout.row()  <<  'X:'  <<  extrusionVectorEdit.x
		layout.row()  <<  'Y:'  <<  extrusionVectorEdit.y
		layout.row()  <<  'Z:'  <<  extrusionVectorEdit.z
		layout.row()
		layout.finalise()




	# SECOND CLASS

	class TestProcedure2 (TestProcedure):
		expansion = FieldColumn( float, 0.0 )


	class TestTool2 (TestTool):
		functionClass = TestProcedure2
		pageTitle = 'Test2'
		titleText = 'Test2'
		expansionEdit = FloatColumnEditEntry( TestProcedure2.expansion, -1.0e16, 1.0e16, 0.1, 1.0, 4 )

		layout = FormLayoutFactory()
		layout.row()  <<  'Expansion:'  <<  expansionEdit
		layout.row()
		layout.finalise()





	class ProcToolTest (object):
		destroySignal = ClassSignal()


		def delete_event(self, widget, event, data=None):
			print( "delete_event" )
			return False

		def destroy(self, widget, data=None):
			self.tool.shutdown()
			self.tool.destroyUI()
			self.destroySignal.emit( self )


		def __init__(self, proc, painterFactory):
			self.window = gtk.Window( gtk.WINDOW_TOPLEVEL );
			self.window.connect( 'delete-event', self.delete_event )
			self.window.connect( 'destroy', self.destroy )
			self.window.set_border_width( 10 )


			self.tool = createFunctionTool( proc )
			self.tool.attachFunction( proc, None, None, self.window, None, painterFactory )
			self.tool.createUI()
			self.tool.initialise()

			w = self.tool.getWidget()
			w.show()
			w.set_size_request( -1, 400 )

			self.window.add( w );
			self.window.show();


	class ToolMonitor (object):
		def __init__(self, tools):
			self._tools = set( tools )
			for t in tools:
				t.destroySignal.connect( self.onToolDestroy )


		def onToolDestroy(self, tool):
			self._tools.remove( tool )

			if len( self._tools )  ==  0:
				gtk.main_quit()




	proc1 = TestProcedure()
	proc2 = TestProcedure2()

	painterFactory = ObjectPainterFactory()

	tool1 = ProcToolTest( proc1, painterFactory )
	tool2 = ProcToolTest( proc2, painterFactory )
	tool2b = ProcToolTest( proc2, painterFactory )

	mon = ToolMonitor( [ tool1, tool2, tool2b ] )



	gtk.main()


