##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import deepcopy

from gSculpt.Util.SignalSlot import ClassSignal, blockSlot, unblockSlot

from gSculpt.Painter import Painter
from gSculpt.Painter.RequestPainter import RequestPainterObjectSource
from gSculpt.Painter.RequestViewportPainter import RequestViewportPainter

from gSculpt.Kernel.CachedObjectSource import CachedObjectSource

from gSculpt.ProceduralCore.GSSceneNode import GSSceneNode

from gSculpt.ProceduralTool.GSFunctionTool import GSFunctionTool, FunctionToolProductView






class SceneNodeInitialProductSource (CachedObjectSource):
	def __init__(self):
		super( SceneNodeInitialProductSource, self ).__init__()
		self._sceneNode = None
		self.errorHandler = None



	def attachSceneNode(self, sceneNode):
		assert self._sceneNode is None, 'node already attached'
		self._sceneNode = sceneNode
		self._sceneNode.initialProductModifiedSignal.connect( self._p_onInitialProductModified )


	def detachSceneNode(self):
		assert self._sceneNode is not None, 'no node attached'
		self._sceneNode.initialProductModifiedSignal.disconnect( self._p_onInitialProductModified )
		self._sceneNode = None



	def _p_evaluateSource(self):
		assert self._sceneNode is not None, 'no node attached'
		result = None

		if self._sceneNode is not None:
			result = self._sceneNode.getInitialProduct( self.errorHandler )

		unblockSlot( self._p_onInitialProductModified )

		return result


	def _p_onInitialProductModified(self):
		assert self._sceneNode is not None, 'no node attached'

		if self._p_isCached():
			blockSlot( self._p_onInitialProductModified )
			self._p_invalidateCache()




class SceneNodeToolInitialProductView (FunctionToolProductView, RequestPainterObjectSource):
	def __init__(self, painterFactory):
		super( SceneNodeToolInitialProductView, self ).__init__()

		self._initialProductSource = SceneNodeInitialProductSource()
		self._painter = RequestViewportPainter( painterFactory, False )

		self._view = None
		self._sceneNode = None



	def initialise(self, view, sceneNode):
		self._view = view
		self._sceneNode = sceneNode

		self._painter.painterObjectSource = self
		self._initialProductSource.attachSceneNode( sceneNode )
		self._initialProductSource.changedSignal.connect( self._p_onInitialProductValueChanged )


	def shutdown(self):
		self._initialProductSource.changedSignal.disconnect( self._p_onInitialProductValueChanged )
		self._initialProductSource.detachSceneNode()
		self._view = None
		self._sceneNode = None

		self._painter.painterObjectSource = None



	def installPainters(self, view):
		self._painter.addToView( view )

	def uninstallPainters(self, view):
		self._painter.removeFromView( view )


	def requestObjectToPaint(self):
		return self.objectToPaint


	def _p_onInitialProductValueChanged(self):
		self._view.postRedraw()



	def _p_setErrorHandler(self, errorHandler):
		self._initialProductSource.errorHandler = errorHandler

	def _p_getInitialProductSource(self):
		return self._initialProductSource

	def _p_getDisplayedProduct(self):
		return self._initialProductSource.getValue()

	def _p_getObjectToPaint(self):
		return self._initialProductSource.getValue()

	errorHandler = property( None, _p_setErrorHandler )
	initialProductSource = property( _p_getInitialProductSource )
	outputProductSource = property( _p_getInitialProductSource )
	displayedProduct = property( _p_getDisplayedProduct )
	objectToPaint = property( _p_getObjectToPaint )






class GSSceneNodeTool (GSFunctionTool):
	functionClass = GSSceneNode

	def _p_createProductView(self):
		v = SceneNodeToolInitialProductView( self.painterFactory )
		v.initialise( self.view, self.sceneNode )
		return v


	def _p_shutdownProductView(self, productView):
		productView.shutdown()


	def _p_getSceneNode(self):
		return self._func


	sceneNode = property( _p_getSceneNode )






if __name__ == '__main__':
	from gSculpt.Kernel.DB import FieldColumn
	from gSculpt.ProceduralCore.GSSceneNode import GSSceneNode, GSSceneGraph, SinkSingleDecl, SinkMultipleDecl

	from gSculpt.Util.SignalSlot import Signal

	from gSculpt.Math import Math
	from gSculpt.Math.Math import Vector3
	from gSculpt.Math import MathVariable

	from gSculpt.UI.FormLayout import FormLayoutFactory

	from gSculpt.Kernel.ColumnEditors import *
	from gSculpt.ProceduralTool.GSFunctionTool import createFunctionTool

	from gSculpt.Painter.ObjectPainterFactory import ObjectPainterFactory


	# Make the procedure class
	class TestSceneNode (GSSceneNode):
		segments = FieldColumn( int, 0 )
		extrusionVector = FieldColumn( Vector3, Vector3() )
		in1 = SinkSingleDecl()



	class TestTool (GSSceneNodeTool):
		functionClass = TestSceneNode
		pageTitle = 'Test'
		titleText = 'Test'
		segmentsEdit = IntColumnEditEntry( TestSceneNode.segments, 0, 1048576, 1, 10 )
		extrusionVectorEdit = Vector3ColumnEditEntry( TestSceneNode.extrusionVector, 0.0, 1.0e16, 0.1, 1.0, 4 )

		layout = FormLayoutFactory()
		layout.row()  <<  'Segments:'  <<  segmentsEdit
		layout.row()
		layout.row()  <<  'Extrusion:'
		layout.row()  <<  'X:'  <<  extrusionVectorEdit.x
		layout.row()  <<  'Y:'  <<  extrusionVectorEdit.y
		layout.row()  <<  'Z:'  <<  extrusionVectorEdit.z
		layout.row()
		layout.finalise()




	# SECOND CLASS

	class TestSceneNode2 (TestSceneNode):
		expansion = FieldColumn( float, 0.0 )
		in2 = SinkMultipleDecl()


	class TestTool2 (TestTool):
		functionClass = TestSceneNode2
		pageTitle = 'Test2'
		titleText = 'Test2'
		expansionEdit = FloatColumnEditEntry( TestSceneNode2.expansion, -1.0e16, 1.0e16, 0.1, 1.0, 4 )

		layout = FormLayoutFactory()
		layout.row()  <<  'Expansion:'  <<  expansionEdit
		layout.row()
		layout.finalise()





	class NodeToolTest (object):
		destroySignal = ClassSignal()


		def delete_event(self, widget, event, data=None):
			print( "delete_event" )
			return False

		def destroy(self, widget, data=None):
			self.tool.shutdown()
			self.tool.destroyUI()
			self.destroySignal.emit( self )


		def __init__(self, proc, painterFactory):
			self.window = gtk.Window( gtk.WINDOW_TOPLEVEL );
			self.window.connect( 'delete-event', self.delete_event )
			self.window.connect( 'destroy', self.destroy )
			self.window.set_border_width( 10 )


			self.tool = createFunctionTool( proc )
			self.tool.attachFunction( proc, None, None, self.window, None, painterFactory )
			self.tool.createUI()
			self.tool.initialise()

			w = self.tool.getWidget()
			w.show()
			w.set_size_request( -1, 400 )

			self.window.add( w );
			self.window.show();


	class ToolMonitor (object):
		def __init__(self, tools):
			self._tools = set( tools )
			for t in tools:
				t.destroySignal.connect( self.onToolDestroy )


		def onToolDestroy(self, tool):
			self._tools.remove( tool )

			if len( self._tools )  ==  0:
				gtk.main_quit()




	proc1 = TestSceneNode()
	proc2 = TestSceneNode2()

	painterFactory = ObjectPainterFactory()

	tool1 = NodeToolTest( proc1, painterFactory )
	tool2 = NodeToolTest( proc2, painterFactory )
	tool2b = NodeToolTest( proc2, painterFactory )

	mon = ToolMonitor( [ tool1, tool2, tool2b ] )



	gtk.main()


