##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import deepcopy

from gSculpt.Math.Math import Quaternion, Matrix4

from gSculpt.FileIO.IOXml import IOXmlEnum, ioObjectFactoryRegisterPrimitive

from gSculpt.Kernel.DB import *
from gSculpt.Kernel.Enum import Enum

from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool.GSProcedureTool import GSProcedureTool
from gSculpt.ProceduralTool.GSProcedureAdjustableInvoker import *
from gSculpt.Kernel.ColumnEditors import *

from gSculpt.Model import Model
from gSculpt.Model.Model import GSProductModel




# TransformationTarget enumeration
class TransformationTarget (Enum):
	BODY = 0
	POINTS = 1




# XML IO for TransformationTarget

xformTargetIOXmlEnum = IOXmlEnum( TransformationTarget, 'TransformationTarget', TransformationTarget.BODY )
xformTargetIOXmlEnum.value( TransformationTarget.BODY, 'body' )
xformTargetIOXmlEnum.value( TransformationTarget.POINTS, 'points' )

ioXmlReadTransformationTargetProp = xformTargetIOXmlEnum.propReader
ioXmlReadTransformationTargetNode = xformTargetIOXmlEnum.nodeReader
ioXmlWriteTransformationTargetProp = xformTargetIOXmlEnum.propWriter
ioXmlWriteTransformationTargetNode = xformTargetIOXmlEnum.nodeWriter

ioObjectFactoryRegisterPrimitive( 'TransformationTarget', TransformationTarget )



# Field for TransformationTarget

class TransformationTargetField (Field):
	valueClass = TransformationTarget
	defaultValue = TransformationTarget.BODY
	xmlNodeValueReader = ioXmlReadTransformationTargetNode
	xmlNodeValueWriter = ioXmlWriteTransformationTargetNode

registerFieldType( TransformationTarget, TransformationTargetField )




class ProcTransformationBase (GSProcedure):
	description = _( 'Abstract transformation base' )

	target = FieldColumn( TransformationTarget, TransformationTarget.BODY )

	inputClass = GSProductModel


	@staticmethod
	def transform(inputObject, matrixOrTransformation, target):
		inputObject.transform( matrixOrTransformation, target == TransformationTarget.BODY, False, target == TransformationTarget.POINTS )

	@staticmethod
	def transformAboutCentre(inputObject, matrix, target):
		centre = inputObject.getBBox( target == TransformationTarget.POINTS ).getCentre()
		matrix = Matrix4.translate( -centre.toVector3() )  *  matrix  *  Matrix4.translate( centre.toVector3() )
		inputObject.transform( matrix, target == TransformationTarget.BODY, False, target == TransformationTarget.POINTS )




class TransformationBaseAdjustableInvoker (GSProcedureAdjustableInvoker):
	functionClass = ProcTransformationBase

	def main(self):
		self.model = deepcopy( self._inputObject )


	def restore(self):
		self.model.transformRestore()


	def getResult(self):
		return self.model


	def transform(self, matrixOrTransformation, target):
		self.model.transform( matrixOrTransformation, target == TransformationTarget.BODY, True, target == TransformationTarget.POINTS )

	def transformAboutCentre(self, matrix, target):
		centre = self.model.getBBox( target == TransformationTarget.POINTS ).getCentre()
		matrix = Matrix4.translate( -centre.toVector3() )  *  matrix  *  Matrix4.translate( centre.toVector3() )
		self.model.transform( matrix, target == TransformationTarget.BODY, True, target == TransformationTarget.POINTS )




class ToolTransformationBase (GSProcedureTool):
	row = RowRefColumn( ProcTransformationBase )

	functionClass = ProcTransformationBase
	pageTitle = _( 'Transformation' )
	titleText = _( 'Abstract Transformation' )
	adjustableInvoker = TransformationBaseAdjustableInvoker


	row_target = ObjectSourceProxyColumn( row.target )

	def _p_focalPoint(self):
		return self.inputProduct.getBBox( self.row_target == TransformationTarget.POINTS ).getCentre()

	focalPoint = FunctionColumn( _p_focalPoint )

