##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import deepcopy

import math

from gSculpt.Math.Math import Vector3, Matrix4, Vector2, Point3

from gSculpt.UI.FormLayout import FormLayoutFactory
from gSculpt.UI.UIAction import UIAction

from gSculpt.Kernel.DB import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool.GSProcedureTool import GSProcedureTool
from gSculpt.Kernel.ColumnEditors import *

from gSculpt.Manipulator.ColumnManipulators import *
from gSculpt.Manipulator.RealManipulator import *

from gSculpt.TransformationTools.ToolTransformation import *
from gSculpt.TransformationTools.ToolTransformationBase import *
from gSculpt.TransformationTools.ToolRotatedTransformation import *
from gSculpt.TransformationTools import TransformationToolActions
from gSculpt.Transformation.Transformation import ShrinkTransformation, MatrixTransformation




def makeTransformation(model, target, xformRotation, shrink, anchor, lower, upper):
	centre = model.getBBox( target == TransformationTarget.POINTS ).getCentre()
	lower *= 0.01
	upper *= 0.01
	anchor *= 0.01

	rotationMatrix = Matrix4.rotate( xformRotation )
	modelToXformFrameMatrix = Matrix4.translate( -centre.toVector3() )  *  rotationMatrix.rigidBodyInverse()
	xformToModelFrameMatrix = modelToXformFrameMatrix.rigidBodyInverse()

	box = model.getBBox( modelToXformFrameMatrix, target == TransformationTarget.POINTS )
	lz = box.getLower().z
	uz = box.getUpper().z
	lowerZ = lz  +  ( uz - lz )  *  lower
	upperZ = lz  +  ( uz - lz )  *  upper
	anchorZ = lz  +  ( uz - lz )  *  anchor

	x0 = MatrixTransformation( modelToXformFrameMatrix )
	x1 = ShrinkTransformation( lowerZ, upperZ, anchorZ, shrink.x, shrink.y )
	x2 = MatrixTransformation( xformToModelFrameMatrix )
	x1.chainTo( x2 )
	x0.chainTo( x1 )

	return x0




class ProcTransformationShrink (ProcRotatedTransformation):
	description = _( 'Shrink' )

	shrink = FieldColumn( Vector2, Vector2( 1.0, 1.0 ) )
	anchor = FieldColumn( float, 50.0 )
	lower = FieldColumn( float, 0.0 )
	upper = FieldColumn( float, 100.0 )


	def procedureInvoke(self, inputObject, errorHandler):
		x0 = makeTransformation( inputObject, self.target, self.xformRotation, self.shrink, self.anchor, self.lower, self.upper )
		ProcTransformationShrink.transform( inputObject, x0, self.target )
		return inputObject




class TransformationShrinkAdjustableInvoker (RotatedTransformationAdjustableInvoker):
	functionClass = ProcTransformationShrink

	def adjust(self):
		x0 = makeTransformation( self.model, self._proc.target, self._proc.xformRotation, self._proc.shrink, self._proc.anchor, self._proc.lower, self._proc.upper )
		self.transform( x0, self._proc.target )




class ToolTransformationShrink (ToolRotatedTransformation):
	row = RowRefColumn( ProcTransformationShrink )

	functionClass = ProcTransformationShrink
	pageTitle = _( 'Shrink' )
	titleText = _( 'Shrink' )
	adjustableInvoker = TransformationShrinkAdjustableInvoker


	row_xformRotation = ObjectSourceProxyColumn( row.xformRotation )
	row_anchor = ObjectSourceProxyColumn( row.anchor )

	def _p_shrinkFocalPoint(self):
		centre = self.inputProduct.getBBox( self.row_target == TransformationTarget.POINTS ).getCentre()

		rotationMatrix = Matrix4.rotate( self.row_xformRotation )
		modelToXformFrameMatrix = Matrix4.translate( -centre.toVector3() )  *  rotationMatrix.rigidBodyInverse()
		xformToModelFrameMatrix = modelToXformFrameMatrix.rigidBodyInverse()

		box = self.inputProduct.getBBox( modelToXformFrameMatrix, self.row_target == TransformationTarget.POINTS )
		lz = box.getLower().z
		uz = box.getUpper().z
		anchorZ = lz  +  ( uz - lz ) * self.row_anchor * 0.01
		return Point3( 0.0, 0.0, anchorZ )  *  xformToModelFrameMatrix

	shrinkFocalPoint = FunctionColumn( _p_shrinkFocalPoint )



	shrinkManip = Scale2ColumnManipulator3d( row.shrink, shrinkFocalPoint, row.xformRotation,
									snapping=ColumnManipulatorSnappingFraction() )

	rotationManip = RotationColumnManipulator( row.xformRotation, ToolRotatedTransformation.focalPoint, snapping=ColumnManipulatorSnappingAngle() )

	manipulatorSwitch = ColumnManipulatorSwitch( [ shrinkManip, rotationManip ] )
	manipulatorSwitch.enable()

	manipulatorSelector = ColumnManipulatorSwitchRadioGroup( manipulatorSwitch )
	shrinkButton = manipulatorSelector.addChoice( shrinkManip, TransformationToolActions.deformAction )
	rotationButton = manipulatorSelector.addChoice( rotationManip, TransformationToolActions.rotationAction )


	rotationEntry = QuaternionColumnEditEntry( row.xformRotation )
	rotationPopup = QuaternionColumnEditDirectionPopup( row.xformRotation, _( 'Set rotation' ) )

	anchorEntry = FloatColumnEditEntry( row.anchor, 0.0, 100.0, 0.01, 1.0, 4 )
	anchorSlider = FloatColumnEditSlider( row.anchor, 0.0, 100.0, 0.01, 1.0 )

	shrinkEntry = Vector2ColumnEditEntry( row.shrink, -1.0e8, 1.0e8, 0.01, 0.1, 4 )

	lowerEntry = FloatColumnEditEntry( row.lower, 0.0, 100.0, 0.01, 1.0, 4 )
	upperEntry = FloatColumnEditEntry( row.upper, 0.0, 100.0, 0.01, 1.0, 4 )

	lowerSlider = FloatColumnEditSlider( row.lower, 0.0, 100.0, 0.01, 1.0 )
	upperSlider = FloatColumnEditSlider( row.upper, 0.0, 100.0, 0.01, 1.0 )


	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Mouse function:' )
	layout.row()  <<  shrinkButton
	layout.row()  <<  rotationButton
	layout.row()
	layout.row()  <<  _( 'Rotation:' )
	layout.row()  <<  rotationPopup.popup
	layout.row()  <<  'X:'  <<  rotationEntry.x
	layout.row()  <<  'Y:'  <<  rotationEntry.y
	layout.row()  <<  'Z:'  <<  rotationEntry.z
	layout.row()
	layout.row()  <<  _( 'Anchor:' )  <<  anchorEntry.entry
	layout.row()  <<  anchorSlider.slider
	layout.row()
	layout.row()  <<  _( 'Shrink:' )
	layout.row()  <<  'U:'  <<  shrinkEntry.x
	layout.row()  <<  'V:'  <<  shrinkEntry.y
	layout.row()
	layout.row()  <<  _( 'Lower (%%):' )  <<  lowerEntry.entry
	layout.row()  <<  lowerSlider.slider
	layout.row()  <<  _( 'Upper (%%):' )  <<  upperEntry.entry
	layout.row()  <<  upperSlider.slider
	layout.row()
	layout.finalise()

