##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from copy import deepcopy

import math

from gSculpt.Math.Math import Vector3, Matrix4, Vector2, Point3

from gSculpt.UI.FormLayout import FormLayoutFactory
from gSculpt.UI.UIAction import UIAction

from gSculpt.Kernel.DB import *
from gSculpt.ProceduralCore.GSProcedure import *

from gSculpt.ProceduralTool.GSProcedureTool import GSProcedureTool
from gSculpt.Kernel.ColumnEditors import *

from gSculpt.Manipulator.ColumnManipulators import *
from gSculpt.Manipulator.RealManipulator import *

from gSculpt.TransformationTools.ToolTransformation import *
from gSculpt.TransformationTools.ToolTransformationBase import *
from gSculpt.TransformationTools.ToolRotatedTransformation import *
from gSculpt.TransformationTools import TransformationToolActions
from gSculpt.Transformation.Transformation import TaperTransformation, MatrixTransformation




def makeTransformation(model, target, xformRotation, taper, lower, upper):
	centre = model.getBBox( target == TransformationTarget.POINTS ).getCentre()
	lower *= 0.01
	upper *= 0.01

	rotationMatrix = Matrix4.rotate( xformRotation )
	modelToXformFrameMatrix = Matrix4.translate( -centre.toVector3() )  *  rotationMatrix.rigidBodyInverse()
	xformToModelFrameMatrix = modelToXformFrameMatrix.rigidBodyInverse()

	box = model.getBBox( modelToXformFrameMatrix, target == TransformationTarget.POINTS )
	lz = box.getLower().z
	uz = box.getUpper().z
	lowerZ = lz  +  ( uz - lz )  *  lower
	upperZ = lz  +  ( uz - lz )  *  upper

	x0 = MatrixTransformation( modelToXformFrameMatrix )
	x1 = TaperTransformation( lowerZ, upperZ, taper.x, taper.y )
	x2 = MatrixTransformation( xformToModelFrameMatrix )
	x1.chainTo( x2 )
	x0.chainTo( x1 )

	return x0




class ProcTransformationTaper (ProcRotatedTransformation):
	description = _( 'Taper' )

	taper = FieldColumn( Vector2, Vector2( 1.0, 1.0 ) )
	lower = FieldColumn( float, 0.0 )
	upper = FieldColumn( float, 100.0 )


	def procedureInvoke(self, inputObject, errorHandler):
		x0 = makeTransformation( inputObject, self.target, self.xformRotation, self.taper, self.lower, self.upper )
		ProcTransformationTaper.transform( inputObject, x0, self.target )
		return inputObject




class TransformationTaperAdjustableInvoker (RotatedTransformationAdjustableInvoker):
	functionClass = ProcTransformationTaper

	def adjust(self):
		x0 = makeTransformation( self.model, self._proc.target, self._proc.xformRotation, self._proc.taper, self._proc.lower, self._proc.upper )
		self.transform( x0, self._proc.target )




class ToolTransformationTaper (ToolRotatedTransformation):
	row = RowRefColumn( ProcTransformationTaper )

	functionClass = ProcTransformationTaper
	pageTitle = _( 'Taper' )
	titleText = _( 'Taper' )
	adjustableInvoker = TransformationTaperAdjustableInvoker


	row_xformRotation = ObjectSourceProxyColumn( row.xformRotation )

	def _p_taperFocalPoint(self):
		centre = self.inputProduct.getBBox( self.row_target == TransformationTarget.POINTS ).getCentre()

		rotationMatrix = Matrix4.rotate( self.row_xformRotation )
		modelToXformFrameMatrix = Matrix4.translate( -centre.toVector3() )  *  rotationMatrix.rigidBodyInverse()
		xformToModelFrameMatrix = modelToXformFrameMatrix.rigidBodyInverse()

		box = self.inputProduct.getBBox( modelToXformFrameMatrix, self.row_target == TransformationTarget.POINTS )
		return Point3( 0.0, 0.0, box.getUpper().z )  *  xformToModelFrameMatrix

	taperFocalPoint = FunctionColumn( _p_taperFocalPoint )



	taperManip = Scale2ColumnManipulator3d( row.taper, taperFocalPoint, row.xformRotation,
									snapping=ColumnManipulatorSnappingFraction() )

	rotationManip = RotationColumnManipulator( row.xformRotation, ToolRotatedTransformation.focalPoint, snapping=ColumnManipulatorSnappingAngle() )

	manipulatorSwitch = ColumnManipulatorSwitch( [ taperManip, rotationManip ] )
	manipulatorSwitch.enable()

	manipulatorSelector = ColumnManipulatorSwitchRadioGroup( manipulatorSwitch )
	taperButton = manipulatorSelector.addChoice( taperManip, TransformationToolActions.deformAction )
	rotationButton = manipulatorSelector.addChoice( rotationManip, TransformationToolActions.rotationAction )


	rotationEntry = QuaternionColumnEditEntry( row.xformRotation )
	rotationPopup = QuaternionColumnEditDirectionPopup( row.xformRotation, _( 'Set rotation' ) )
	taperEntry = Vector2ColumnEditEntry( row.taper, -1.0e8, 1.0e8, 0.01, 0.1, 4 )

	lowerEntry = FloatColumnEditEntry( row.lower, 0.0, 100.0, 0.01, 1.0, 4 )
	upperEntry = FloatColumnEditEntry( row.upper, 0.0, 100.0, 0.01, 1.0, 4 )

	lowerSlider = FloatColumnEditSlider( row.lower, 0.0, 100.0, 0.01, 1.0 )
	upperSlider = FloatColumnEditSlider( row.upper, 0.0, 100.0, 0.01, 1.0 )



	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Mouse function:' )
	layout.row()  <<  taperButton
	layout.row()  <<  rotationButton
	layout.row()
	layout.row()  <<  _( 'Rotation:' )
	layout.row()  <<  rotationPopup.popup
	layout.row()  <<  'X:'  <<  rotationEntry.x
	layout.row()  <<  'Y:'  <<  rotationEntry.y
	layout.row()  <<  'Z:'  <<  rotationEntry.z
	layout.row()
	layout.row()  <<  _( 'Taper:' )
	layout.row()  <<  'U:'  <<  taperEntry.x
	layout.row()  <<  'V:'  <<  taperEntry.y
	layout.row()
	layout.row()  <<  _( 'Lower (%%):' )  <<  lowerEntry.entry
	layout.row()  <<  lowerSlider.slider
	layout.row()  <<  _( 'Upper (%%):' )  <<  upperEntry.entry
	layout.row()  <<  upperSlider.slider
	layout.row()
	layout.finalise()

