##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
import pygtk
pygtk.require( '2.0' )
import gtk

import math

from gSculpt.Math.Math import Axis, Quaternion, Vector3, Colour3f

from gSculpt.Kernel import Variable

from gSculpt.UI import RadioGroup, ButtonPopupMenu, ComboBox, GtkAccelEntry




def floatToInt(v):
	return int( v + 0.5 )

def floatToFloat(v):
	return v




class VarEdit (object):
	def __init__(self):
		super( VarEdit, self ).__init__()
		self._sensitivityVariable = None
		self._bInvertSensitivity = False
		self._bUserSensitivity = True


	def setSensitive(self, bSensitive):
		self._bUserSensitivity = bSensitive
		self._o_refreshSensitivity()

	def _o_setUISensitivity(self, bSensitive):
		assert False, 'abstract'


	def _o_refreshSensitivity(self):
		bSensitive = self._bUserSensitivity  and  self._o_computeSensitivity()
		self._o_setUISensitivity( bSensitive )


	def _o_computeSensitivity(self):
		assert False, 'abstract'


	def attachSensitivityVariable(self, sensitive, bInvert):
		assert self._sensitivityVariable is None, 'sensitivity variable already attached'
		self._sensitivityVariable = sensitive
		sensitive.valueSignal.connect( self._p_onSensitivityVar )
		self._bInvertSensitivity = bInvert
		self._p_onSensitivityVar( sensitive.value )

	def detachSensitivityVariable(self, sensitive):
		assert self._sensitivityVariable is not None, 'not sensitivity variable attached'
		self._o_setUISensitivity( True )
		sensitive.valueSignal.disconnect( self._p_onSensitivityVar )
		self._sensitivityVariable = None


	def _p_onSensitivityVar(self, value):
		if self._bInvertSensitivity:
			value = not value
		self.setSensitive( value )


	def _o_watchVarValidity(self, var):
		var.validitySignal.connect( self._o_onVarValidity )

	def _o_stopWatchingVarValidity(self, var):
		var.validitySignal.disconnect( self._o_onVarValidity )

	def _o_onVarValidity(self, bValid):
		self._o_refreshSensitivity()




class WidgetVarEdit (VarEdit):
	def show(self):
		assert False, 'abstract'

	def hide(self):
		assert False, 'abstract'




class WidgetSingleVarEdit (WidgetVarEdit):
	def __init__(self):
		super( WidgetSingleVarEdit, self ).__init__()
		self._var = None


	def attachVariable(self, var):
		assert self._var is None, 'variable already attached'
		self._var = var
		self._o_watchVarValidity( self._var )
		self._o_refreshSensitivity()

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._o_stopWatchingVarValidity( self._var )
		self._var = None
		self._o_refreshSensitivity()


	def _o_computeSensitivity(self):
		if self._var is not None:
			return self._var.isValid()
		else:
			return False



class BoolVarEditCheck (WidgetSingleVarEdit):
	"""Boolean variable editor - check box
	@check - the check box widget"""

	def __init__(self, labelText):
		super( BoolVarEditCheck, self ).__init__()
		self.check = gtk.CheckButton( labelText )
		self.check.connect( 'clicked', self._p_onCheck )
		self._bIgnoreCheck = False
		self._o_setUISensitivity( False )


	def attachVariable(self, var):
		super( BoolVarEditCheck, self ).attachVariable( var )
		if self._var.isValid():
			self._p_onVarValue( self._var.value )
		self._var.valueSignal.connect( self._p_onVarValue )

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._var.valueSignal.disconnect( self._p_onVarValue )
		super( BoolVarEditCheck, self ).detachVariable()


	def show(self):
		self.check.show()

	def hide(self):
		self.check.hide()


	def _o_setUISensitivity(self, bSensitive):
		self.check.set_sensitive( bSensitive )


	def _p_onCheck(self, check):
		if not self._bIgnoreCheck:
			assert self._var is not None
			value = self.check.get_active()
			if value != self._var.value:
				self._var.value = value

	def _p_onVarValue(self, value):
		if value is not None:
			self._bIgnoreCheck = True
			self.check.set_active( value )
			self._bIgnoreCheck = False




class NumVarEditEntry (WidgetSingleVarEdit):
	"""Numeric variable editor - entry
	@entry - the entry widget"""
	def __init__(self, minValue, maxValue, step, page, numDigits, valueConverter):
		super( NumVarEditEntry, self ).__init__()
		self.entry = gtk.SpinButton( None, climb_rate = 0.15, digits = numDigits )
		self.entry.set_increments( step, page )
		self.entry.set_range( minValue, maxValue )
		self.entry.connect( 'value-changed', self._p_onSpin )
		self._bIgnoreSpin = False
		self._valueConverter = valueConverter
		self._o_setUISensitivity( False )


	def attachVariable(self, var):
		super( NumVarEditEntry, self ).attachVariable( var )
		if self._var.isValid():
			self._p_onVarValue( self._var.value )
		self._var.valueSignal.connect( self._p_onVarValue )

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._var.valueSignal.disconnect( self._p_onVarValue )
		super( NumVarEditEntry, self ).detachVariable()


	def show(self):
		self.entry.show()

	def hide(self):
		self.entry.hide()


	def _o_setUISensitivity(self, bSensitive):
		self.entry.set_sensitive( bSensitive )


	def _p_onSpin(self, spinButton, data = None):
		if not self._bIgnoreSpin:
 			assert self._var is not None
			value = self._valueConverter( self.entry.get_value() )
			if value != self._var.value:
				self._var.value = value

	def _p_onVarValue(self, value):
		if value is not None:
			self._bIgnoreSpin = True
			self.entry.set_value( value )
			self._bIgnoreSpin = False


class IntVarEditEntry (NumVarEditEntry):
	"""Integer variable editor - entry
	@entry - the entry widget"""
	def __init__(self, minValue, maxValue, step, page):
		NumVarEditEntry.__init__(self, minValue, maxValue, step, page, 0, floatToInt )

class FloatVarEditEntry (NumVarEditEntry):
	"""Floating point variable editor - entry
	@entry - the entry widget"""
	def __init__(self, minValue, maxValue, step, page, numDigits):
		NumVarEditEntry.__init__(self, minValue, maxValue, step, page, numDigits, floatToFloat )




class NumVarEditSlider (WidgetSingleVarEdit):
	"""Numeric variable editor - slider
	@slider - the slider widget"""
	def __init__(self, minValue, maxValue, step, page, valueConverter):
		super( NumVarEditSlider, self ).__init__()
		self.slider = gtk.HScale()
		self.slider.set_increments( step, page )
		self.slider.set_range( minValue, maxValue )
		self.slider.set_draw_value( False )
		self.slider.set_size_request( 125, -1 )
		self.slider.connect( 'value-changed', self._p_onScale )
		self._bIgnoreScale = False
		self._valueConverter = valueConverter
		self._o_setUISensitivity( False )


	def attachVariable(self, var):
		super( NumVarEditSlider, self ).attachVariable( var )
		if self._var.isValid():
			self._p_onVarValue( self._var.value )
		self._var.valueSignal.connect( self._p_onVarValue )

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._var.valueSignal.disconnect( self._p_onVarValue )
		super( NumVarEditSlider, self ).detachVariable()


	def show(self):
		self.slider.show()

	def hide(self):
		self.slider.hide()


	def _o_setUISensitivity(self, bSensitive):
		self.slider.set_sensitive( bSensitive )


	def _p_onScale(self, scale, data = None):
		if not self._bIgnoreScale:
 			assert self._var is not None
			value = self._valueConverter( self.slider.get_value() )
			if value != self._var.value:
				self._var.value = value

	def _p_onVarValue(self, value):
		if value is not None:
			self._bIgnoreScale = True
			self.slider.set_value( value )
			self._bIgnoreScale = False



class IntVarEditSlider (NumVarEditSlider):
	"""Integer variable editor - slider
	@slider - the slider widget"""
	def __init__(self, minValue, maxValue, step, page):
		NumVarEditSlider.__init__(self, minValue, maxValue, step, page, floatToInt )

class FloatVarEditSlider (NumVarEditSlider):
	"""Floating point variable editor - slider
	@slider - the slider widget"""
	def __init__(self, minValue, maxValue, step, page):
		NumVarEditSlider.__init__(self, minValue, maxValue, step, page, floatToFloat )






class StringVarEditEntry (WidgetSingleVarEdit):
	"""String variable editor - entry
	@entry - the entry widget"""
	def __init__(self):
		super( StringVarEditEntry, self ).__init__()
		self.entry = gtk.Entry()
		self.entry.connect( 'activate', self._p_onActivate )
		self._bIgnoreEntry = False
		self._o_setUISensitivity( False )


	def attachVariable(self, var):
		super( StringVarEditEntry, self ).attachVariable( var )
		if self._var.isValid():
			self._p_onVarValue( self._var.value )
		self._var.valueSignal.connect( self._p_onVarValue )

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._var.valueSignal.disconnect( self._p_onVarValue )
		super( StringVarEditEntry, self ).detachVariable()


	def show(self):
		self.entry.show()

	def hide(self):
		self.entry.hide()


	def _o_setUISensitivity(self, bSensitive):
		self.entry.set_sensitive( bSensitive )


	def _p_onActivate(self, entry, data = None):
		if not self._bIgnoreEntry:
 			assert self._var is not None
			value = self.entry.get_text()
			if value != self._var.value:
				self._var.value = value

	def _p_onVarValue(self, value):
		if value is not None:
			self._bIgnoreEntry = True
			self.entry.set_text( value )
			self._bIgnoreEntry = False





class StringVarViewLabel (WidgetSingleVarEdit):
	"""String variable view - label
	@entry - the entry widget"""
	def __init__(self):
		super( StringVarViewLabel, self ).__init__()
		self.label = gtk.Label()
		self._o_setUISensitivity( False )


	def attachVariable(self, var):
		super( StringVarViewLabel, self ).attachVariable( var )
		if self._var.isValid():
			self._p_onVarValue( self._var.value )
		self._var.valueSignal.connect( self._p_onVarValue )

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._var.valueSignal.disconnect( self._p_onVarValue )
		super( StringVarViewLabel, self ).detachVariable()


	def show(self):
		self.label.show()

	def hide(self):
		self.label.hide()


	def _o_setUISensitivity(self, bSensitive):
		self.label.set_sensitive( bSensitive )


	def _p_onVarValue(self, value):
		if value is not None:
			self.label.set_text( value )




class StringVarEditAccelEntry (WidgetSingleVarEdit):
	"""String variable editor - accelerator entry
	@entry - the entry widget"""
	def __init__(self):
		super( StringVarEditAccelEntry, self ).__init__()
		self.entry = GtkAccelEntry.GtkAccelEntry()
		self.entry.connect( 'accel-set', self._p_onAccel )
		self._bIgnoreEntry = False
		self._o_setUISensitivity( False )


	def attachVariable(self, var):
		super( StringVarEditAccelEntry, self ).attachVariable( var )
		if self._var.isValid():
			self._p_onVarValue( self._var.value )
		self._var.valueSignal.connect( self._p_onVarValue )

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._var.valueSignal.disconnect( self._p_onVarValue )
		super( StringVarEditAccelEntry, self ).detachVariable()


	def show(self):
		self.entry.show()

	def hide(self):
		self.entry.hide()


	def _o_setUISensitivity(self, bSensitive):
		self.entry.set_sensitive( bSensitive )


	def _p_onAccel(self, entry, value, data = None):
		if not self._bIgnoreEntry:
 			assert self._var is not None
			if value != self._var.value:
				self._var.value = value

	def _p_onVarValue(self, value):
		if value is not None:
			self._bIgnoreEntry = True
			self.entry.set_text( value )
			self._bIgnoreEntry = False







class Vector2OrPoint2VarEditEntry (WidgetSingleVarEdit):
	"""Vector2 or Point2 variable editor - entry
	@x - the X entry
	@y - the Y entry
	@z - the Z entry"""
	def __init__(self, minValue, maxValue, step, page, numDigits):
		super( Vector2OrPoint2VarEditEntry, self ).__init__()
		self._var = None

		self.x = gtk.SpinButton( None, climb_rate = 0.15, digits = numDigits )
		self.x.set_increments( step, page )
		self.x.set_range( minValue, maxValue )
		self.x.connect( 'value-changed', self._p_onXSpin )

		self.y = gtk.SpinButton( None, climb_rate = 0.15, digits = numDigits )
		self.y.set_increments( step, page )
		self.y.set_range( minValue, maxValue )
		self.y.connect( 'value-changed', self._p_onYSpin )

		self._bIgnoreSpin = False

		self._o_setUISensitivity( False )


	def attachVariable(self, var):
		super( Vector2OrPoint2VarEditEntry, self ).attachVariable( var )
		if self._var.isValid():
			self._p_onVarValue( self._var.value )
		self._var.valueSignal.connect( self._p_onVarValue )

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._var.valueSignal.disconnect( self._p_onVarValue )
		super( Vector2OrPoint2VarEditEntry, self ).detachVariable()


	def show(self):
		self.x.show()
		self.y.show()

	def hide(self):
		self.x.hide()
		self.y.hide()


	def _o_setUISensitivity(self, bSensitive):
		self.x.set_sensitive( bSensitive )
		self.y.set_sensitive( bSensitive )


	def _p_onXSpin(self, spinButton, data = None):
		if not self._bIgnoreSpin:
 			assert self._var is not None
			varValue = self._var.getValue()
			value = self.x.get_value()
			if value != varValue.x:
				varValue.x = value
				self._var.value = varValue

	def _p_onYSpin(self, spinButton, data = None):
		if not self._bIgnoreSpin:
 			assert self._var is not None
			varValue = self._var.getValue()
			value = self.y.get_value()
			if value != varValue.y:
				varValue.y = value
				self._var.value = varValue


	def _p_onVarValue(self, value):
		if value is not None:
			self._bIgnoreSpin = True
			self.x.set_value( value.x )
			self.y.set_value( value.y )
			self._bIgnoreSpin = False


class Vector2VarEditEntry (Vector2OrPoint2VarEditEntry):
	"""Vector2 variable editor - entry
	@x - the X entry
	@y - the Y entry
	@z - the Z entry"""
	pass


class Point2VarEditEntry (Vector2OrPoint2VarEditEntry):
	"""Point2 variable editor - entry
	@x - the X entry
	@y - the Y entry
	@z - the Z entry"""
	pass




class Vector3OrPoint3VarEditEntry (WidgetSingleVarEdit):
	"""Vector3 or Point3 variable editor - entry
	@x - the X entry
	@y - the Y entry
	@z - the Z entry"""
	def __init__(self, minValue, maxValue, step, page, numDigits):
		super( Vector3OrPoint3VarEditEntry, self ).__init__()
		self._var = None

		self.x = gtk.SpinButton( None, climb_rate = 0.15, digits = numDigits )
		self.x.set_increments( step, page )
		self.x.set_range( minValue, maxValue )
		self.x.connect( 'value-changed', self._p_onXSpin )

		self.y = gtk.SpinButton( None, climb_rate = 0.15, digits = numDigits )
		self.y.set_increments( step, page )
		self.y.set_range( minValue, maxValue )
		self.y.connect( 'value-changed', self._p_onYSpin )

		self.z = gtk.SpinButton( None, climb_rate = 0.15, digits = numDigits )
		self.z.set_increments( step, page )
		self.z.set_range( minValue, maxValue )
		self.z.connect( 'value-changed', self._p_onZSpin )

		self._bIgnoreSpin = False

		self._o_setUISensitivity( False )


	def attachVariable(self, var):
		super( Vector3OrPoint3VarEditEntry, self ).attachVariable( var )
		if self._var.isValid():
			self._p_onVarValue( self._var.value )
		self._var.valueSignal.connect( self._p_onVarValue )

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._var.valueSignal.disconnect( self._p_onVarValue )
		super( Vector3OrPoint3VarEditEntry, self ).detachVariable()


	def show(self):
		self.x.show()
		self.y.show()
		self.z.show()

	def hide(self):
		self.x.hide()
		self.y.hide()
		self.z.hide()


	def _o_setUISensitivity(self, bSensitive):
		self.x.set_sensitive( bSensitive )
		self.y.set_sensitive( bSensitive )
		self.z.set_sensitive( bSensitive )


	def _p_onXSpin(self, spinButton, data = None):
		if not self._bIgnoreSpin:
 			assert self._var is not None
			varValue = self._var.getValue()
			value = self.x.get_value()
			if value != varValue.x:
				varValue.x = value
				self._var.value = varValue

	def _p_onYSpin(self, spinButton, data = None):
		if not self._bIgnoreSpin:
 			assert self._var is not None
			varValue = self._var.getValue()
			value = self.y.get_value()
			if value != varValue.y:
				varValue.y = value
				self._var.value = varValue

	def _p_onZSpin(self, spinButton, data = None):
		if not self._bIgnoreSpin:
 			assert self._var is not None
			varValue = self._var.getValue()
			value = self.z.get_value()
			if value != varValue.z:
				varValue.z = value
				self._var.value = varValue


	def _p_onVarValue(self, value):
		if value is not None:
			self._bIgnoreSpin = True
			self.x.set_value( value.x )
			self.y.set_value( value.y )
			self.z.set_value( value.z )
			self._bIgnoreSpin = False


class Vector3VarEditEntry (Vector3OrPoint3VarEditEntry):
	"""Vector3 variable editor - entry
	@x - the X entry
	@y - the Y entry
	@z - the Z entry"""
	pass


class Point3VarEditEntry (Vector3OrPoint3VarEditEntry):
	"""Point3 variable editor - entry
	@x - the X entry
	@y - the Y entry
	@z - the Z entry"""
	pass




class QuaternionVarEditEntry (WidgetSingleVarEdit):
	"""Quaternion variable editor - entry
	@x - the X entry
	@y - the Y entry
	@z - the Z entry"""
	def __init__(self):
		super( QuaternionVarEditEntry, self ).__init__()
		self._var = None

		self.x = gtk.SpinButton( None, climb_rate = 0.15, digits = 4 )
		self.x.set_increments( 0.1, 10.0 )
		self.x.set_range( -180.0, 180.0 )
		self.x.connect( 'value-changed', self._p_onXSpin )

		self.y = gtk.SpinButton( None, climb_rate = 0.15, digits = 4 )
		self.y.set_increments( 0.1, 10.0 )
		self.y.set_range( -180.0, 180.0 )
		self.y.connect( 'value-changed', self._p_onYSpin )

		self.z = gtk.SpinButton( None, climb_rate = 0.15, digits = 4 )
		self.z.set_increments( 0.1, 10.0 )
		self.z.set_range( -180.0, 180.0 )
		self.z.connect( 'value-changed', self._p_onZSpin )

		self._bIgnoreSpin = False

		self._o_setUISensitivity( False )


	def attachVariable(self, var):
		super( QuaternionVarEditEntry, self ).attachVariable( var )
		if self._var.isValid():
			self._p_onVarValue( self._var.value )
		self._var.valueSignal.connect( self._p_onVarValue )

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._var.valueSignal.disconnect( self._p_onVarValue )
		super( QuaternionVarEditEntry, self ).detachVariable()


	def show(self):
		self.x.show()
		self.y.show()
		self.z.show()

	def hide(self):
		self.x.hide()
		self.y.hide()
		self.z.hide()


	def _o_setUISensitivity(self, bSensitive):
		self.x.set_sensitive( bSensitive )
		self.y.set_sensitive( bSensitive )
		self.z.set_sensitive( bSensitive )


	def _p_onXSpin(self, spinButton, data = None):
		if not self._bIgnoreSpin:
 			assert self._var is not None
			q = self._var.value
			euler = q.toEuler()
			euler.x = math.radians( -spinButton.get_value() )
			self._var.value = Quaternion.rotateEuler( euler )

	def _p_onYSpin(self, spinButton, data = None):
		if not self._bIgnoreSpin:
 			assert self._var is not None
			q = self._var.value
			euler = q.toEuler()
			euler.y = math.radians( -spinButton.get_value() )
			self._var.value = Quaternion.rotateEuler( euler )

	def _p_onZSpin(self, spinButton, data = None):
		if not self._bIgnoreSpin:
 			assert self._var is not None
			q = self._var.value
			euler = q.toEuler()
			euler.z = math.radians( -spinButton.get_value() )
			self._var.value = Quaternion.rotateEuler( euler )


	def _p_onVarValue(self, value):
		if value is not None:
			self._bIgnoreSpin = True
			euler = value.toEuler()
			self.x.set_value( math.degrees( -euler.x ) )
			self.y.set_value( math.degrees( -euler.y ) )
			self.z.set_value( math.degrees( -euler.z ) )
			self._bIgnoreSpin = False





class QuaternionVarEditDirectionPopup (WidgetSingleVarEdit):
	"""Quaternion variable editor - direction popup menu
	@popup - popup menu button"""
	def __init__(self, modellerView, labelText):
		super( QuaternionVarEditDirectionPopup, self ).__init__()
		self._var = None

		self.popup = ButtonPopupMenu.ButtonPopupMenu( labelText=labelText )
		self.popup.menuSignal.connect( self._p_onPopup )
		self.popup.addMenuItem( Vector3( -1.0, 0.0, 0.0 ), '-X' )
		self.popup.addMenuItem( Vector3( 1.0, 0.0, 0.0 ), '+X' )
		self.popup.addMenuItem( Vector3( 0.0, -1.0, 0.0 ), '-Y' )
		self.popup.addMenuItem( Vector3( 0.0, 1.0, 0.0 ), '+Y' )
		self.popup.addMenuItem( Vector3( 0.0, 0.0, -1.0 ), '-Z' )
		self.popup.addMenuItem( Vector3( 0.0, 0.0, 1.0 ), '+Z' )
		self.popup.addMenuItem( lambda: modellerView.currentViewport.camera.cameraPosition.forwardAxis, _( 'Camera forward' ) )
		self.popup.addMenuItem( lambda: -modellerView.currentViewport.camera.cameraPosition.forwardAxis, _( 'Camera back' ) )
		self.popup.addMenuItem( lambda: -modellerView.currentViewport.camera.cameraPosition.rightAxis, _( 'Camera left' ) )
		self.popup.addMenuItem( lambda: modellerView.currentViewport.camera.cameraPosition.rightAxis, _( 'Camera right' ) )
		self.popup.addMenuItem( lambda: -modellerView.currentViewport.camera.cameraPosition.upAxis, _( 'Camera down' ) )
		self.popup.addMenuItem( lambda: modellerView.currentViewport.camera.cameraPosition.upAxis, _( 'Camera up' ) )

		self._o_setUISensitivity( False )


	def show(self):
		self.popup.getWidget().show()

	def hide(self):
		self.popup.getWidget().hide()


	def _o_setUISensitivity(self, bSensitive):
		self.popup.getWidget().set_sensitive( bSensitive )


	def _p_onPopup(self, event, choice):
		if event.bUserEvent:
			direction = None
			if isinstance( choice, Vector3 ):
				direction = choice
			elif callable( choice ):
				direction = choice()
			else:
				raise TypeError, 'choice is invalid'
			self._var.value = Quaternion.axisToAxis( Vector3( 0.0, 0.0, 1.0 ), direction )





class Colour3fVarEditColourButton (WidgetSingleVarEdit):
	"""Colour variable editor - colour button
	@colourButton - the colour button widget"""
	def __init__(self):
		super( Colour3fVarEditColourButton, self ).__init__()
		self.colourButton = gtk.ColorButton()
		self.colourButton.connect( 'color-set', self._p_onColourSet )
		self._bIgnoreButton = False
		self._o_setUISensitivity( False )


	def attachVariable(self, var):
		super( Colour3fVarEditColourButton, self ).attachVariable( var )
		if self._var.isValid():
			self._p_onVarValue( self._var.value )
		self._var.valueSignal.connect( self._p_onVarValue )

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._var.valueSignal.disconnect( self._p_onVarValue )
		super( Colour3fVarEditColourButton, self ).detachVariable()


	def show(self):
		self.colourButton.show()

	def hide(self):
		self.colourButton.hide()


	def _o_setUISensitivity(self, bSensitive):
		self.colourButton.set_sensitive( bSensitive )


	def _p_onColourSet(self, colourButton, data = None):
		if not self._bIgnoreButton:
 			assert self._var is not None
			gdkColour = self.colourButton.get_color()
			value = Colour3f( gdkColour.red / 65535.0,  gdkColour.green / 65535.0,  gdkColour.blue / 65535.0 )
			if value != self._var.value:
				self._var.value = value

	def _p_onVarValue(self, value):
		if value is not None:
			self._bIgnoreButton = True
			gdkColour = gtk.gdk.Color( int( value.r * 65535.0 ), int( value.g * 65535.0 ), int( value.b * 65535.0 ) )
			self.colourButton.set_color( gdkColour )
			self._bIgnoreButton = False





class ChoiceVarEditRadioGroup (WidgetSingleVarEdit):
	"""Choice variable editor - radio group"""
	def __init__(self):
		super( ChoiceVarEditRadioGroup, self ).__init__()
		self._var = None
		self._radioGroup = RadioGroup.RadioGroup()
		self._radioGroup.radioSignal.connect( self._p_onRadio )

		self._o_setUISensitivity( False )


	def attachVariable(self, var):
		super( ChoiceVarEditRadioGroup, self ).attachVariable( var )
		if self._var.isValid():
			self._p_onVarValue( self._var.value )
		self._var.valueSignal.connect( self._p_onVarValue )

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._var.valueSignal.disconnect( self._p_onVarValue )
		super( ChoiceVarEditRadioGroup, self ).detachVariable()


	def addChoice(self, choice, labelText=None, gsImageFilename=None, contents=None):
		return self._radioGroup.addRadioButton( choice, labelText, gsImageFilename, contents )


	def show(self):
		self._radioGroup.show()

	def hide(self):
		self._radioGroup.hide()


	def getButtonForChoice(self, choice):
		return self._radioGroup.getButtonForChoice( choice )

	def getButtonByIndex(self, index):
		return self._radioGroup.getButtonByIndex( index )


	def _o_setUISensitivity(self, bSensitive):
		self._radioGroup.setSensitive( bSensitive )


	def _p_onRadio(self, event, choice):
		assert self._var is not None
		if event.bUserEvent:
			if choice != self._var.value:
				self._var.value = choice

	def _p_onVarValue(self, value):
		self._radioGroup.currentChoice = value



class AxisVarEditRadioGroup (ChoiceVarEditRadioGroup):
	def __init__(self):
		super( AxisVarEditRadioGroup, self ).__init__()
		self.x = self.addChoice( Axis.AXIS_X, 'X' )
		self.y = self.addChoice( Axis.AXIS_Y, 'Y' )
		self.z = self.addChoice( Axis.AXIS_Z, 'Z' )
		self.negativeX = self.addChoice( Axis.AXIS_NEGATIVE_X, '-X' )
		self.negativeY = self.addChoice( Axis.AXIS_NEGATIVE_Y, '-Y' )
		self.negativeZ = self.addChoice( Axis.AXIS_NEGATIVE_Z, '-Z' )






class ChoiceVarEditComboBox (WidgetSingleVarEdit):
	"""Choice variable editor - radio group"""
	def __init__(self, columnSpecs=[ str ], choices=[]):
		super( ChoiceVarEditComboBox, self ).__init__()
		self._var = None
		self.comboBox = ComboBox.ComboBox( columnSpecs, choices )
		self.comboBox.choiceSignal.connect( self._p_onCombo )

		self._o_setUISensitivity( False )



	def attachVariable(self, var):
		super( ChoiceVarEditComboBox, self ).attachVariable( var )
		if self._var.isValid():
			self._p_onVarValue( self._var.value )
		self._var.valueSignal.connect( self._p_onVarValue )

	def detachVariable(self):
		assert self._var is not None, 'no variable attached'
		self._var.valueSignal.disconnect( self._p_onVarValue )
		super( ChoiceVarEditComboBox, self ).detachVariable()


	def addChoice(self, choice, labelText=None, gsImageFilename=None, contents=None):
		return self._radioGroup.addRadioButton( choice, labelText, gsImageFilename, contents )



	def append(self, choice, *row):
		self.comboBox.append( choice, *row )

	def insert(self, index, choice, *row):
		self.comboBox.insert( index, choice, *row )

	def remove(self, choice):
		self.comboBox.remove( choice )

	def clear(self):
		self.comboBox.clear()



	def show(self):
		self.comboBox.getWidget().show()

	def hide(self):
		self.comboBox.getWidget().hide()


	def _o_setUISensitivity(self, bSensitive):
		self.comboBox.getWidget().set_sensitive( bSensitive )


	def _p_onCombo(self, event, choice):
		assert self._var is not None
		if event.bUserEvent:
			if choice != self._var.value:
				self._var.value = choice


	def _p_onVarValue(self, value):
		self.comboBox.choice = value












if __name__ == '__main__':
	from gSculpt.UI import FormLayout

	class HelloWorld:
		def delete_event(self, widget, event, data=None):
			print( "delete_event" )
			return gtk.FALSE

		def destroy(self, widget, data=None):
			gtk.main_quit()

		def onValue(self,value):
			pass

		def __init__(self):
			self.var = Variable.FloatVar( 10.0 )
			self.var.valueSignal.connect( self.onValue )

			self.window = gtk.Window( gtk.WINDOW_TOPLEVEL );
			self.window.connect( 'delete-event', self.delete_event )
			self.window.connect( 'destroy', self.destroy )
			self.window.set_border_width( 10 )

			self.entry = FloatVarEditEntry( -100.0, 100.0, 1.0, 10.0, 4 )
			self.entry.attachVariable( self.var )
			self.entry.getWidget().show()

			self.slider = FloatVarEditSlider( -100.0, 100.0, 1.0, 10.0 )
			self.slider.attachVariable( self.var )
			self.slider.getWidget().show()


			self.layout = FormLayout.FormLayout()
			self.layout.row() << 'Entry' << self.entry.getWidget()
			self.layout.row() << 'Slider' << self.slider.getWidget()
			self.layout.row() << 'A' << 'B' << 'C'
			self.layout.finalise()
			self.layout.getWidget().show()

			self.window.add( self.layout.getWidget() );
			self.window.show();

		def main(self):
			gtk.main()

	hello = HelloWorld()
	hello.main()



