##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from OpenGL.GL import *

from gSculpt.FileIO.IOXml import ioObjectFactoryRegister

from gSculpt.Graphics.Graphics import glMultMatrix

from gSculpt.Util.SignalSlot import ClassSignal

from gSculpt.Math.Math import Point3, Quaternion, Vector2, Matrix4
from gSculpt.Math.MathField import Point3Field, QuaternionField, Vector2Field

from gSculpt.Kernel.DB import Row, FieldColumn, FloatField, BoolField

from gSculpt.Texture.TextureRef import TextureRef
from gSculpt.Texture.TextureRefVar import TextureRefField

from gSculpt.WorkArea.Viewport3d import Viewport3dPainter

from gSculpt.View.View import PaintLayer
from gSculpt.View import BackgroundImageCommandTracker




class BackgroundImage (Row):
	trackerClass = BackgroundImageCommandTracker.BackgroundImageCommandTracker

	texture = FieldColumn( TextureRef, TextureRef() )
	position = FieldColumn( Point3, Point3() )
	rotation = FieldColumn( Quaternion, Quaternion() )
	size = FieldColumn( Vector2, Vector2( 100.0, 100.0 ) )
	brightness = FieldColumn( float, 1.0 )
	opacity = FieldColumn( float, 1.0 )
	bOverlay = FieldColumn( bool, False )
	bVisible = FieldColumn( bool, True )
	bBackfaceCull = FieldColumn( bool, False )

	textureSignal = ClassSignal()


	def __init__(self):
		super( BackgroundImage, self ).__init__()
		self.fields.texture.valueSignal.connect( self._p_onTexture )



	def paint3d(self, paintLayer):
		if self.bVisible:
			if ( paintLayer == PaintLayer.PAINTLAYER_OVERLAY_BLEND and not self.bOverlay )  or  ( paintLayer == PaintLayer.PAINTLAYER_OVERLAY_BLEND_NODEPTH and self.bOverlay ):
				if self.bBackfaceCull:
					glCullFace( GL_BACK )
					glEnable( GL_CULL_FACE )

				matrix = Matrix4.scale( self.size )  *  Matrix4.rotate( self.rotation )  *  Matrix4.translate( self.position.toVector3() )
				glPushMatrix()
				glMultMatrix( matrix )

				tex = None
				if self.texture is not None:
					tex = self.texture.texture

				if tex is not None:
					glEnable( GL_TEXTURE_2D )
					glBindTexture( GL_TEXTURE_2D, tex.glTextureHandle )

				b = self.brightness

				glColor4d( b, b, b, 1.0 - self.opacity )

				if tex is not None:
					glBegin( GL_QUADS )
				else:
					glBegin( GL_LINE_LOOP )

				if tex is not None:
					glTexCoord2d( 0.0, 0.0 )
				glVertex3d( -0.5, -0.5, 0.0 )

				if tex is not None:
					glTexCoord2d( 1.0, 0.0 )
				glVertex3d( 0.5, -0.5, 0.0 )

				if tex is not None:
					glTexCoord2d( 1.0, 1.0 )
				glVertex3d( 0.5, 0.5, 0.0 )

				if tex is not None:
					glTexCoord2d( 0.0, 1.0 )
				glVertex3d( -0.5, 0.5, 0.0 )

				glEnd()


				if tex is not None:
					glDisable( GL_TEXTURE_2D )

				glPopMatrix()

				if self.bBackfaceCull:
					glDisable( GL_CULL_FACE )


	def setSizeToPreferredSize(self):
		self.size = self.computePreferredSize()


	def computePreferredSize(self):
		if self.texture is not None:
			tex = self.texture.texture

			if tex is not None:
				size = tex.size

				if size is not None:
					if size.x > 0  and  size.y > 0:
						if size.x > size.y:
							return Vector2( 100.0, ( size.y * 100.0 )  /  size.x )
						else:
							return Vector2( ( size.x * 100.0 )  /  size.y, 100.0 )
		return Vector2( 100.0, 100.0 )


	def _p_onTexture(self, value):
		self.setSizeToPreferredSize()
		self.textureSignal.emit( self, value )




class BackgroundImageList (Viewport3dPainter):
	trackerClass = BackgroundImageCommandTracker.BackgroundImageListCommandTracker

	imageTextureSignal = ClassSignal()
	imageAddedSignal = ClassSignal()
	imageInsertedSignal = ClassSignal()
	imageRemoveSignal = ClassSignal()
	imageRemovedSignal = ClassSignal()

	def __init__(self):
		self._images = []
		self._viewports = []

		self._commandTracker_ = None


	def newImage(self):
		image = BackgroundImage()
		self.appendImage( image )
		return image


	def appendImage(self, image):
		image.textureSignal.connect( self._p_onImageTexture )
		image.fieldModifiedSignal.connect( self._p_onImageModified )
		self._images.append( image )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onImageListImageAppended( self, image )
		self.imageAddedSignal.emit( self, image )
		self._p_postRedraw()


	def insertImage(self, index, image):
		image.textureSignal.connect( self._p_onImageTexture )
		image.fieldModifiedSignal.connect( self._p_onImageModified )
		self._images.insert( index, image )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onImageListImageInserted( self, index, image )
		self.imageInsertedSignal.emit( self, index, image )
		self._p_postRedraw()


	def removeImage(self, image):
		assert image in self._images
		self.imageRemoveSignal.emit( self, image )
		if self._commandTracker_ is not None:
			self._commandTracker_._f_onImageListImageRemoved( self, image )
		image.fieldModifiedSignal.disconnect( self._p_onImageModified )
		image.textureSignal.disconnect( self._p_onImageTexture )
		self._images.remove( image )
		self.imageRemovedSignal.emit( self, image )
		self._p_postRedraw()


	def clear(self):
		while len( self._images ) > 0:
			self.removeImage( self._images[-1] )


	def evPaint3d(self, viewport, paintLayer):
		for image in self._images:
			image.paint3d( paintLayer )


	def __getitem__(self, index):
		return self._images[index]

	def __len__(self):
		return len( self._images )

	def __iter__(self):
		return iter( self._images )

	def index(self, image):
		return self._images.index( image )



	def __readxml__(self, xmlNode):
		self.clear()
		for node in xmlNode.childrenNamed( 'image' ):
			image = self.newImage()
			node  >>  image

	def __writexml__(self, xmlNode):
		for image in self._images:
			xmlNode.addChild( 'image' )  <<  image



	def painterAttachedToViewport(self, viewport):
		super( BackgroundImageList, self ).painterAttachedToViewport( viewport )
		self._viewports.append( viewport )


	def painterDetachedFromViewport(self, viewport):
		super( BackgroundImageList, self ).painterDetachedFromViewport( viewport )
		self._viewports.remove( viewport )


	def _p_postRedraw(self):
		for viewport in self._viewports:
			viewport.postRedraw()


	def _p_onImageModified(self, image, fieldName):
		self._p_postRedraw()

	def _p_onImageTexture(self, image, texture):
		self.imageTextureSignal.emit( self, image, texture )


ioObjectFactoryRegister( 'BackgroundImageList', BackgroundImageList )

