##-*************************
##-* This program is free software; you can use it, redistribute it and/or modify it
##-* under the terms of the GNU General Public License version 2 as published by the
##-* Free Software Foundation. The full text of the GNU General Public License
##-* version 2 can be found in the file named 'COPYING' that accompanies this
##-* program. This source code is (C)copyright Geoffrey French 1999-2007.
##-*************************
from gSculpt.Math.Math import Point3, Quaternion, Vector2, Matrix4, Colour3f, Segment3
from gSculpt.Math.MathField import Point3Field, QuaternionField

from gSculpt.Kernel.DB import Row, FieldColumn, FunctionColumn
from gSculpt.Kernel.DBEditorTool import RowEditorTool
from gSculpt.Kernel.ColumnEditors import *

from gSculpt.Manipulator.ColumnManipulators import *

from gSculpt.Model.Model import VisualPlane

from gSculpt.Model import ModelDraw

from gSculpt.UI.FormLayout import FormLayoutFactory
from gSculpt.UI.UIAction import UIAction

from gSculpt.WorkArea.Viewport3d import Viewport3dPainter

from gSculpt.View.View import PaintLayer
from gSculpt.View.DrawingTemplate import DrawingTemplate




class DrawingPlane (Row, Viewport3dPainter, DrawingTemplate):
	position = FieldColumn( Point3, Point3() )
	rotation = FieldColumn( Quaternion, Quaternion() )
	size = FieldColumn( float, 100.0 )
	bEnabled = FieldColumn( bool, False )


	def _p_visualPlane(self):
		return VisualPlane( self.position, self.rotation, self.size, self.bEnabled )

	def _p_plane(self):
		return self.visualPlane.plane


	visualPlane = FunctionColumn( _p_visualPlane )
	plane = FunctionColumn( _p_plane )





	def __init__(self):
		super( DrawingPlane, self ).__init__()
		self.fields.visualPlane.changedSignal.connect( self._p_onVisualPlane )
		self._viewports = []



	def evPaint3d(self, viewport, paintLayer):
		self.visualPlane.paint( ModelDraw.DRAWING_PLANE_COLOUR, ModelDraw.DRAWING_PLANE_ALPHA, paintLayer )


	def painterAttachedToViewport(self, viewport):
		super( DrawingPlane, self ).painterAttachedToViewport( viewport )
		self._viewports.append( viewport )


	def painterDetachedFromViewport(self, viewport):
		super( DrawingPlane, self ).painterDetachedFromViewport( viewport )
		self._viewports.remove( viewport )



	def templateDrawingPoint3d(self, ray, bBackfaceCulling):
		if self.bEnabled:
			bSuccess, t, intersection = ray.intersect( self.plane )
			if bSuccess:
				return True, t, intersection, Segment3( ray.a, intersection )
		return False, None, None, None



	def _p_postRedraw(self):
		for viewport in self._viewports:
			viewport.postRedraw()

	def _p_onVisualPlane(self):
		self._p_postRedraw()





_positionAction = UIAction( 'Tools.Main.DrawingPlaneMode.Position', ( _( 'Tools' ), _( 'Main' ), _( 'Drawing plane mode' ), _( 'Position mode' ) ),
			_( 'Position' ), _( 'Position' ), 'w', _( 'Position mode' ) )
_rotationAction = UIAction( 'Tools.Main.DrawingPlaneMode.Rotation', ( _( 'Tools' ), _( 'Main' ), _( 'Drawing plane mode' ), _( 'Rotation mode' ) ),
			_( 'Rotation' ), _( 'Rotation' ), 'r', _( 'Rotation mode' ) )
_sizeAction = UIAction( 'Tools.Main.DrawingPlaneMode.Size', ( _( 'Tools' ), _( 'Main' ), _( 'Drawing plane mode' ), _( 'Size mode' ) ),
			_( 'Size' ), _( 'Size' ), 'e', _( 'Size mode' ) )



class DrawingPlaneTool (RowEditorTool):
	row = RowRefColumn( DrawingPlane )

	rowClass = DrawingPlane
	pageTitle = _( 'DP' )
	titleText = _( 'Drawing plane' )


	positionManip = Point3ColumnManipulator( row.position, snapping=ColumnManipulatorSnappingDistance() )
	rotationManip = RotationColumnManipulator( row.rotation, row.position, snapping=ColumnManipulatorSnappingAngle() )
	sizeManip = Real1ColumnManipulator( row.size, 0.0, 1.0e18, RealColumnManipulator.WorldSpaceScaleFactory( row.position ),
									snapping=ColumnManipulatorSnappingDistance() )

	manipulatorSwitch = ColumnManipulatorSwitch( [ positionManip, rotationManip, sizeManip ] )
	manipulatorSwitch.enable()

	manipulatorSelector = ColumnManipulatorSwitchRadioGroup( manipulatorSwitch )
	positionButton = manipulatorSelector.addChoice( positionManip, _positionAction )
	rotationButton = manipulatorSelector.addChoice( rotationManip, _rotationAction )
	sizeButton = manipulatorSelector.addChoice( sizeManip, _sizeAction )

	positionEntry = Point3ColumnEditEntry( row.position, -1.0e18, 1.0e18, 0.1, 10.0, 4 )
	rotationPopup = QuaternionColumnEditDirectionPopup( row.rotation, _( 'Set normal' ) )
	rotationEntry = QuaternionColumnEditEntry( row.rotation )
	sizeEntry = FloatColumnEditEntry( row.size, 0.0, 1.0e18, 0.1, 10.0, 4 )



	layout = FormLayoutFactory()
	layout.row()  <<  _( 'Mouse function:' )
	layout.row()  <<  positionButton
	layout.row()  <<  rotationButton
	layout.row()  <<  sizeButton
	layout.row()
	layout.row()  <<  _( 'Position:' )
	layout.row()  <<  'X:'  <<  positionEntry.x
	layout.row()  <<  'Y:'  <<  positionEntry.y
	layout.row()  <<  'Z:'  <<  positionEntry.z
	layout.row()
	layout.row()  <<  _( 'Rotation:' )
	layout.row()  <<  rotationPopup.popup
	layout.row()  <<  'X:'  <<  rotationEntry.x
	layout.row()  <<  'Y:'  <<  rotationEntry.y
	layout.row()  <<  'Z:'  <<  rotationEntry.z
	layout.row()
	layout.row()  <<  _( 'Size:' )  <<  sizeEntry.entry
	layout.row()
	layout.finalise()
